/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (SDS), Denmark
 * (http://sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (SDS), Denmark (http://sundhedsdatastyrelsen.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.consumer;

import java.util.List;

import org.openehealth.ipf.commons.ihe.xds.core.metadata.Identifiable;
import org.openehealth.ipf.commons.ihe.xds.core.requests.RetrieveDocument;
import org.openehealth.ipf.commons.ihe.xds.core.requests.query.FindDocumentsQuery;
import org.openehealth.ipf.commons.ihe.xds.core.requests.query.GetDocumentsQuery;
import org.openehealth.ipf.commons.ihe.xds.core.responses.RetrievedDocument;
import org.openehealth.ipf.commons.ihe.xds.core.responses.RetrievedDocumentSet;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.query.AdhocQueryResponse;

/**
 * Enables querying for document metadata on a given patient and subsequent retrieval of the documents themselves.
 * 
 * Metadata come in two guises: a simplified version, {@link SimpleDocumentMetadata}, containing just enough information to do subsequent retrieval, and the more comprehensive {@link DocumentMetadata}
 * which could form the basis for selection (by a user) of documents to retrieve.
 */
public interface DocumentConsumer {

	/**
	 * Creates a FindDocuments query for finding document metadata.
	 * 
	 * @param patientId identification of patient about which to query for document metadata
	 * @return a query for the given patient. The query can be further parametrised using methods on {@link FindDocumentsQuery}.
	 */
	FindDocumentsQuery createQuery(Identifiable patientId);

	/**
	 * Creates a FindDocuments query for finding document metadata.
	 * 
	 * @param patientId identification of patient about which to query for document metadata
	 * @return a query for the given patient. The query can be further parametrised using methods on {@link FindDocumentsQuery}.
	 */
	FindDocumentsQuery createFindDocumentsQuery(Identifiable patientId);

	/**
	 * Creates a GetDocuments query for obtaining document metadata for one or more particular documents.
	 * 
	 * @return a query. The query can be further parametrised using methods on {@link GetDocumentsQuery}.
	 */
	GetDocumentsQuery createGetDocumentsQuery();

	/**
	 * Performs the given query for finding document metadata and returns the full metadata.
	 * 
	 * @param query
	 * @return
	 * @throws Exception
	 */
	AdhocQueryResponse doQuery(FindDocumentsQuery query) throws Exception;

	/**
	 * Performs the given query for finding document metadata and returns the full metadata.
	 * 
	 * @param query
	 * @param patientId identification of patient about which to query for document metadata
	 * @return resulting document metadata matching the query
	 * @throws Exception
	 */
	AdhocQueryResponse doQuery(GetDocumentsQuery query) throws Exception;

	/**
	 * Creates a controller used for retrieving documents by the given document metadata
	 * 
	 * @param toRetrieve
	 * @return
	 */
	List<RetrieveDocument> prepareRetrieve(AdhocQueryResponse toRetrieve);

	/**
	 * Performs retrieval of documents
	 * 
	 * @param toRetrieve specifies which documents to retrieve by document metadata
	 * @return
	 */
	List<RetrievedDocument> doRetrieve(List<RetrieveDocument> toRetrieve) throws Exception;

	/**
	 * Performs retrieval of documents
	 * 
	 * @param toRetrieve specifies which documents to retrieve by document metadata
	 * @return
	 */
	RetrievedDocumentSet doRetrieveAll(List<RetrieveDocument> toRetrieve) throws Exception;
}