/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.integrationtest;

import java.io.File;
import java.util.List;

import javax.xml.ws.soap.SOAPFaultException;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.junit.Assert;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Ignore;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.openehealth.ipf.commons.ihe.xds.core.requests.RetrieveDocument;
import org.openehealth.ipf.commons.ihe.xds.core.requests.query.FindDocumentsQuery;
import org.openehealth.ipf.commons.ihe.xds.core.responses.RetrievedDocument;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.query.AdhocQueryResponse;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.SpringRunner;
import org.springframework.test.context.support.AnnotationConfigContextLoader;

import dk.nsi.dds.projects.sxa.common.configuration.IntegrationTestConfiguration;
import dk.nsi.dds.projects.sxa.common.testutil.DocumentSharingTestUtil;
import dk.nsi.dds.projects.sxa.consumer.DocumentConsumer;
import dk.nsi.dds.projects.sxa.testhelper.SxaLoggerHelper;
import dk.nsi.dds.projects.sxa.testhelper.SxaQueryHelper;
import dk.nsi.dds.projects.sxa.testhelper.SxaTestDataHelper;


/**
 * Tests performed by citizen type user
 * 
 * Tests of SXA documentsharing not affected by presence of other document metadata
 * providers or registered document metadata.
 * 
 * Please note that MappedCodedValue are used for specifying practiceSettingCode values where
 * CodedValue ought to be used. This was chosen to enable easier verification of actual outcome.
 */
@RunWith(SpringRunner.class)
@ContextConfiguration(classes = { IntegrationTestConfiguration.class }, loader = AnnotationConfigContextLoader.class)
public class CitizenQueryAndRetrieveIT {

	@Autowired
	private DocumentConsumer docConsumer;
	@Autowired
	private SxaTestDataHelper testDataHelper;
	@Autowired
	private SxaQueryHelper queryHelper;

	private static final Logger logger = Logger.getLogger(CitizenQueryAndRetrieveIT.class);
	private DocumentSharingTestUtil testutil = new DocumentSharingTestUtil();
	private static String userId = "1112132016";

	@BeforeClass
	public static void setupTestClientProperties() {
		if (System.getProperty("testclient.properties") == null) {
			System.setProperty("testclient.properties", new File("src/test/resources/testclient.properties").getAbsolutePath());
		}
	}

	@Before
	public void setupClass() throws Exception {
		SxaLoggerHelper.setupLogger(Level.DEBUG); // Set to Level.DEBUG to print out
												  // queries, returned metadata etc.
	}

	// ----------------------- Query Metadata Tests ------------------------------------

	/**
	 * Query DocumentMetadata performed by the citizen him/herself. Expects a non-empty response
	 */
	@Test
	public void metadataReturnedForQueryPerformedByCitizen() throws Exception {
		// Arrange
		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getPatientId(userId));
		// Act
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		// Assert
		Assert.assertNotNull(queryResult.getRegistryObjectList());
		Assert.assertTrue(queryResult.getRegistryObjectList().getIdentifiable().size() > 0);
	}

	/**
	 * Query DocumentMetadata performed by the other citizen but with self relation. Expects SoapFaultException
	 */
	@Test
	@Ignore("Denne test kan kun udføres, hvis der er en DDS mellem klient og Labsvar.")
	public void metadataReturnedForQueryPerformedByOtherCitizen() throws Exception {
		// Arrange
		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId());

		// Act
		try {
			docConsumer.doQuery(query);
			Assert.fail("Forventede en fejl ved kald af DDSRegistry");
		} catch (Throwable t) {
			Assert.assertTrue("Forventede en SOAPFaultException", t instanceof SOAPFaultException);
		}
	}

	/**
	 * Query DocumentMetadata performed by parent of the the patient. Expects a non-empty response
	 */
	@Test
	public void metadataReturnedForQueryPerformedByParent() throws Exception {
		// Arrange
		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId());

		// Act
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		// Assert
		Assert.assertNotNull(queryResult.getRegistryObjectList());
		Assert.assertTrue(queryResult.getRegistryObjectList().getIdentifiable().size() > 0);
	}

	/**
	 * Query DocumentMetadata performed by guardian of the the patient. Expects a non-empty response
	 */
	@Test
	public void metadataReturnedForQueryPerformedByGuardian() throws Exception {
		// Arrange
		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId());

		// Act
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		// Assert
		Assert.assertNotNull(queryResult.getRegistryObjectList());
		Assert.assertTrue(queryResult.getRegistryObjectList().getIdentifiable().size() > 0);
	}

	// ----------------------- Retrieve Document(s) Tests ------------------------------------

	/**
	 * Retrieve document performed by patient's proxy holder. Expects a non-empty response
	 */
	@Test
	public void retrieveDocumentsPerformedByProxyHolder() throws Exception {
		// Arrange
		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId());
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		List<RetrieveDocument> fetchList = docConsumer.prepareRetrieve(queryResult);
		// Act
		List<RetrievedDocument> retrievedDocuments = docConsumer.doRetrieve(fetchList);
		// Assert
		logger.debug(testutil.prettyPrint(retrievedDocuments));
		System.out.println(retrievedDocuments.size());
		Assert.assertTrue(retrievedDocuments.size() == 1);
		Assert.assertNotNull(retrievedDocuments);
	}

	/**
	 * Retrieve document performed by patient's parent. Expects a non-empty response
	 */
	@Test
	public void retrieveDocumentsPerformedByParent() throws Exception {
		// Arrange
		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId());
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		List<RetrieveDocument> fetchList = docConsumer.prepareRetrieve(queryResult);
		// Act
		List<RetrievedDocument> retrievedDocuments = docConsumer.doRetrieve(fetchList);
		// Assert
		logger.debug(testutil.prettyPrint(retrievedDocuments));
		Assert.assertNotNull(retrievedDocuments);
		Assert.assertTrue(retrievedDocuments.size() > 0);
	}

	/**
	 * Retrieve document performed by patient's guardian. Expects a non-empty response
	 */
	@Test
	public void retrieveDocumentsPerformedByGuardian() throws Exception {
		// Arrange
		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId());
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		System.out.println(queryResult.getRegistryObjectList().getIdentifiable().get(0).getValue().getId());
		List<RetrieveDocument> fetchList = docConsumer.prepareRetrieve(queryResult);
		for (RetrieveDocument retrieveDocument : fetchList) {
			System.out.println(retrieveDocument.getHomeCommunityId());
			System.out.println(retrieveDocument.getRepositoryUniqueId());
			System.out.println(retrieveDocument.getDocumentUniqueId());
		}
		// Act
		List<RetrievedDocument> retrievedDocuments = docConsumer.doRetrieve(fetchList);
		// Assert
		logger.debug(testutil.prettyPrint(retrievedDocuments));
		Assert.assertNotNull(retrievedDocuments);
		Assert.assertTrue(retrievedDocuments.size() > 0);
	}

	/**
	 * Retrieve document performed by the citizen him/herself. Expects a non-empty response
	 */
	@Test
	public void retrieveDocumentsPerformedBySelf() throws Exception {
		// Arrange
		queryHelper = new SxaQueryHelper(docConsumer, testDataHelper);
		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId());
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		List<RetrieveDocument> fetchList = docConsumer.prepareRetrieve(queryResult);
		// Act
		List<RetrievedDocument> retrievedDocuments = docConsumer.doRetrieve(fetchList);
		// Assert
		logger.debug(testutil.prettyPrint(retrievedDocuments));
		Assert.assertNotNull(retrievedDocuments);
		Assert.assertTrue(retrievedDocuments.size() > 0);
	}
}
