/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.integrationtest;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.joda.time.DateTime;
import org.junit.Assert;
import org.junit.BeforeClass;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.openehealth.ipf.commons.ihe.xds.core.requests.RetrieveDocument;
import org.openehealth.ipf.commons.ihe.xds.core.requests.query.FindDocumentsQuery;
import org.openehealth.ipf.commons.ihe.xds.core.responses.RetrievedDocument;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.query.AdhocQueryResponse;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.SpringRunner;
import org.springframework.test.context.support.AnnotationConfigContextLoader;

import dk.nsi.dds.projects.sxa.common.TestDataHelper;
import dk.nsi.dds.projects.sxa.common.configuration.IntegrationTestConfiguration;
import dk.nsi.dds.projects.sxa.common.configuration.SXAMappedValue;
import dk.nsi.dds.projects.sxa.common.testutil.DocumentSharingTestUtil;
import dk.nsi.dds.projects.sxa.consumer.DocumentConsumer;
import dk.nsi.dds.projects.sxa.testhelper.SxaDocumentHelper;
import dk.nsi.dds.projects.sxa.testhelper.SxaLoggerHelper;
import dk.nsi.dds.projects.sxa.testhelper.SxaQueryHelper;
import dk.nsi.dds.projects.sxa.testhelper.SxaTestDataHelper;


/**
 * Tests of SXA documentsharing not affected by presence of other document metadata providers or registered document metadata.
 * 
 * Please note that SXAMappedValue are used for specifying practiceSettingCode values where CodedValue ought to be used. This was chosen to enable easier verification of actual outcome.
 */
@RunWith(SpringRunner.class)
@ContextConfiguration(classes = { IntegrationTestConfiguration.class }, loader = AnnotationConfigContextLoader.class)
public class QueryAndRetrieveIT {
	private static final Logger logger = Logger.getLogger(QueryAndRetrieveIT.class);

	@Autowired
	private DocumentConsumer docConsumer;
	@Autowired
	private SxaTestDataHelper testDataHelper;
	@Autowired
	private SxaQueryHelper queryHelper;

	
	private DocumentSharingTestUtil testutil = new DocumentSharingTestUtil();
	private static TestDataHelper practiceSettingCodesDataHelper;

	@BeforeClass
	public static void setupClass() throws Exception {
		SxaLoggerHelper.setupLogger(Level.INFO); // Set to Level.DEBUG to print out
													// queries, returned metadata etc.
		logger.info("Info log enabled");

		if (System.getProperty("testclient.properties") == null) {
			System.setProperty("testclient.properties", new File("src/test/resources/testclient.properties").getAbsolutePath());
		}

		practiceSettingCodesDataHelper = new TestDataHelper();
	}

	// ----------------------- Query Metadata Tests ------------------------------------

	/**
	 * Query DocumentMetadata for an unknown patient. Expects an non-empty response
	 * 
	 * @throws Exception
	 */
	@Test
	public void metadataReturnedForUnknownPatient() throws Exception {
		// Arrange
		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getPatientIdForUnknownPatient());
		// Act
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		// Assert
		Assert.assertNotNull(queryResult.getRegistryObjectList().getIdentifiable());
		Assert.assertTrue(queryResult.getRegistryObjectList().getIdentifiable().size() > 0);
	}

	@Test
	public void metadataReturnedForPatientHavingNoLabReports() throws Exception {
		// Arrange
		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getPatientIdForPatientWithNoData());
		// Act
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		// Assert
		Assert.assertNotNull(queryResult.getRegistryObjectList().getIdentifiable());
		Assert.assertTrue(queryResult.getRegistryObjectList().getIdentifiable().size() > 0);
	}

	/**
	 * Verifies that a query: - using query parameters designating SXA-data - with list of only valid practiceSettingCodes results in a single SXA document metadata entry where: - practiceSettingCodes
	 * match the selected practiceSettingCodes
	 */
	@Test
	public void metadataReturnedForMixedValidAndInvalidPracticeSettingCodes() throws Exception {
		// Arrange
		List<SXAMappedValue> mixedValidAndInvalidPracticeSettingCodes = new ArrayList<>();
		mixedValidAndInvalidPracticeSettingCodes.add(practiceSettingCodesDataHelper.practiceSettingCodeForMicrobiology);
		mixedValidAndInvalidPracticeSettingCodes.add(practiceSettingCodesDataHelper.practiceSettingCodeForPathology);
		mixedValidAndInvalidPracticeSettingCodes.add(practiceSettingCodesDataHelper.practiceSettingCodeNonSxa);
		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId(), mixedValidAndInvalidPracticeSettingCodes);
		// Act
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		// Assert
		Assert.assertNotNull(queryResult.getRegistryObjectList().getIdentifiable());
		Assert.assertNotNull(queryResult.getRegistryObjectList().getIdentifiable());
		Assert.assertFalse(queryResult.getRegistryObjectList().getIdentifiable().isEmpty());
		Assert.assertTrue(queryResult.getRegistryObjectList().getIdentifiable().size() > 0);
	}

	/**
	 * Query DocumentMetadata with a list of only invalid PracticeSettingCodes. Expects an empty response
	 * 
	 * @throws Exception
	 */
	@Test
	public void metadataReturnedForOnlyInvalidPracticeSettingCodes() throws Exception {
		// Arrange
		List<SXAMappedValue> selectedInvalidPracticeSettingCodes = new ArrayList<>();
		selectedInvalidPracticeSettingCodes.add(practiceSettingCodesDataHelper.practiceSettingCodeNonSxa);
		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId(), selectedInvalidPracticeSettingCodes);
		// Act
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		// Assert
		Assert.assertNotNull(queryResult.getRegistryObjectList().getIdentifiable());
		Assert.assertTrue(queryResult.getRegistryObjectList().getIdentifiable().isEmpty());
	}

	/**
	 * Verifies that a query: - using other query parameters than those designating SXA-data results in no SXA document metadata entry.
	 */
	@Test
	public void noSxaMetadataWhenQueryForTypeCodeNotUsedForSxa() throws Exception {
		// Arrange
		FindDocumentsQuery query = queryHelper.prepareQueryForNonSxaData();
		// Act
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		// Assert
		Assert.assertNotNull(queryResult.getRegistryObjectList());
		Assert.assertTrue(queryResult.getRegistryObjectList().getIdentifiable().isEmpty());
	}

	/**
	 * Retrieve documents from a patient with a subset of possible PracticeSettingCodes. Expects to retrieve only documents that match selected PracticeSettingCodes
	 * 
	 * @throws Exception
	 */

	// -------------------------- Retrieve Documents Tests -------------------------------

	@Test
	public void retrieveDocumentsWithSubsetOfPracticeSettingCodes() throws Exception {
		// Arrange
		List<SXAMappedValue> selectedValidPracticeSettingCodes = new ArrayList<>();
		selectedValidPracticeSettingCodes.add(practiceSettingCodesDataHelper.practiceSettingCodeForMicrobiology);
		selectedValidPracticeSettingCodes.add(practiceSettingCodesDataHelper.practiceSettingCodeForBiochemistry);
		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId(), selectedValidPracticeSettingCodes);
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		List<RetrieveDocument> fetchList = docConsumer.prepareRetrieve(queryResult);
		// Act
		List<RetrievedDocument> retrievedDocuments = docConsumer.doRetrieve(fetchList);
		// Assert
		logger.debug(testutil.prettyPrint(retrievedDocuments));
		Assert.assertNotNull(queryResult.getRegistryObjectList());
		Assert.assertTrue(SxaDocumentHelper.ensureRetrievedDocumentsMatchSpecialities(selectedValidPracticeSettingCodes, retrievedDocuments));
	}

	/**
	 * Retrieve documents from a patient with all possible PracticeSettingCodes. Expects to retrieve all documents from patient
	 * 
	 * @throws Exception
	 */
	@Test
	public void retrieveDocumentsWithAllPracticeSettingCodes() throws Exception {
		// Arrange
		List<SXAMappedValue> allPracticeSettingCodes = new ArrayList<>();
		allPracticeSettingCodes.add(practiceSettingCodesDataHelper.practiceSettingCodeForMicrobiology);
		allPracticeSettingCodes.add(practiceSettingCodesDataHelper.practiceSettingCodeForBiochemistry);
		allPracticeSettingCodes.add(practiceSettingCodesDataHelper.practiceSettingCodeForPathology);

		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId(), allPracticeSettingCodes);
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		List<RetrieveDocument> fetchList = docConsumer.prepareRetrieve(queryResult);
		// Act
		List<RetrievedDocument> retrievedDocuments = docConsumer.doRetrieve(fetchList);
		// Assert
		// logger.info(testutil.prettyPrint(retrievedDocuments));
		Assert.assertNotNull(queryResult.getRegistryObjectList());

		boolean ensureRetrievedDocumentsMatchSpecialities = SxaDocumentHelper.ensureRetrievedDocumentsMatchSpecialities(allPracticeSettingCodes, retrievedDocuments);
		Assert.assertTrue("Forventede match!", ensureRetrievedDocumentsMatchSpecialities);
	}

	/**
	 * Retrieve documents from a patient with single valid PracticeSettingCode. Expects to retrieve documents for patient
	 * 
	 * @throws Exception
	 */
	@Test
	public void retrieveDocumentsWithSinglePracticeSettingCode() throws Exception {
		// Arrange
		List<SXAMappedValue> allPracticeSettingCodes = new ArrayList<>();
		allPracticeSettingCodes.add(practiceSettingCodesDataHelper.practiceSettingCodeForMicrobiology);

		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId(), allPracticeSettingCodes);
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		List<RetrieveDocument> fetchList = docConsumer.prepareRetrieve(queryResult);
		// Act
		List<RetrievedDocument> retrievedDocuments = docConsumer.doRetrieve(fetchList);
		// Assert
		logger.debug(testutil.prettyPrint(retrievedDocuments));
		Assert.assertNotNull(retrievedDocuments);
		Assert.assertTrue(SxaDocumentHelper.ensureRetrievedDocumentsMatchSpecialities(allPracticeSettingCodes, retrievedDocuments));
	}

	/**
	 * Retrieve documents from a patient with no PracticeSettingCodes. Expects to retrieve all documents from patient
	 * 
	 * @throws Exception
	 */
	@Test
	public void retrieveDocumentsWithNoPracticeSettingCodes() throws Exception {

		// Arrange
		List<SXAMappedValue> emptyList = new ArrayList<>();

		List<SXAMappedValue> fullList = new ArrayList<>();
		fullList.add(practiceSettingCodesDataHelper.practiceSettingCodeForMicrobiology);
		fullList.add(practiceSettingCodesDataHelper.practiceSettingCodeForBiochemistry);
		fullList.add(practiceSettingCodesDataHelper.practiceSettingCodeForPathology);

		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId(), emptyList);
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		List<RetrieveDocument> fetchList = docConsumer.prepareRetrieve(queryResult);

		// Act
		List<RetrievedDocument> retrievedDocumentsResult = docConsumer.doRetrieve(fetchList);

		// Assert
		Assert.assertTrue(retrievedDocumentsResult.size() > 0);
		Assert.assertTrue(SxaDocumentHelper.ensureRetrievedDocumentsMatchSpecialities(fullList, retrievedDocumentsResult));
	}

	/**
	 * Retrieve documents from a patient with both valid sxa practice setting codes and invalid sxa practice setting codes. Expects to retrieve only documents that match the valid practice setting
	 * codes
	 * 
	 * @throws Exception
	 */
	@Test
	public void retrieveDocumentsWithMixedValidAndInvalidPracticeSettingCodes() throws Exception {

		// Arrange
		List<SXAMappedValue> practiceSettingCodesWithInvalid = new ArrayList<>();
		practiceSettingCodesWithInvalid.add(practiceSettingCodesDataHelper.practiceSettingCodeForBiochemistry);
		practiceSettingCodesWithInvalid.add(practiceSettingCodesDataHelper.practiceSettingCodeNonSxa);

		List<SXAMappedValue> validList = new ArrayList<>();
		validList.add(practiceSettingCodesDataHelper.practiceSettingCodeForBiochemistry);

		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId(), practiceSettingCodesWithInvalid);
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		List<RetrieveDocument> fetchList = docConsumer.prepareRetrieve(queryResult);

		// Act
		List<RetrievedDocument> retrievedDocumentsResult = docConsumer.doRetrieve(fetchList);

		// Assert
		Assert.assertTrue(retrievedDocumentsResult.size() > 0);
		Assert.assertTrue(SxaDocumentHelper.ensureRetrievedDocumentsMatchSpecialities(validList, retrievedDocumentsResult));
	}

	/**
	 * Retrieve documents from a patient with only invalid PracticeSettingCodes. Expects to retrieve response with no documents
	 * 
	 * @throws Exception
	 */
	@Test
	public void retrieveDocumentsWithOnlyInvalidPracticeSettingCodes() throws Exception {

		// Arrange
		List<SXAMappedValue> invalidList = new ArrayList<>();
		invalidList.add(practiceSettingCodesDataHelper.practiceSettingCodeNonSxa);

		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId(), invalidList);
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);

		// Assert
		Assert.assertNotNull(queryResult.getRegistryObjectList());
		Assert.assertEquals(0, queryResult.getRegistryObjectList().getIdentifiable().size());
	}

	/**
	 * Retrieve documents from a patient with service start and stop time set. Expects to retrieve response
	 * 
	 * @throws Exception
	 */
	@Test
	public void retrieveDocumentsWithStartAndStopTime() throws Exception {

		// Arrange
		FindDocumentsQuery query = queryHelper.prepareQuery(testDataHelper.getDefaultPatientId(), DateTime.now(), DateTime.now());
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		List<RetrieveDocument> fetchList = docConsumer.prepareRetrieve(queryResult);

		// Act
		List<RetrievedDocument> retrievedDocumentsResult = docConsumer.doRetrieve(fetchList);

		// Assert
		Assert.assertTrue(retrievedDocumentsResult.size() > 0);
	}

	/**
	 * Retrieve documents from a patient with service start time explicitly set and stop time set to default value. Expects to retrieve response
	 * 
	 * @throws Exception
	 */
	@Test
	public void retrieveDocumentsWithStartTime() throws Exception {

		// Arrange
		FindDocumentsQuery query = queryHelper.prepareQuery(testDataHelper.getDefaultPatientId(), DateTime.now(), null);
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		List<RetrieveDocument> fetchList = docConsumer.prepareRetrieve(queryResult);

		// Act
		List<RetrievedDocument> retrievedDocumentsResult = docConsumer.doRetrieve(fetchList);

		// Assert
		Assert.assertTrue(retrievedDocumentsResult.size() > 0);
	}

	/**
	 * Retrieve documents from a patient with service stop time explicitly set and start time set to default value. Expects to retrieve response
	 * 
	 * @throws Exception
	 */
	@Test
	public void retrieveDocumentsWithStopTime() throws Exception {

		// Arrange
		FindDocumentsQuery query = queryHelper.prepareQuery(testDataHelper.getDefaultPatientId(), null, DateTime.now());
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		List<RetrieveDocument> fetchList = docConsumer.prepareRetrieve(queryResult);

		// Act
		List<RetrievedDocument> retrievedDocumentsResult = docConsumer.doRetrieve(fetchList);

		// Assert
		Assert.assertTrue(retrievedDocumentsResult.size() > 0);
	}

	/**
	 * Retrieve documents from a patient with service start and stop time set to. Expects to retrieve response
	 * 
	 * @throws Exception
	 */
	@Test
	public void retrieveDocumentsWithoutStartAndStopTime() throws Exception {

		// Arrange
		FindDocumentsQuery query = queryHelper.prepareQuery(testDataHelper.getDefaultPatientId(), null, null);
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		List<RetrieveDocument> fetchList = docConsumer.prepareRetrieve(queryResult);

		// Act
		List<RetrievedDocument> retrievedDocumentsResult = docConsumer.doRetrieve(fetchList);

		// Assert
		Assert.assertTrue(retrievedDocumentsResult.size() > 0);
	}
}
