/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.integrationtest;

import java.util.List;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.joda.time.DateTime;
import org.joda.time.Interval;
import org.junit.Assert;
import org.junit.BeforeClass;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.TimeRange;
import org.openehealth.ipf.commons.ihe.xds.core.requests.RetrieveDocument;
import org.openehealth.ipf.commons.ihe.xds.core.requests.query.FindDocumentsQuery;
import org.openehealth.ipf.commons.ihe.xds.core.responses.RetrievedDocument;
import org.openehealth.ipf.commons.ihe.xds.core.responses.Status;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.query.AdhocQueryResponse;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.SpringRunner;
import org.springframework.test.context.support.AnnotationConfigContextLoader;

import dk.nsi.dds.projects.sxa.common.configuration.IntegrationTestConfiguration;
import dk.nsi.dds.projects.sxa.common.testutil.DocumentSharingTestUtil;
import dk.nsi.dds.projects.sxa.common.xds.impl.ValueConverter;
import dk.nsi.dds.projects.sxa.consumer.DocumentConsumer;
import dk.nsi.dds.projects.sxa.testhelper.SxaLoggerHelper;
import dk.nsi.dds.projects.sxa.testhelper.SxaQueryHelper;

/**
 * These tests assume that SXA is the only document metadata provider responding to queries. Presence of other document metadata providers or document metadata registered by such in a shared registry
 * are likely to interfere with these tests.
 */
@RunWith(SpringRunner.class)
@ContextConfiguration(classes = { IntegrationTestConfiguration.class }, loader = AnnotationConfigContextLoader.class)
public class SxaExclusiveIT {
	private static final Logger logger = Logger.getLogger(SxaExclusiveIT.class);

	@Autowired
	private DocumentConsumer docConsumer;
	@Autowired
	private SxaQueryHelper queryHelper;

	private DocumentSharingTestUtil testutil = new DocumentSharingTestUtil();
	private static DateTime start = ValueConverter.convertFromDTM("20140127" + "104056");
	private static DateTime stop = start.plusYears(1);

	@BeforeClass
	public static void setupClass() throws Exception {
		SxaLoggerHelper.setupLogger(Level.DEBUG); // Set to Level.DEBUG to print out
													// queries, returned metadata etc.
	}

	/**
	 * Verifies that a query: - using query parameters designating SXA-data - without service start and stop query parameters specified results in a single SXA document metadata entry.
	 */
	@Test
	public void sxaMetadataReturnedWithoutStartAndStop() throws Exception {
		// Arrange
		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData();
		logger.debug(testutil.prettyPrint(query));
		// Act
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		logger.debug(testutil.prettyPrint(queryResult.getRegistryObjectList().getIdentifiable()));
		// Assert
		Assert.assertNotNull(queryResult.getRegistryObjectList());
		Assert.assertTrue(queryResult.getRegistryObjectList().getIdentifiable().size() > 0);
		Assert.assertEquals(1, queryResult.getRegistryObjectList().getIdentifiable().size());
	}

	/**
	 * Verifies that a query: - using query parameters designating SXA-data - with service start and stop query parameters specified results in a single SXA document metadata entry where: -
	 * serviceStartTime is set to start time of the query - serviceStopTime is set to stop time of the query
	 */
	@Test
	public void sxaMetadataReturnedWithServiceStartAndStop() throws Exception {
		// Arrange
		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData();
		query.getServiceStartTime().setFrom(start);
		query.getServiceStopTime().setTo(stop);

		logger.debug(testutil.prettyPrint(query));
		// Act
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		logger.debug(testutil.prettyPrint(queryResult.getRegistryObjectList().getIdentifiable()));
		// Assert
		Assert.assertNotNull(queryResult.getRegistryObjectList());
		Assert.assertTrue(queryResult.getRegistryObjectList().getIdentifiable().size() > 0);
	}

	/**
	 * Verifies that a query: - using query parameters designating SXA-data - with service start interval query parameters specified results in a single SXA document metadata entry where: -
	 * serviceStartTime is set to the beginning of the start interval of the query - serviceStopTime is set to null
	 */
	@Test
	public void sxaMetadataReturnedWithServiceStartInterval() throws Exception {
		// Arrange
		DateTime startBegin = start;
		DateTime startEnd = stop;
		Interval interval = new Interval(startBegin, startEnd);
		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData();
		// query = query.setServiceStartTime(interval);
		TimeRange serviceStartTime = query.getServiceStartTime();
		serviceStartTime.setFrom(interval.getStart());
		serviceStartTime.setTo(interval.getEnd());

		logger.debug(testutil.prettyPrint(query));
		// Act
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		logger.debug(testutil.prettyPrint(queryResult.getRegistryObjectList().getIdentifiable()));
		// Assert
		Assert.assertNotNull(queryResult.getRegistryObjectList());
		Assert.assertEquals(1, queryResult.getRegistryObjectList().getIdentifiable().size());
	}

	/**
	 * Verifies that a query: - using query parameters designating SXA-data - with service stop query parameter specified results in a single SXA document metadata entry where: - serviceStartTime is
	 * set to null - serviceStopTime is set to the stop value of the query
	 */
	@Test
	public void sxaMetadataReturnedWithServiceStop() throws Exception {
		// Arrange
		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData();

		TimeRange serviceStopTime = query.getServiceStopTime();
		serviceStopTime.setTo(stop);

		logger.debug(testutil.prettyPrint(query));
		// Act
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		logger.debug(testutil.prettyPrint(queryResult.getRegistryObjectList().getIdentifiable()));
		// Assert
		Assert.assertNotNull(queryResult.getRegistryObjectList());
	}

	/**
	 * Verifies that a query: - using query parameters designating SXA-data - with service start and stop query parameters specified and subsequent document retrieval results in a single SXA document
	 */
	@Test
	public void sxaReturnedWithStartAndStop() throws Exception {
		// Arrange
		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData();
		query.getServiceStartTime().setFrom(start);
		query.getServiceStopTime().setTo(stop);

		logger.debug(testutil.prettyPrint(query));
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		logger.debug(testutil.prettyPrint(queryResult.getRegistryObjectList().getIdentifiable()));
		List<RetrieveDocument> fetchList = docConsumer.prepareRetrieve(queryResult);
		// Act
		List<RetrievedDocument> retrievedDocuments = docConsumer.doRetrieve(fetchList);
		logger.debug(testutil.prettyPrint(retrievedDocuments));
		// Assert
		Assert.assertNotNull(queryResult.getRegistryObjectList());
		Assert.assertEquals(1, retrievedDocuments.size());
	}

	/**
	 * Verifies that a query: - using query parameters designating SXA-data - with service start query parameter specified and subsequent document retrieval results in a single SXA document
	 */
	@Test
	public void sxaDocumentReturnedWithStart() throws Exception {
		// Arrange
		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData();
		query.getServiceStartTime().setFrom(start);

		logger.debug(testutil.prettyPrint(query));
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		logger.debug(testutil.prettyPrint(queryResult.getRegistryObjectList().getIdentifiable()));
		List<RetrieveDocument> fetchList = docConsumer.prepareRetrieve(queryResult);
		// Act
		List<RetrievedDocument> retrievedDocuments = docConsumer.doRetrieve(fetchList);
		logger.debug(testutil.prettyPrint(retrievedDocuments));
		// Assert
		Assert.assertNotNull(queryResult.getRegistryObjectList());
		Assert.assertTrue(queryResult.getRegistryObjectList().getIdentifiable().size() > 0);
		Assert.assertEquals(1, retrievedDocuments.size());
	}

	/**
	 * Verifies that a query: - using query parameters designating SXA-data - with service stop query parameter specified and subsequent document retrieval results in a single SXA document
	 */
	@Test
	public void sxaDocumentReturnedWithStop() throws Exception {
		// Arrange
		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData();
		query.getServiceStopTime().setTo(stop);
		logger.debug(testutil.prettyPrint(query));
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		Assert.assertEquals(Status.FAILURE.getOpcode30(), queryResult.getStatus());

		logger.debug(testutil.prettyPrint(queryResult.getRegistryObjectList().getIdentifiable()));
		List<RetrieveDocument> fetchList = docConsumer.prepareRetrieve(queryResult);
		// Act
		List<RetrievedDocument> retrievedDocuments = docConsumer.doRetrieve(fetchList);
		logger.debug(testutil.prettyPrint(retrievedDocuments));
		// Assert
		Assert.assertNotNull(queryResult.getRegistryObjectList());
		Assert.assertTrue(queryResult.getRegistryObjectList().getIdentifiable().size() == 0);
		Assert.assertEquals(0, retrievedDocuments.size());
	}

	/**
	 * Verifies that a query: - using query parameters designating SXA-data - without start or stop query parameter specified and subsequent document retrieval results in a single SXA document
	 */
	@Test
	public void sxaDocumentReturnedWithoutStartAndStop() throws Exception {
		// Arrange
		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData();
		logger.debug(testutil.prettyPrint(query));
		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		logger.debug(testutil.prettyPrint(queryResult.getRegistryObjectList().getIdentifiable()));
		List<RetrieveDocument> fetchList = docConsumer.prepareRetrieve(queryResult);
		// Act
		List<RetrievedDocument> retrievedDocuments = docConsumer.doRetrieve(fetchList);
		logger.debug(testutil.prettyPrint(retrievedDocuments));
		// Assert
		Assert.assertNotNull(retrievedDocuments);
		Assert.assertTrue(retrievedDocuments.size() >0);
		Assert.assertEquals(1, retrievedDocuments.size());
	}

	/**
	 * Verifies that a query: - using query parameters designating SXA-data - with (past) start and future stop (one year from now) and subsequent document retrieval results in a single SXA document
	 */
	@Test
	public void sxaDocumentReturnedForFutureStop() throws Exception {
		// Arrange
		FindDocumentsQuery query = queryHelper.prepareQueryForSxaData();
		query.getServiceStartTime().setFrom(DateTime.now().minusYears(1));
		query.getServiceStopTime().setFrom(DateTime.now().plusYears(1));

		AdhocQueryResponse queryResult = docConsumer.doQuery(query);
		List<RetrieveDocument> fetchList = docConsumer.prepareRetrieve(queryResult);
		// Act
		List<RetrievedDocument> retrievedDocuments = docConsumer.doRetrieve(fetchList);
		// Assert
		Assert.assertNotNull(retrievedDocuments);
		Assert.assertTrue(retrievedDocuments.size() > 0);
	}
}
