/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.testhelper;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringReader;
import java.io.UnsupportedEncodingException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.List;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.stream.StreamSource;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.openehealth.ipf.commons.ihe.xds.core.responses.RetrievedDocument;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import dk.nsi.dds.projects.sxa.common.configuration.SXAMappedValue;
import dk.nsi.dds.projects.sxa.common.util.SxaDataHandlerUtil;
import dk.nsi.schema.privacy.piw._1.PrivacyInformationWrapperType;

public class SxaDocumentHelper {

	public static final String mappedXstring = "/Rekvisition/LaboratorieSpeciale";
	public static final String rawXstring = "/*[local-name()='LaboratoryReport']/*[local-name()='Letter']/*[local-name()='TypeCode']";

	public static boolean ensureRetrievedDocumentsMatchSpecialities(List<SXAMappedValue> inList, List<RetrievedDocument> retrievedDocumentsResult)
			throws XPathExpressionException, ParserConfigurationException, SAXException, IOException, JAXBException {

		byte[] document = SxaDataHandlerUtil.readDataFrom(retrievedDocumentsResult.get(0));
		List<String> mappedDocuments = getDocumentsArrayFromRetrievedDocument(document, "/ResultatPakkePIH/MappedeSvar/DocumentWrapper");

		boolean mappedDocumentsMatch = ensureDocumentsMatchSpecialities(inList, mappedDocuments, mappedXstring, false)
				&& ensureAllSpecialitiesIsInListOfDocuments(inList, mappedDocuments, mappedXstring, false);

		List<String> rawDocuments = getDocumentsArrayFromRetrievedDocument(document, "/ResultatPakkePIH/RaaSvar/LeverandoerSvar/LeverandoerSvarXml/DocumentWrapper");
		boolean rawDocumentsMatch = ensureDocumentsMatchSpecialities(inList, rawDocuments, rawXstring, true) && ensureAllSpecialitiesIsInListOfDocuments(inList, rawDocuments, rawXstring, true);

		return mappedDocumentsMatch && rawDocumentsMatch;
	}

	private static boolean ensureDocumentsMatchSpecialities(List<SXAMappedValue> inList, List<String> documents, String xString, boolean xrptType)
			throws UnsupportedEncodingException, XPathExpressionException {
		for (String document : documents) {
			if (!documentContainsValidSpeciality(document, inList, xString, xrptType)) {
				return false;
			}
		}
		return true;
	}

	private static boolean ensureAllSpecialitiesIsInListOfDocuments(List<SXAMappedValue> inList, List<String> documents, String xString, boolean xrptType) {
		for (SXAMappedValue speciality : inList) {
			if (!specialityIsInDocuments(speciality, documents, xString, xrptType)) {
				return false;
			}
		}
		return true;
	}

	private static boolean specialityIsInDocuments(SXAMappedValue speciality, List<String> documents, String xString, boolean xrptType) {

		for (String doc : documents) {
			String specialityInDocument;
			try {
				specialityInDocument = findSpeciality(doc, xString);

				if (xrptType) {
					if (speciality.getMappedValue().equals(convertXrptToMappedValue(specialityInDocument))) {
						return true;
					}
				} else {
					if (speciality.getMappedValue().equals(specialityInDocument)) {
						return true;
					}
				}

			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		return false;
	}

	private static boolean documentContainsValidSpeciality(String document, List<SXAMappedValue> specialities, String xString, boolean xrptType)
			throws UnsupportedEncodingException, XPathExpressionException {
		String specialityInDocument = findSpeciality(document, xString);
		for (SXAMappedValue codedValue : specialities) {
			if (xrptType) {
				if (codedValue.getMappedValue().equals(convertXrptToMappedValue(specialityInDocument))) {
					return true;
				}
			} else {
				if (codedValue.getMappedValue().equals(specialityInDocument)) {
					return true;
				}
			}
		}

		return false;
	}

	private static List<String> getDocumentsArrayFromRetrievedDocument(byte[] document, String xString) throws ParserConfigurationException, SAXException, IOException, XPathExpressionException, JAXBException {
		List<String> resultArray = new ArrayList<String>();

		String xml = new String(document, "utf-8");		
		String trimmedXml = xml.trim().replaceFirst("^([\\W]+)<", "<");
		
		PrivacyInformationWrapperType piw = unmarshalPIW(new StringReader(trimmedXml));;

		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder = factory.newDocumentBuilder();
		Document doc = builder.parse(new InputSource(new ByteArrayInputStream(piw.getDocument().getValue())));
		
		NodeList nodeList = (NodeList) XPathFactory.newInstance().newXPath().compile(xString).evaluate(doc, XPathConstants.NODESET);

		for (int i = 0; i < nodeList.getLength(); i++) {
			Node node = nodeList.item(i);
			String nodeValue = node.getFirstChild().getNodeValue();
			String nodeXml = new String(java.util.Base64.getDecoder().decode(nodeValue));
			resultArray.add(nodeXml);
		}

		return resultArray;
	}

	private static PrivacyInformationWrapperType unmarshalPIW(InputStream is) throws JAXBException {
		StreamSource source = new StreamSource(new InputStreamReader(is, StandardCharsets.UTF_8));
		JAXBElement<PrivacyInformationWrapperType> piwElement = JAXBContext.newInstance(PrivacyInformationWrapperType.class).createUnmarshaller().unmarshal(source, PrivacyInformationWrapperType.class);
		return piwElement.getValue();
	}
	
	private static PrivacyInformationWrapperType unmarshalPIW(Reader r) throws JAXBException {
		StreamSource source = new StreamSource(r);
		JAXBElement<PrivacyInformationWrapperType> piwElement = JAXBContext.newInstance(PrivacyInformationWrapperType.class).createUnmarshaller().unmarshal(source, PrivacyInformationWrapperType.class);
		return piwElement.getValue();
	}

	private static String findSpeciality(String doc, String xString) throws UnsupportedEncodingException, XPathExpressionException {
		//byte[] raw = Base64.decode(doc.getBytes());

		String xml = doc;//new String(raw, "utf-8");
		String trimmedXml = xml.trim().replaceFirst("^([\\W]+)<", "<");
		XPath xPath = XPathFactory.newInstance().newXPath();
		String result = xPath.evaluate(xString, new InputSource(new StringReader(trimmedXml)));

		return result;
	}

	private static String convertXrptToMappedValue(String xprt) {
		switch (xprt) {
		case "XRPT01":
			return "KliniskBiokemi";
		case "XRPT02":
			return "Mikrobiologi";
		case "XRPT04":
			return "Patologi";
		default:
			return "";
		}
	}
}
