/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.testhelper;

import java.util.List;

import org.joda.time.DateTime;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.DocumentEntryType;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Identifiable;
import org.openehealth.ipf.commons.ihe.xds.core.requests.query.FindDocumentsQuery;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import dk.nsi.dds.projects.sxa.common.configuration.SXAMappedValue;
import dk.nsi.dds.projects.sxa.common.util.QueryUtil;
import dk.nsi.dds.projects.sxa.consumer.DocumentConsumer;

@Component
public class SxaQueryHelper {

	public static final String PATIENT_CPR = "2512484916";
	public static final String UNKNOWN_PATIENT_CPR = "0101010101";

	private static SXAMappedValue sxaTypeCode = new SXAMappedValue("11502-2", "2.16.840.1.113883.6.1", "LABORATORY REPORT.TOTAL");
	private static SXAMappedValue nonsxaTypeCode = new SXAMappedValue("53576-5", "2.16.840.1.113883.6.1", "Personal Health Monitoring Report");
	private static SXAMappedValue sxaFormatCode = new SXAMappedValue("urn:ad:dk:medcom:labreports:svareksponeringsservice", "1.2.208.184.100.10", "Laboratoriesvar (samling af svar)");

	private DocumentConsumer docConsumer;
	private SxaTestDataHelper testDataHelper;

	@Autowired 
	public SxaQueryHelper(DocumentConsumer docConsumer, SxaTestDataHelper testDataHelper) {
		super();
		this.docConsumer = docConsumer;
		this.testDataHelper = testDataHelper;
	}

	public FindDocumentsQuery prepareQueryForSxaData() {
		return prepareQueryForSxaData(testDataHelper.getDefaultPatientId());
	}

	public FindDocumentsQuery prepareQueryForSxaData(Identifiable patientId) {
		return prepareQuery(patientId, sxaTypeCode, sxaFormatCode);
	}

	public FindDocumentsQuery prepareQueryForNonSxaData() {
		return prepareQuery(testDataHelper.getDefaultPatientId(), nonsxaTypeCode, sxaFormatCode);
	}

	public FindDocumentsQuery prepareQueryForSxaData(Identifiable patientId, List<SXAMappedValue> practiceSettingCodes) {
		FindDocumentsQuery query = prepareQuery(patientId, sxaTypeCode, sxaFormatCode);
		for (SXAMappedValue mappedCode : practiceSettingCodes) {
			QueryUtil.addPracticeSettingCode(query, mappedCode);
		}
		return query;
	}

	public FindDocumentsQuery prepareQueryForSxaDataWithOnlyUnsupportedPracticeSettingCodes() {
		FindDocumentsQuery query = prepareQueryForSxaData(testDataHelper.getDefaultPatientId());
		QueryUtil.addPracticeSettingCode(query, testDataHelper.practiceSettingCodeInvalid);
		return query;
	}

	private FindDocumentsQuery prepareQuery(Identifiable patientId, SXAMappedValue typeCode, SXAMappedValue formatCode) {
		return prepareQuery(patientId, typeCode, formatCode, DateTime.now(), DateTime.now());
	}

	private FindDocumentsQuery prepareQuery(Identifiable patientId, SXAMappedValue typeCode, SXAMappedValue formatCode, DateTime startTime, DateTime stopTime) {
		FindDocumentsQuery query = docConsumer.createQuery(patientId);
		QueryUtil.addDocumentEntryType(query, DocumentEntryType.ON_DEMAND);
		QueryUtil.addTypeCode(query, typeCode);
		QueryUtil.addFormatCode(query, formatCode);

		if (startTime != null) {
			query.getServiceStartTime().setFrom(startTime);
		}
		if (stopTime != null) {
			query.getServiceStopTime().setTo(stopTime);
		}
		return query;
	}

	public FindDocumentsQuery prepareQuery(Identifiable patientId, DateTime startTime, DateTime stopTime) {
		return prepareQuery(patientId, sxaTypeCode, sxaFormatCode, startTime, stopTime);
	}

}
