/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2012 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.labreportservicestub.service;

import java.util.List;
import java.util.Map;

import javax.annotation.Resource;
import javax.inject.Inject;
import javax.jws.WebMethod;
import javax.jws.WebParam;
import javax.jws.WebResult;
import javax.jws.WebService;
import javax.jws.soap.SOAPBinding;
import javax.xml.bind.annotation.XmlSeeAlso;
import javax.xml.soap.MessageFactory;
import javax.xml.soap.Name;
import javax.xml.soap.SOAPConstants;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPFault;
import javax.xml.soap.SOAPMessage;
import javax.xml.ws.Holder;
import javax.xml.ws.WebServiceContext;
import javax.xml.ws.handler.MessageContext;
import javax.xml.ws.soap.SOAPFaultException;

import org.apache.log4j.Logger;

import com.cgi.svareksponering.ArrayOfString;
import com.cgi.svareksponering.GetDataFromPeriodParameters;
import com.cgi.svareksponering.ISvarEksponering;
import com.cgi.svareksponering.ResultatPakkePIH;

import dk.nsi.dds.projects.sxa.labreportservicestub.configuration.HttpHeaderControl;
import dk.nsi.dds.projects.sxa.labreportservicestub.configuration.SxaConfiguration;
import dk.nsi.dds.projects.sxa.labreportservicestub.configuration.logic.SxaLogic;
import dk.nsi.dds.projects.sxa.labreportservicestub.exception.SxaException;
import dk.nsi.schema.privacy.pih._1.PrivacyInformationHeaderType;

@WebService(serviceName = "SvarEksponering", targetNamespace = "http://cgi.com/SvarEksponering", portName = "BasicHttpBinding_ISvarEksponering")
@SOAPBinding(parameterStyle = SOAPBinding.ParameterStyle.BARE)
@XmlSeeAlso({ dk.nsi.schema.privacy.pih._1.ObjectFactory.class, com.cgi.svareksponering.ObjectFactory.class })
public class SvarEksponering implements ISvarEksponering {
  private static final Logger logger = Logger.getLogger(SvarEksponering.class);
  
  @Inject String startupTime;
  @Inject SxaLogic sxaLogic; 
  @Inject SxaConfiguration sxaConfiguration;
  @Inject HttpHeaderControl httpHeaderControl;
  
  @Resource
  WebServiceContext wsCtxt;

  @WebMethod(operationName = "GetDataFromPeriod", action = "http://cgi.com/SvarEksponering/ISvarEksponering/GetDataFromPeriod")
  @WebResult(name = "ResultatPakkePIH", targetNamespace = "http://cgi.com/SvarEksponering", partName = "parameters")
  @SOAPBinding(parameterStyle = SOAPBinding.ParameterStyle.BARE)
  public ResultatPakkePIH getDataFromPeriod(
    @WebParam(name = "GetDataFromPeriodParameters", targetNamespace = "http://cgi.com/SvarEksponering", partName = "parameters") GetDataFromPeriodParameters parameters,
    @WebParam(name = "Header", targetNamespace = "http://cgi.com/SvarEksponering", header = true, mode = WebParam.Mode.OUT, partName = "Header") Holder<PrivacyInformationHeaderType> header) {
    try {
    	Map<String, List<String>> httpHeaders = getHttpHeaders();
    	if(httpHeaderControl.doPrint(httpHeaders)) {
    		httpHeaderControl.print(httpHeaders);
    	}
    	
    	final boolean doValidateHttpHeaders = httpHeaderControl.doValidate(httpHeaders);
		logger.info("Validating HTTP_REQUEST_HEADERS: "+ doValidateHttpHeaders);
    	if(doValidateHttpHeaders) {
    		httpHeaderControl.validate(httpHeaders);
    		logger.info("HTTP_REQUEST_HEADERS are valid");
    	}
      return sxaLogic
          .authenticateUser(parameters.getLoginUsername(), parameters.getLoginPassword())
          .verifyPatientCpr(parameters.getPatientCpr())
          .doLogic(parameters, header);
    } catch (SxaException e) {
      logger.debug(e.getMessage());
      throw new SOAPFaultException(generateSOAPFault(true, e.getMessage(), "DataError"));
    } catch (Exception e) {
      logger.error(e.getMessage(), e);
      throw new SOAPFaultException(generateSOAPFault(false, e.getMessage(), "SystemError"));
    }
  }
  
  @SuppressWarnings("unchecked")
  private Map<String, List<String>> getHttpHeaders(){
	MessageContext messageContext = wsCtxt.getMessageContext();
	Map<String, List<String>> headers = (Map<String, List<String>>) messageContext.get(MessageContext.HTTP_REQUEST_HEADERS);
	return headers;
  }
  

  @Override
  public ArrayOfString getCustomMessages() {
    return null;
  }

  @Override
  public ArrayOfString dumpXmlFileForJournalSystem(
    ArrayOfString svarTilAfsendelseArray, String lokationsNummer, String yderNummer, String sksKode) {
    return null;
  }
  
  private static SOAPFault generateSOAPFault(boolean client, String message, String faultCode) {
    SOAPFault fault = null;
    try {
        SOAPMessage soapMessage = MessageFactory.newInstance().createMessage();
        final SOAPEnvelope envelope = soapMessage.getSOAPPart().getEnvelope();
        final Name name = envelope.createName(faultCode, null, SOAPConstants.URI_NS_SOAP_1_1_ENVELOPE);
        fault = soapMessage.getSOAPBody().addFault(name, message != null ? message : "unknown message");
    } catch (SOAPException e1) {
        logger.error("Error generating SOAPFault: " + e1.getMessage());
    }
    return fault;
}

}
