/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2014 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.labreportservicestub.xml;

import java.io.IOException;
import java.io.InputStream;
import java.util.List;

import javax.inject.Inject;

import org.apache.log4j.Logger;

import com.cgi.svareksponering.ArrayOfDocumentWrapper;
import com.cgi.svareksponering.ArrayOfLeverandoerSvar;
import com.cgi.svareksponering.DocumentWrapper;
import com.cgi.svareksponering.LeverandoerSvar;
import com.cgi.svareksponering.ResultatPakkePIH;
import com.cgi.svareksponering.Speciale;
import com.thoughtworks.xstream.XStream;
import com.thoughtworks.xstream.converters.extended.ToAttributedValueConverter;

import dk.nsi.dds.projects.sxa.common.util.xmlhandling.XStreamFileException;
import dk.nsi.dds.projects.sxa.common.util.xmlhandling.XStreamInputStreamHandler;
import dk.nsi.dds.projects.sxa.labreportservicestub.configuration.SxaConfiguration;
import dk.nsi.dds.projects.sxa.labreportservicestub.configuration.SxaConstants;
import dk.nsi.dds.projects.sxa.labreportservicestub.configuration.SxaCprToResponseFileConfiguration;
import dk.nsi.dds.projects.sxa.labreportservicestub.exception.SxaException;
import dk.nsi.dds.projects.sxa.labreportservicestub.xml.entity.SxaSoapXmlBody;
import dk.nsi.dds.projects.sxa.labreportservicestub.xml.entity.SxaSoapXmlEnvelope;
import dk.nsi.dds.projects.sxa.labreportservicestub.xml.entity.SxaSoapXmlHeader;
import dk.nsi.schema.privacy.pih._1.AttributeType;
import dk.nsi.schema.privacy.pih._1.DeclarationType;
import dk.nsi.schema.privacy.pih._1.LocationType;
import dk.nsi.schema.privacy.pih._1.NamespaceType;
import dk.nsi.schema.privacy.pih._1.PrivacyInformationHeaderType;
import dk.nsi.schema.privacy.pih._1.SourceType;

public class SxaXmlReader {

	@Inject
	SxaConfiguration sxaConfiguration;

	private static final Logger logger = Logger.getLogger(SxaXmlReader.class);
	private static final XStream xstream = createXstreamForResponse();
	private SxaCprToResponseFileConfiguration cprConfiguration;

	private static XStream createXstreamForResponse() {
		XStream xstream = new XStream();
		xstream.ignoreUnknownElements();
		xstream.registerConverter(new ToAttributedValueConverter(DocumentWrapper.class, xstream.getMapper(), xstream.getReflectionProvider(), xstream.getConverterLookup(), "value"));
		xstream.registerConverter(new ToAttributedValueConverter(NamespaceType.class, xstream.getMapper(), xstream.getReflectionProvider(), xstream.getConverterLookup(), "value"));
		xstream.registerConverter(new ToAttributedValueConverter(AttributeType.class, xstream.getMapper(), xstream.getReflectionProvider(), xstream.getConverterLookup(), "value"));
		xstream.alias("s:Envelope", SxaSoapXmlEnvelope.class);
		xstream.aliasField("s:Header", SxaSoapXmlEnvelope.class, "header");
		xstream.aliasField("h:header", SxaSoapXmlHeader.class, "pih");
		xstream.alias("source", SourceType.class);
		xstream.alias("attribute", AttributeType.class);
		xstream.addImplicitCollection(SourceType.class, "attribute");
		xstream.alias("location", LocationType.class);
		xstream.alias("declaration", DeclarationType.class);
		xstream.addImplicitCollection(PrivacyInformationHeaderType.class, "declaration");
		xstream.alias("element", String.class);
		xstream.addImplicitCollection(LocationType.class, "element");
		xstream.aliasField("s:Body", SxaSoapXmlEnvelope.class, "body");
		xstream.aliasField("ResultatPakkePIH", SxaSoapXmlBody.class, "resultatPakkePIH");
		xstream.aliasField("MappedeSvar", ResultatPakkePIH.class, "mappedeSvar");
		xstream.aliasField("RaaSvar", ResultatPakkePIH.class, "raaSvar");
		xstream.aliasField("LeverandoerSvar", ArrayOfLeverandoerSvar.class, "leverandoerSvar");
		xstream.addImplicitCollection(ArrayOfLeverandoerSvar.class, "leverandoerSvar", LeverandoerSvar.class);
		xstream.aliasField("LeverandoerSvarXml", LeverandoerSvar.class, "leverandoerSvarXml");

		xstream.aliasField("DocumentWrapper", ArrayOfDocumentWrapper.class, "documentWrapper");
		xstream.addImplicitCollection(ArrayOfDocumentWrapper.class, "documentWrapper", DocumentWrapper.class);
		return xstream;
	}

	private SxaCprToResponseFileConfiguration readCprToResponseFile() {
		try {
			String cpr2ResponceFileMap = sxaConfiguration.getProperty(SxaConstants.CPR_TO_RESPONSEFILE_MAP_FILE);

			InputStream cprReader = SxaXmlReader.class.getClassLoader().getResourceAsStream(cpr2ResponceFileMap);
			return new SxaCprToResponseFileConfiguration(cprReader);

		} catch (IOException e) {
			throw new SxaException("Configuration loading failed: " + e.getMessage(), e);
		}
	}

	private void loadCprConfiguration() {
		if (cprConfiguration == null) {
			cprConfiguration = readCprToResponseFile();
		}
	}

	public SxaSoapXmlEnvelope getEnvelopeFromFile(String patientId, List<Speciale> specialities) throws Exception {
		loadCprConfiguration();
		String stubFilename = getStubFilenameForPatientWithSpecialities(patientId, specialities);

		InputStream inputStream = SxaXmlReader.class.getClassLoader().getResourceAsStream(stubFilename);
		XStreamInputStreamHandler xStreamHandler = new XStreamInputStreamHandler(inputStream, xstream);

		SxaSoapXmlEnvelope envelope;

		try {
			envelope = xStreamHandler.readXmlFileFromStream(SxaSoapXmlEnvelope.class);
			return envelope;
		} catch (XStreamFileException e) {
			throw new SxaException("Cant read stub_data_file: " + e.getMessage(), e);
		}
	}

	private String getStubFilenameForPatientWithSpecialities(String patientId, List<Speciale> specialities) throws Exception {
		String patientWithSpeciality = patientId;
		patientWithSpeciality += getSpecialitySuffix(specialities);
		String responeStubFilename = cprConfiguration.getCprToResponseFilePairs().get(patientWithSpeciality);
		logFileLoading(responeStubFilename, patientId, "");
		return responeStubFilename;
	}

	private String getSpecialitySuffix(List<Speciale> specialities) {
		Integer result = 0;
		for (Speciale speciality : specialities) {
			switch (speciality.name()) {
			case "MIKROBIOLOGI":
				result += 1;
				break;
			case "KLINISK_BIOKEMI":
				result += 2;
				break;
			case "PATOLOGI":
				result += 4;
				break;
			default:
				return "_0";
			}
		}

		return "_" + result.toString();
	}

	private void logFileLoading(String responseStubFilename, String patientId, String fileName) {
		if (responseStubFilename == null || responseStubFilename.trim().isEmpty()) {
			throw new SxaException("Patient with cpr: " + patientId + " is not mapped in " + fileName);
		}
		logger.info("Patient with cpr: " + patientId + " is mapped as " + responseStubFilename + " in " + fileName);
	}
}
