package dk.nsi.dds.projects.sxa.labreportservicestub.service;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.ws.Holder;
import javax.xml.ws.WebServiceContext;
import javax.xml.ws.handler.MessageContext;
import javax.xml.ws.soap.SOAPFaultException;

import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.runners.MockitoJUnitRunner;

import com.cgi.svareksponering.GetDataFromPeriodParameters;
import com.cgi.svareksponering.ResultatPakkePIH;

import dk.nsi.dds.projects.sxa.labreportservicestub.configuration.HttpHeaderControl;
import dk.nsi.dds.projects.sxa.labreportservicestub.configuration.logic.SxaLogic;
import dk.nsi.dds.projects.sxa.labreportservicestub.exception.SxaException;
import dk.nsi.schema.privacy.pih._1.PrivacyInformationHeaderType;

@RunWith(MockitoJUnitRunner.class)
public class SvarEksponeringTest {
	
	@Mock
	private SxaLogic sxaLogic;	
	
	@Mock
	private WebServiceContext wsCtx;
	
	@Mock
	private MessageContext mCtx;
	
	@Mock
    private GetDataFromPeriodParameters parameters;
	
	@Mock
    private PrivacyInformationHeaderType pihType;
	
	@Mock
	private ResultatPakkePIH resultatPakkePIH;
	
	@Mock
	private HttpHeaderControl httpHeaderControl;
	
	@InjectMocks
	private SvarEksponering svarEksponering;
	
    private Holder<PrivacyInformationHeaderType> header;
    Map<String, List<String>> httpHeaders;
    
	@Before
	public void setup() throws Exception {
		header = new Holder<PrivacyInformationHeaderType>(pihType);
		
		httpHeaders = new HashMap<>();
		
		Mockito.when(wsCtx.getMessageContext()).thenReturn(mCtx);
		
		Mockito.when(mCtx.get(MessageContext.HTTP_REQUEST_HEADERS)).thenReturn(httpHeaders);
		
		Mockito.when(sxaLogic.authenticateUser(Mockito.anyString(), Mockito.anyString())).thenReturn(sxaLogic);
		Mockito.when(sxaLogic.verifyPatientCpr(Mockito.anyString())).thenReturn(sxaLogic);
		Mockito.when(sxaLogic.doLogic(parameters, header)).thenReturn(resultatPakkePIH);
		
		Mockito.when(httpHeaderControl.doPrint(httpHeaders)).thenReturn(true);
		Mockito.when(httpHeaderControl.doValidate(httpHeaders)).thenReturn(true);
		
	}
	@Test
	public void testGetDataFromPeriodNoHeader() throws Exception {
		
		httpHeaders = null;
		
		Mockito.when(mCtx.get(MessageContext.HTTP_REQUEST_HEADERS)).thenReturn(httpHeaders);
		
		svarEksponering.getDataFromPeriod(parameters, header);
		
		Mockito.verify(wsCtx, Mockito.times(1)).getMessageContext();
		Mockito.verify(mCtx, Mockito.times(1)).get(MessageContext.HTTP_REQUEST_HEADERS);
		
		Mockito.verify(sxaLogic, Mockito.times(1)).authenticateUser(Mockito.anyString(), Mockito.anyString());
		Mockito.verify(sxaLogic, Mockito.times(1)).verifyPatientCpr(Mockito.anyString());
		Mockito.verify(sxaLogic, Mockito.times(1)).doLogic(parameters, header);
		
		Mockito.verify(httpHeaderControl, Mockito.times(1)).doPrint(httpHeaders);
		Mockito.verify(httpHeaderControl, Mockito.times(1)).doValidate(httpHeaders);
		
		Mockito.verifyNoMoreInteractions(wsCtx, mCtx, sxaLogic, httpHeaderControl);
	}
	
	@Test
	public void testGetDataFromPeriodEmptyHeader() throws Exception {
		
		svarEksponering.getDataFromPeriod(parameters, header);
		
		Mockito.verify(wsCtx, Mockito.times(1)).getMessageContext();
		Mockito.verify(mCtx, Mockito.times(1)).get(MessageContext.HTTP_REQUEST_HEADERS);
		
		Mockito.verify(sxaLogic, Mockito.times(1)).authenticateUser(Mockito.anyString(), Mockito.anyString());
		Mockito.verify(sxaLogic, Mockito.times(1)).verifyPatientCpr(Mockito.anyString());
		Mockito.verify(sxaLogic, Mockito.times(1)).doLogic(parameters, header);
		
		Mockito.verify(httpHeaderControl, Mockito.times(1)).doPrint(httpHeaders);
		Mockito.verify(httpHeaderControl, Mockito.times(1)).print(httpHeaders);
		Mockito.verify(httpHeaderControl, Mockito.times(1)).doValidate(httpHeaders);
		Mockito.verify(httpHeaderControl, Mockito.times(1)).validate(httpHeaders);
		
		Mockito.verifyNoMoreInteractions(wsCtx, mCtx, sxaLogic, httpHeaderControl);
	}
	
	@Test
	public void testGetDataFromPeriodValidateHeader() throws Exception {
				
		svarEksponering.getDataFromPeriod(parameters, header);
		
		Mockito.verify(wsCtx, Mockito.times(1)).getMessageContext();
		
		Mockito.verify(mCtx, Mockito.times(1)).get(MessageContext.HTTP_REQUEST_HEADERS);
		
		Mockito.verify(sxaLogic, Mockito.times(1)).authenticateUser(Mockito.anyString(), Mockito.anyString());
		Mockito.verify(sxaLogic, Mockito.times(1)).verifyPatientCpr(Mockito.anyString());
		Mockito.verify(sxaLogic, Mockito.times(1)).doLogic(parameters, header);
		
		Mockito.verify(httpHeaderControl, Mockito.times(1)).doPrint(httpHeaders);
		Mockito.verify(httpHeaderControl, Mockito.times(1)).print(httpHeaders);
		Mockito.verify(httpHeaderControl, Mockito.times(1)).doValidate(httpHeaders);
		Mockito.verify(httpHeaderControl, Mockito.times(1)).validate(httpHeaders);
		
		Mockito.verifyNoMoreInteractions(wsCtx, mCtx, sxaLogic, httpHeaderControl);
	}
	
	@SuppressWarnings("unchecked")
	@Test(expected=SOAPFaultException.class)
	public void testGetDataFromPeriodThrowSxaException() throws Exception {			
		Mockito.when(sxaLogic.authenticateUser(Mockito.isNull(String.class), Mockito.anyString())).thenThrow(SxaException.class);
		
		svarEksponering.getDataFromPeriod(parameters, header);
	}
	
	@Test
	public void testGetCustomMessages() throws Exception {
		Assert.assertNull(svarEksponering.getCustomMessages());
	}
	
	@Test
	public void testDumpXmlFileForJournalSystem() throws Exception {
		Assert.assertNull(svarEksponering.dumpXmlFileForJournalSystem(null, null, null, null));
	}
}
