/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentmetadataprovider;

import java.io.File;
import java.io.FileNotFoundException;
import java.net.URISyntaxException;
import java.util.Collections;
import java.util.List;

import javax.xml.bind.JAXBElement;

import org.joda.time.DateTime;
import org.junit.Assert;
import org.junit.BeforeClass;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.AssigningAuthority;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.AvailabilityStatus;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.DocumentEntry;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.DocumentEntryType;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Identifiable;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Person;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.XcnName;
import org.openehealth.ipf.commons.ihe.xds.core.requests.QueryRegistry;
import org.openehealth.ipf.commons.ihe.xds.core.requests.query.FindDocumentsQuery;
import org.openehealth.ipf.commons.ihe.xds.core.responses.QueryResponse;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.query.AdhocQueryRequest;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.query.AdhocQueryResponse;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.rim.ClassificationType;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.rim.ExternalIdentifierType;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.rim.ExtrinsicObjectType;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.rim.IdentifiableType;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.rim.RegistryObjectListType;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.rim.SlotListType;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.rim.SlotType1;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.rs.RegistryErrorList;
import org.openehealth.ipf.commons.ihe.xds.iti18.Iti18PortType;
import org.openehealth.ipf.platform.camel.ihe.xds.core.converters.EbXML30Converters;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;
import org.springframework.test.context.support.AnnotationConfigContextLoader;

import dk.nsi.dds.projects.sxa.common.TestDataHelper;
import dk.nsi.dds.projects.sxa.common.configuration.IntegrationTestConfiguration;
import dk.nsi.dds.projects.sxa.common.testutil.AssertUtil;
import dk.nsi.dds.projects.sxa.common.util.QueryUtil;
import dk.nsi.dds.projects.sxa.common.util.xmlhandling.XStreamFileException;
import dk.nsi.dds.projects.sxa.documentmetadataprovider.configuration.MetadataConfiguration;
import dk.nsi.dds.projects.sxa.documentmetadataprovider.configuration.MetadataConfigurationHandler;
import dk.nsi.dds.projects.sxa.documentmetadataprovider.configuration.MetadataContentConfiguration;
import dk.nsi.util.Resource;

@RunWith(SpringJUnit4ClassRunner.class)
@ContextConfiguration(classes = { IntegrationTestConfiguration.class }, loader = AnnotationConfigContextLoader.class)
public class MetadataProviderWSIT {
	@Autowired
	private Iti18PortType iti18PortType;

	@BeforeClass
	public static void setupTestClientProperties() {
		// Denne opsætning bruges, hvis man vil afvikle testen af udviklingsvækrtøj og man ikke har kaldt 'mvn verify -P external-test'
		// Ved kald af 'mvn verify -P external-test' så kopieres testclient.properties fra integrationtest-modulet til classpath for denne test.
		if (System.getProperty(IntegrationTestConfiguration.clientPropertiesFileName) == null) {
			System.setProperty(IntegrationTestConfiguration.clientPropertiesFileName, new File("../../../integrationtest/src/test/resources/testclient.properties").getAbsolutePath());
		}
	}

	@Test
	public void testSxaMetadataProviderWS() throws Exception {
		MetadataConfiguration metadataConfiguration = loadConfiguration();

		// Arrange
		FindDocumentsQuery query = createDefaultQuery(metadataConfiguration.getMetadataContentConfiguration());
		query.setStatus(Collections.singletonList(AvailabilityStatus.APPROVED));

		// Act
		QueryResponse result = callWebservice(query);
		// Assert
		Assert.assertNotNull(result.getDocumentEntries());
		Assert.assertFalse(result.getDocumentEntries().isEmpty());
		Assert.assertEquals(1, result.getDocumentEntries().size());

		DocumentEntry metadataResult = result.getDocumentEntries().get(0);

		Assert.assertNotNull(metadataResult);
		Assert.assertEquals(getPatientId(), metadataResult.getPatientId());

		if (query.getClassCodes() != null && !query.getClassCodes().isEmpty()) {
			AssertUtil.assertEqualsIgnoreLangAndCharset(query.getClassCodes().get(0), metadataResult.getClassCode());
		}

		AssertUtil.assertEqualsIgnoreLangAndCharset(query.getTypeCodes().get(0), metadataResult.getTypeCode());
		if (query.getEventCodes() != null && !query.getEventCodes().getOuterList().isEmpty()) {
			AssertUtil.assertEqualsIgnoreLangAndCharset(query.getEventCodes().getOuterList().get(0).get(0), metadataResult.getEventCodeList().get(0));
		}
		DocumentMetadataProviderTestDataHelper testHelper = new DocumentMetadataProviderTestDataHelper(new TestDataHelper());
		testHelper.compareConfigurationWithDocumentMetadata(query, metadataConfiguration, metadataResult);
	}

	@Test
	public void testMinimalQuery() throws Exception {
		// Arrange
		FindDocumentsQuery query = new FindDocumentsQuery();

		query.setPatientId(getPatientId());
		query.setDocumentEntryTypes(Collections.singletonList(DocumentEntryType.ON_DEMAND));

		query.setStatus(Collections.singletonList(AvailabilityStatus.APPROVED));

		// Act
		QueryResponse result = callWebservice(query, new AdhocQueryResponseValidator() {
			@Override
			public void validate(AdhocQueryResponse adhocQueryResponse) {
				AssertUtil.assertEquals("urn:uuid:14d4debf-8f97-4251-9a74-a90016b0af0d", adhocQueryResponse.getRequestId());
				AssertUtil.assertEquals("urn:oasis:names:tc:ebxml-regrep:ResponseStatusType:Success", adhocQueryResponse.getStatus());
				AssertUtil.assertEquals(0, adhocQueryResponse.getStartIndex());
				AssertUtil.assertEquals(1, adhocQueryResponse.getTotalResultCount());

				RegistryErrorList registryErrorList = adhocQueryResponse.getRegistryErrorList();
				AssertUtil.assertNull(registryErrorList);

				RegistryObjectListType registryObjectList = adhocQueryResponse.getRegistryObjectList();
				AssertUtil.assertNotNull(registryObjectList);
				AssertUtil.assertEquals(1, registryObjectList.getIdentifiable().size());
				final JAXBElement<? extends IdentifiableType> jaxbElement = registryObjectList.getIdentifiable().get(0);
				IdentifiableType value = jaxbElement.getValue();
				AssertUtil.assertEquals("urn:uuid:1056d72b39b9c48097e5e", value.getId());
				AssertUtil.assertEquals("1.2.3.5.6", value.getHome());
				List<SlotType1> slots = value.getSlot();
				AssertUtil.assertNotNull(slots);
				AssertUtil.assertEquals(4, slots.size());
				for (SlotType1 slotType : slots) {
					AssertUtil.assertOneOf(slotType.getName(), "languageCode", "sourcePatientId", "repositoryUniqueId", "serviceStartTime");
					AssertUtil.assertNull(slotType.getSlotType());
					AssertUtil.assertNotNull(slotType.getValueList());
					List<String> slotTypeValues = slotType.getValueList().getValue();
					AssertUtil.assertNotNull(slotTypeValues);
					AssertUtil.assertEquals(1, slotTypeValues.size());
					AssertUtil.assertOneOf(slotTypeValues.get(0), "da-DK","2512484916^^^&1.2.3.4&ISO","1.2.3.4.55.66","19481225000000");
				}

				AssertUtil.assertEquals(ExtrinsicObjectType.class, jaxbElement.getDeclaredType());
				ExtrinsicObjectType extrinsicObjectType = (ExtrinsicObjectType) jaxbElement.getValue();
				AssertUtil.assertEquals("Lab Reports for 2512484916", extrinsicObjectType.getName());
				AssertUtil.assertNull(extrinsicObjectType.getDescription());
				AssertUtil.assertEquals(value.getHome(), extrinsicObjectType.getHome());
				AssertUtil.assertEquals(value.getId(), extrinsicObjectType.getId());
				
				AssertUtil.assertNull(extrinsicObjectType.getLid());
				
				AssertUtil.assertEquals("text/xml", extrinsicObjectType.getMimeType());
				AssertUtil.assertEquals("urn:uuid:34268e47-fdf5-41a6-ba33-82133c465248", extrinsicObjectType.getObjectType());
				AssertUtil.assertEquals("urn:oasis:names:tc:ebxml-regrep:StatusType:Approved", extrinsicObjectType.getStatus());
				AssertUtil.assertNotNull(extrinsicObjectType.getClassification());
				AssertUtil.assertEquals(6, extrinsicObjectType.getClassification().size());
				for (ClassificationType classificationType : extrinsicObjectType.getClassification()) {
					AssertUtil.assertNull(classificationType.getClassificationNode());
					AssertUtil.assertOneOf(classificationType.getClassificationScheme(), "urn:uuid:41a5887f-8865-4c09-adf7-e362475b143a", "urn:uuid:a09d5840-386c-46f2-b5ad-9c3699a4309d",
							"urn:uuid:f33fb8ac-18af-42cc-ae0e-ed0b0bdb91e1", "urn:uuid:f0306f51-975f-434e-a61c-c59651d33983", "urn:uuid:f4f85eac-e6cb-4883-b524-f2705394840f",
							"urn:uuid:93606bcf-9494-43ec-9b4e-a7748d1a838d");
					AssertUtil.assertEquals(value.getId(), classificationType.getClassifiedObject());
					AssertUtil.assertNull(classificationType.getHome());
					AssertUtil.assertTrue(classificationType.getId().startsWith("urn:uuid:"));					
					AssertUtil.assertNull(classificationType.getLid());				
					AssertUtil.assertOneOf(classificationType.getNodeRepresentation(), "", "001", "urn:ad:dk:medcom:labreports:svareksponeringsservice", "22232009", "11502-2", "N");				
					AssertUtil.assertNull(classificationType.getObjectType());
					AssertUtil.assertNull(classificationType.getStatus());
					AssertUtil.assertEmpty(classificationType.getClassification());
					AssertUtil.assertNull(classificationType.getDescription());
					AssertUtil.assertEmpty(classificationType.getExternalIdentifier());
					AssertUtil.assertOneOf(classificationType.getName(), true, "Klinisk rapport", "Laboratoriesvar (samling af svar)", "empty", "LABORATORY REPORT.TOTAL", "Normal");
					AssertUtil.assertNotNull(classificationType.getSlot());
					AssertUtil.assertEquals(1, classificationType.getSlot().size());
					SlotType1 slotType = classificationType.getSlot().get(0);
					AssertUtil.assertNotNull(slotType);
					AssertUtil.assertOneOf(slotType.getName(), "codingScheme", "authorInstitution");
					AssertUtil.assertNull(slotType.getSlotType());
					AssertUtil.assertNotNull(slotType.getValueList());
					List<String> slotTypeValues = slotType.getValueList().getValue();
					AssertUtil.assertNotNull(slotTypeValues);
					AssertUtil.assertEquals(1, slotTypeValues.size());
					AssertUtil.assertOneOf(slotTypeValues.get(0), "1.2.208.184.100.9", "1.2.208.184.100.10", "2.16.840.1.113883.6.96", "2.16.840.1.113883.6.1", "2.16.840.1.113883.5.25",
							"Region Test^^^^^&1.2.208.176.1.1&ISO^^^^291000016008");
					AssertUtil.assertNull(classificationType.getVersionInfo());
				}
				AssertUtil.assertNull(extrinsicObjectType.getContentVersionInfo());
				AssertUtil.assertNull(extrinsicObjectType.getDataHandler());
				AssertUtil.assertNotNull(extrinsicObjectType.getExternalIdentifier());
				AssertUtil.assertEquals(2, extrinsicObjectType.getExternalIdentifier().size());
				for (ExternalIdentifierType externalIdentifierType : extrinsicObjectType.getExternalIdentifier()) {
					AssertUtil.assertNull(externalIdentifierType.getHome());
					AssertUtil.assertTrue(externalIdentifierType.getId().startsWith("urn:uuid:"));
					AssertUtil.assertNull(externalIdentifierType.getLid());					
					AssertUtil.assertOneOf(externalIdentifierType.getIdentificationScheme(), "urn:uuid:58a6f841-87b3-4a3e-92fd-a8ffeff98427", "urn:uuid:2e82c1f6-a085-4c72-9da3-8640a32e42ab");
					AssertUtil.assertNull(externalIdentifierType.getObjectType());
					AssertUtil.assertEquals(value.getId(), externalIdentifierType.getRegistryObject());				
					AssertUtil.assertNull(externalIdentifierType.getStatus());
					AssertUtil.assertOneOf(externalIdentifierType.getValue(), "2512484916^^^&1.2.3.4&ISO", "1.2.3.4.55.66.12253848422678.1.1.0");
					AssertUtil.assertEmpty(externalIdentifierType.getClassification());
					AssertUtil.assertNull(externalIdentifierType.getDescription());
					AssertUtil.assertEmpty(externalIdentifierType.getExternalIdentifier());
					AssertUtil.assertOneOf(externalIdentifierType.getName(), "XDSDocumentEntry.patientId", "XDSDocumentEntry.uniqueId");
					AssertUtil.assertEmpty(externalIdentifierType.getSlot());					
					AssertUtil.assertNull(externalIdentifierType.getVersionInfo());
				}
				AssertUtil.assertNull(extrinsicObjectType.getExtraMetadata());
				AssertUtil.assertEquals(slots, extrinsicObjectType.getSlot());
				
				AssertUtil.assertNotNull(extrinsicObjectType.getVersionInfo());
				AssertUtil.assertEquals("1.1", extrinsicObjectType.getVersionInfo().getVersionName());
				AssertUtil.assertNull(extrinsicObjectType.getVersionInfo().getComment());
				
				SlotListType responseSlotList = adhocQueryResponse.getResponseSlotList();
				AssertUtil.assertNull(responseSlotList);
			}
		});
		Assert.assertNotNull(result.getDocumentEntries());
		Assert.assertTrue(result.getDocumentEntries().size() > 0);
		Assert.assertEquals(1, result.getDocumentEntries().size());
	}

	@Test
	public void startAfterStopQueryReturnsError() throws Exception {
		// Arrange
		FindDocumentsQuery query = new FindDocumentsQuery();
		query.setPatientId(getPatientId());
		query.setDocumentEntryTypes(Collections.singletonList(DocumentEntryType.ON_DEMAND));
		query.setStatus(Collections.singletonList(AvailabilityStatus.APPROVED));

		DateTime start = DateTime.now();

		query.getServiceStartTime().setFrom(start);
		query.getServiceStopTime().setTo(start.minusDays(1));

		// Act
		QueryResponse result = callWebservice(query);
		// Assert
		Assert.assertFalse(result.getDocumentEntries().size() > 0);
		Assert.assertTrue(result.getErrors().size() > 0);
		Assert.assertTrue(result.getErrors().stream().anyMatch(e -> e.getCustomErrorCode().contains("ikke ligge efter")));
	}

	private MetadataConfiguration loadConfiguration() throws XStreamFileException, URISyntaxException, FileNotFoundException {
		File confFile = Resource.getFile(this.getClass(), "sxametadataconfiguration.xml");
		File filterFile = Resource.getFile(this.getClass(), "sxafilterconfiguration.xml");
		MetadataConfigurationHandler handler = new MetadataConfigurationHandler(confFile, filterFile);
		return handler.readConfiguration();
	}

	private Identifiable getPatientId() {
		return createDummyPerson().getId();
	}

	public Person createDummyPerson() {
		Identifiable id = new Identifiable("2512484916", new AssigningAuthority("1.2.3.4"));
		XcnName name = new XcnName("surName", "givenName", "middleName", "suffix", "prefix", "HighDegree");
		return new Person(id, name);
	}

	public FindDocumentsQuery createDefaultQuery(MetadataContentConfiguration metadataContentConfiguration) {
		FindDocumentsQuery query = new FindDocumentsQuery();
		query.setPatientId(getPatientId());

		QueryUtil.addClassCode(query, metadataContentConfiguration.getClassCode());
		QueryUtil.addTypeCode(query, metadataContentConfiguration.getTypeCode());
		QueryUtil.addConfidentialityCode(query, metadataContentConfiguration.getConfidentialityCode());
		if (metadataContentConfiguration.getEventCodes() != null && !metadataContentConfiguration.getEventCodes().isEmpty()) {
			QueryUtil.addEventCode(query, metadataContentConfiguration.getEventCodes().get(0));
		}
		QueryUtil.addFormatCode(query, metadataContentConfiguration.getFormatCode());
		QueryUtil.addHealtcareFacilityTypeCode(query, metadataContentConfiguration.getHealthcareFacilityTypeCode());
		QueryUtil.addPracticeSettingCode(query, metadataContentConfiguration.getPracticeSettingCode());
		QueryUtil.addDocumentEntryType(query, DocumentEntryType.ON_DEMAND);

		return query;
	}

	private QueryResponse callWebservice(FindDocumentsQuery query) {
		return callWebservice(query, null);
	}

	private QueryResponse callWebservice(FindDocumentsQuery query, AdhocQueryResponseValidator validator) {
		QueryRegistry registry = new QueryRegistry(query);
		AdhocQueryRequest adhocQueryRequest = EbXML30Converters.convert(registry);

		AdhocQueryResponse adhocQueryResponse = iti18PortType.documentRegistryRegistryStoredQuery(adhocQueryRequest);

		if (validator != null) {
			validator.validate(adhocQueryResponse);
		}

		Assert.assertNotNull(adhocQueryResponse);
		Assert.assertNotNull(adhocQueryResponse.getRegistryObjectList());
		QueryResponse queryResponse = EbXML30Converters.convertToQueryResponse(adhocQueryResponse);
		Assert.assertNotNull(queryResponse.getDocumentEntries());

		return queryResponse;
	}

	interface AdhocQueryResponseValidator {
		public void validate(AdhocQueryResponse adhocQueryResponse);
	}
}
