/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentmetadataprovider.impl;

import java.io.File;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import javax.ejb.ConcurrencyManagement;
import javax.ejb.ConcurrencyManagementType;
import javax.ejb.Singleton;
import javax.ejb.Startup;
import javax.enterprise.inject.Produces;
import javax.xml.ws.WebServiceContext;

import dk.nsi.dds.projects.sxa.common.configuration.SXADocumentProviderConfig;
import dk.nsi.util.Config;
import org.apache.log4j.Logger;

import dk.nsi.common.util.SLALogItemFactory;
import dk.nsi.dds.projects.sxa.common.PracticeSettingCodesHandler;
import dk.nsi.dds.projects.sxa.common.PracticeSettingCodesMappingConfiguration;
import dk.nsi.dds.projects.sxa.common.util.configloading.ConfigFileFinder;
import dk.nsi.dds.projects.sxa.common.util.configloading.FileModifiedWatcher;
import dk.nsi.dds.projects.sxa.common.util.configloading.Log4jConfigurationLoader;
import dk.nsi.dds.projects.sxa.common.util.xmlhandling.XStreamFileException;
import dk.nsi.dds.projects.sxa.documentmetadataprovider.configuration.MetadataConfiguration;
import dk.nsi.dds.projects.sxa.documentmetadataprovider.configuration.MetadataConfigurationHandler;
import dk.nsi.dds.projects.sxa.documentmetadataprovider.ws.MetadataProviderWS;
import dk.sdsd.nsp.slalog.api.SLALogConfig;
import dk.sdsd.nsp.slalog.api.SLALogger;

@Singleton
@Startup
@ConcurrencyManagement(ConcurrencyManagementType.BEAN)
public class ConfigurationLoader {
	
	private static final Logger log = Logger.getLogger(ConfigurationLoader.class);
	
	private static final String CONFIGURATION_FILE_NAME = "sxametadataconfiguration.xml";
	private static final String FILTER_FILE_NAME = "sxafilterconfiguration.xml";
	private static final String PRACTICESETTINGCODES_CONFIGURATION_FILE_NAME = "sxapracticesettingcodes.xml";
	private File practiceSettingCodesFile;
	private PracticeSettingCodesHandler practiceSettingCodesHandler;
	private SLALogItemFactory slaLogItemFactory;
	private FileModifiedWatcher configurationFile;
	private FileModifiedWatcher filterFile;
	private MetadataConfigurationHandler metadataConfigurationHandler;
	private Config config;

	@Resource
	private WebServiceContext webServiceContext;

	@PostConstruct
	public void startup() {		
		Log4jConfigurationLoader.readLog4jConfigurationFile("log4j-sxa-documentmetadataprovider-ws.xml");
		setupSLALogging();
		
		configurationFile = new FileModifiedWatcher(ConfigFileFinder.loadFile(CONFIGURATION_FILE_NAME));
		filterFile = new FileModifiedWatcher(ConfigFileFinder.loadFile(FILTER_FILE_NAME));
		metadataConfigurationHandler = new MetadataConfigurationHandler(configurationFile.getFile(), filterFile.getFile());
		config = new SXADocumentProviderConfig("sxadocumentmetadataprovider");
		
		try {
			createPracticeSettingCodesConfigurationHandler();
		} catch (Exception e) {
			//throw new RuntimeException("Exception during service initialization: " + e.getMessage(), e);
			log.fatal("Kunne ikke initialisere "+PRACTICESETTINGCODES_CONFIGURATION_FILE_NAME, e);
		}
	}

	private void createPracticeSettingCodesConfigurationHandler() throws XStreamFileException {
		practiceSettingCodesFile = ConfigFileFinder.loadFile(PRACTICESETTINGCODES_CONFIGURATION_FILE_NAME);
		practiceSettingCodesHandler = new PracticeSettingCodesHandler(practiceSettingCodesFile);
		practiceSettingCodesHandler.readConfiguration();
	}

	private void setupSLALogging() {
		SLALogger slaLogger = new SLALogConfig(MetadataProviderWS.class.getName(), "sxa-documentmetadataprovider").getSLALogger();
		slaLogItemFactory = new SLALogItemFactory(slaLogger, MetadataProviderWS.class.getName());
	}

	@Produces
	public MetadataConfiguration getMetadataConfiguration() throws XStreamFileException {
		return metadataConfigurationHandler.readConfiguration();
	}

	@Produces
	public PracticeSettingCodesMappingConfiguration getPracticeSettingCodesConfigurationHandler() throws XStreamFileException {
		return practiceSettingCodesHandler.readConfiguration();
	}

	@Produces
	public SLALogItemFactory createSLALogItemFactory() {
		return slaLogItemFactory;
	}

	@Produces
	public Config getConfig() {
		return config;
	}

	@Produces
	public WebServiceContext getWebServiceContext() {
		return webServiceContext;
	}
}
