/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentmetadataprovider.impl;

import java.util.List;

import javax.ejb.Stateless;
import javax.inject.Inject;

import org.apache.log4j.Logger;
import org.joda.time.DateTime;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Author;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.AvailabilityStatus;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.DocumentEntry;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.DocumentEntryType;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Identifiable;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.LocalizedString;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.TimeRange;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Version;
import org.openehealth.ipf.commons.ihe.xds.core.requests.query.FindDocumentsQuery;
import org.openehealth.ipf.commons.ihe.xds.core.responses.QueryResponse;
import org.openehealth.ipf.commons.ihe.xds.core.responses.Severity;

import dk.nsi.dds.projects.sxa.common.PracticeSettingCodesMappingConfiguration;
import dk.nsi.dds.projects.sxa.common.QueryIntervalCreator;
import dk.nsi.dds.projects.sxa.common.configuration.SXAMappedValue;
import dk.nsi.dds.projects.sxa.common.core.interval.UnboundableInterval;
import dk.nsi.dds.projects.sxa.common.id.IdConverter;
import dk.nsi.dds.projects.sxa.common.util.ConverterUtil;
import dk.nsi.dds.projects.sxa.common.util.DocumentUniqueIdUtil;
import dk.nsi.dds.projects.sxa.common.util.perflogger.PerformanceLogger;
import dk.nsi.dds.projects.sxa.common.util.perflogger.PerformanceLogger.Monitor;
import dk.nsi.dds.projects.sxa.common.xds.impl.ValueEncoderDecoder;
import dk.nsi.dds.projects.sxa.documentmetadataprovider.configuration.MetadataConfiguration;
import dk.nsi.dds.projects.sxa.documentmetadataprovider.configuration.MetadataContentConfiguration;
import dk.nsi.dds.projects.sxa.documentmetadataprovider.ws.DocumentMetadataProvider;

@Stateless
public class DocumentMetadataProviderImpl implements DocumentMetadataProvider {
	private static final Logger logger = Logger.getLogger(DocumentMetadataProviderImpl.class);

	@Inject
	private MetadataConfiguration metadataConfiguration;
	
	@Inject
	private PracticeSettingCodesMappingConfiguration practiceSettingCodesMappingConfiguration;

	@Override
	public QueryResponse doQuery(FindDocumentsQuery query) {
		logger.debug("Inside DocumentMetadataProviderImpl.doQuery");
		try (Monitor monitor = new PerformanceLogger.Monitor("DocumentMetadataProviderImpl.doQuery")) {
			QueryResponse response = new QueryResponse();
			try {
				if (query != null) {
					DocumentEntry documentMetadata = getDocumentMetadata(query);
					if (documentMetadata != null) {
						response.getDocumentEntries().add(documentMetadata);
					}
				} else {
					logger.debug("FindDocumentsQuery is null");
					response = MetadataProviderQueryImpl.createAndAddCustomError(response, "Forespørgslen indeholdt ikke noget data", getCodeContext(), Severity.ERROR);
				}
			} catch (Exception e) {				
				response = MetadataProviderQueryImpl.createAndAddCustomError(response, e.getMessage(), getCodeContext(), Severity.ERROR);
			}
			return response;
		}
	}

	public String getCodeContext() {
		return ValueEncoderDecoder.encodeErrorCodeContext(null, getHomeCommunityId(), getRepositoryUniqueId());
	}
	
	protected MetadataContentConfiguration getMetadataContentConfiguration() {
		return metadataConfiguration != null ? metadataConfiguration.getMetadataContentConfiguration() : null;
	}
	
	@Override
	public String getHomeCommunityId() {	
		return getMetadataContentConfiguration() != null ? getMetadataContentConfiguration().getHomeCommunityId() : null;
	}
	
	@Override
	public String getRepositoryUniqueId() {	
		return getMetadataContentConfiguration() != null ? getMetadataContentConfiguration().getRepositoryUniqueId() : null;
	}

	private DocumentEntry getDocumentMetadata(FindDocumentsQuery query) throws Exception {
		if (metadataConfiguration != null) {
			MetadataQueryFilter metadataQueryFilter = new MetadataQueryFilter(metadataConfiguration.getMetadataConfigurationFilter());
			MetadataContentConfiguration metadataContentConfiguration = metadataConfiguration.getMetadataContentConfiguration();

			checkIntervalAndSetDefaultValues(query);

			if (metadataQueryFilter != null && metadataQueryFilter.isValidQuery(query)) {
				if(!MetadataQueryFilter.isSearchForStatus(query, AvailabilityStatus.APPROVED)) {
					//We have a status in query, but none is APPROVED
					logger.info("Returnere tomt svar, da søgning ikke indbefattede APPROVED status.");
					return null;					
				}
				
				DocumentEntry documentMetadata = new DocumentEntry();
				documentMetadata.setPatientId(query.getPatientId());
				documentMetadata.setSourcePatientId(query.getPatientId());

				final String documentUniqueId = getDocumentUniqueId(query, metadataContentConfiguration.getUniqueIdRoot());
				documentMetadata.setUniqueId(documentUniqueId);

				Author author = new Author();
				author.setAuthorPerson(metadataContentConfiguration.getAuthorPerson());
				author.getAuthorInstitution().add(metadataContentConfiguration.getAuthorInstitution());
				documentMetadata.getAuthors().add(author);

				documentMetadata.setClassCode(ConverterUtil.convert(metadataContentConfiguration.getClassCode()));
				documentMetadata.getConfidentialityCodes().add(ConverterUtil.convert(metadataContentConfiguration.getConfidentialityCode()));
				documentMetadata.getEventCodeList().addAll(ConverterUtil.convert(metadataContentConfiguration.getEventCodes()));
				documentMetadata.setFormatCode(ConverterUtil.convert(metadataContentConfiguration.getFormatCode()));
				documentMetadata.setHealthcareFacilityTypeCode(ConverterUtil.convert(metadataContentConfiguration.getHealthcareFacilityTypeCode()));
				documentMetadata.setPracticeSettingCode(ConverterUtil.convert(metadataContentConfiguration.getPracticeSettingCode()));

				UnboundableInterval startStopInterval = QueryIntervalCreator.createQueryInterval(query.getPatientId().getId(), getServiceStartDateTime(query), getServiceStopDateTime(query));
				
				if(startStopInterval.getStart() != null) {
					documentMetadata.setServiceStartTime(startStopInterval.getStart());
				}
				
				if(startStopInterval.getEnd() != null) {
					documentMetadata.setServiceStopTime(startStopInterval.getEnd());				
				}


				documentMetadata.setAvailabilityStatus(AvailabilityStatus.APPROVED);
				documentMetadata.setTypeCode(ConverterUtil.convert(metadataContentConfiguration.getTypeCode()));
				documentMetadata.setMimeType(metadataContentConfiguration.getMimeType());
				documentMetadata.setLanguageCode(metadataContentConfiguration.getLanguageCode());
				
				documentMetadata.setHomeCommunityId(metadataContentConfiguration.getHomeCommunityId());
				documentMetadata.setRepositoryUniqueId(metadataContentConfiguration.getRepositoryUniqueId());

				LocalizedString title = new LocalizedString("Lab Reports for " + (query.getPatientId() != null ? query.getPatientId().getId() : ""), null, null);
				documentMetadata.setTitle(title);
				documentMetadata.setEntryUuid(IdConverter.convertOidToUuid(documentUniqueId.replace(".", "")));
				documentMetadata.setType(DocumentEntryType.ON_DEMAND);
				
				/*Added*/
				documentMetadata.setVersion(new Version("1.1"));
				
				return documentMetadata;
			} else {
				logger.debug("metadataQueryFilter is null or the query is invalid");
			}
		} else {
			logger.debug("MetadataConfiguration is null");
		}
		
		return null;
	}

	private String getDocumentUniqueId(FindDocumentsQuery query, String uniqueIdRoot) {
		List<SXAMappedValue> validPracticeSettingCodes = practiceSettingCodesMappingConfiguration.getPracticeSettingCodesConfiguration().getPracticeSettingCodes();

		final Identifiable patientId = query.getPatientId();
		final DateTime serviceStartDateTime = getServiceStartDateTime(query);
		final DateTime serviceStopDateTime = getServiceStopDateTime(query);

		return DocumentUniqueIdUtil.createUniqueId(uniqueIdRoot, patientId, serviceStartDateTime, serviceStopDateTime, query.getPracticeSettingCodes(), validPracticeSettingCodes);
	}

	private DateTime getServiceStartDateTime(FindDocumentsQuery query) {
		return getDateTime(query.getServiceStartTime(), true);
	}

	private DateTime getServiceStopDateTime(FindDocumentsQuery query) {
		return getDateTime(query.getServiceStopTime(), false);
	}

	private DateTime getDateTime(TimeRange timeRange, boolean preferBeginning) {
		if (timeRange.getFrom() == null && timeRange.getTo() == null) {
			return null;
		}

		if (preferBeginning) {
			return (timeRange.getFrom() != null) ? timeRange.getFrom() : timeRange.getTo();
		} else {
			return (timeRange.getTo() != null) ? timeRange.getTo() : timeRange.getFrom();
		}
	}

	private void checkIntervalAndSetDefaultValues(FindDocumentsQuery query) throws Exception {
		DateTime start = getServiceStartDateTime(query);
		DateTime end = getServiceStopDateTime(query);
		if (start != null && end != null && start.isAfter(end)) {
			throw new Exception("Tidspunktet 'Fra' (" + start.toString() + ") må ikke ligge efter 'Til' (" + end.toString() + ")");
		}
	}
}
