/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentmetadataprovider.impl;

import java.util.List;

import org.apache.commons.collections.CollectionUtils;
import org.apache.log4j.Logger;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.AvailabilityStatus;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Code;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.DocumentEntryType;
import org.openehealth.ipf.commons.ihe.xds.core.requests.query.FindDocumentsQuery;
import org.openehealth.ipf.commons.ihe.xds.core.requests.query.QueryList;

import dk.nsi.dds.projects.sxa.common.configuration.SXAMappedValue;
import dk.nsi.dds.projects.sxa.common.util.ConverterUtil;
import dk.nsi.dds.projects.sxa.documentmetadataprovider.configuration.MetadataFilterConfiguration;

public class MetadataQueryFilter {

	private static final Logger logger = Logger.getLogger(MetadataQueryFilter.class);
	private static final CodedValueIgnoreDescriptionComparator codedValueComparator = new CodedValueIgnoreDescriptionComparator();

	private MetadataFilterConfiguration configurationFilter;

	public MetadataQueryFilter(MetadataFilterConfiguration configurationFilter) {
		this.configurationFilter = configurationFilter;
	}

	public boolean isValidQuery(FindDocumentsQuery query) {
		if (query == null || configurationFilter == null) {
			logger.error("query is null or the filter is null");
			return false;
		}
		try {
			// Check that every condition is true
			// Throw exception otherwise
			checkTrue(isQueryForOnDemandDocuments(query), "Not query on demand documents");
			
			checkTrue(query.getCreationTime() == null || query.getCreationTime().getFrom() == null, "Query has creation time start");
			checkTrue(query.getCreationTime() == null || query.getCreationTime().getTo() == null, "Query has creation time end");
			checkTrue(isEmpty(query.getAuthorPersons()), "Query has authors");
			checkTrue(checkList(query.getClassCodes(), configurationFilter.getClassCodes()), "Query class codes not in filter");
			checkTrue(checkList(query.getTypeCodes(), configurationFilter.getTypeCodes()), "Query type codes not in filter");
			checkTrue(checkList(query.getEventCodes(), configurationFilter.getEventCodes()), "Query event codes not in filter");
			checkTrue(checkList(query.getConfidentialityCodes(), configurationFilter.getConfidentialityCodes()), "Query confidentiality codes not in filter");
			checkTrue(checkList(query.getFormatCodes(), configurationFilter.getFormatCodes()), "Query format codes not in filter");
			checkTrue(checkList(query.getPracticeSettingCodes(), configurationFilter.getPracticeSettingCodes()), "Practice Setting codes not in filter");
			return true;
		} catch (QueryInvalidException e) {
			logger.debug(String.format("Query is invalid due to the following: %s", e.getMessage()));
			return false;
		}
	}

	public static boolean isSearchForAllStatus(FindDocumentsQuery query) {
		return CollectionUtils.isEmpty(query.getStatus());
	}
	
	public static boolean isSearchForStatus(FindDocumentsQuery query, AvailabilityStatus status) {
		return isSearchForAllStatus(query) || contains(query.getStatus(), status);
	}
	
	@SuppressWarnings("unused")
	private static boolean isQueryForStableAndOnDemandDocuments(FindDocumentsQuery query) {
		return isQueryForStableDocuments(query) && isQueryForOnDemandDocuments(query);
	}
	
	private static boolean isQueryForOnDemandDocuments(FindDocumentsQuery query) {
		return contains(query.getDocumentEntryTypes(), DocumentEntryType.ON_DEMAND);
	}

	private static boolean isQueryForStableDocuments(FindDocumentsQuery query) {
		return contains(query.getDocumentEntryTypes(), DocumentEntryType.STABLE);
	}

	public static <E> boolean contains(List<E> list, E e) {
		if (!isEmpty(list)) {
			return list.contains(e);
		}
		return false;
	}

	private static <E> boolean isEmpty(List<E> list) {
		return list == null || list.isEmpty();
	}
	
	private static boolean checkList(QueryList<Code> queryValues, List<SXAMappedValue> filterValues) {
		if (queryValues == null || isEmpty(queryValues.getOuterList())) {
			return true;
		}

		for (List<Code> list : queryValues.getOuterList()) {
			if (checkList(list, filterValues)) {
				return true;
			}
		}

		return false;
	}

	private static boolean checkList(List<Code> queryValues, List<SXAMappedValue> filterValues) {
		if (isEmpty(queryValues) || isEmpty(filterValues)) {
			return true;
		}

		for (SXAMappedValue filterValue : filterValues) {
			for (Code queryValue : queryValues) {
				if (codedValueComparator.compare(ConverterUtil.convert(filterValue), queryValue) == 0) {
					return true;
				}
			}
		}
		return false;
	}

	public void checkTrue(boolean condition, String errorMessage) throws QueryInvalidException {
		if (!condition) {
			throw new QueryInvalidException(errorMessage);
		}
	}

	private class QueryInvalidException extends RuntimeException {
		private static final long serialVersionUID = 1L;

		public QueryInvalidException(String message) {
			super(message);
		}
	}
}
