/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (SDS), Denmark
 * (http://sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (SDS), Denmark (http://sundhedsdatastyrelsen.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentmetadataprovider.ws;

import static javax.xml.ws.soap.SOAPBinding.SOAP12HTTP_BINDING;

import java.math.BigInteger;

import javax.annotation.Resource;
import javax.ejb.EJB;
import javax.enterprise.inject.Produces;
import javax.inject.Inject;
import javax.interceptor.Interceptors;
import javax.jws.WebService;
import javax.xml.ws.BindingType;
import javax.xml.ws.WebServiceContext;
import javax.xml.ws.soap.Addressing;

import dk.nsi.dds.projects.sxa.common.security.AccessValidationInterceptor;
import org.apache.log4j.Logger;
import org.openehealth.ipf.commons.ihe.xds.core.ebxml.EbXMLQueryResponse;
import org.openehealth.ipf.commons.ihe.xds.core.ebxml.ebxml30.EbXMLAdhocQueryRequest30;
import org.openehealth.ipf.commons.ihe.xds.core.ebxml.ebxml30.EbXMLFactory30;
import org.openehealth.ipf.commons.ihe.xds.core.requests.QueryRegistry;
import org.openehealth.ipf.commons.ihe.xds.core.requests.query.QueryType;
import org.openehealth.ipf.commons.ihe.xds.core.responses.ErrorCode;
import org.openehealth.ipf.commons.ihe.xds.core.responses.QueryResponse;
import org.openehealth.ipf.commons.ihe.xds.core.responses.Severity;
import org.openehealth.ipf.commons.ihe.xds.core.responses.Status;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.query.AdhocQueryRequest;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.query.AdhocQueryResponse;
import org.openehealth.ipf.commons.ihe.xds.core.transform.requests.QueryRegistryTransformer;
import org.openehealth.ipf.commons.ihe.xds.core.transform.responses.QueryResponseTransformer;
import org.openehealth.ipf.commons.ihe.xds.core.validate.XDSMetaDataException;
import org.openehealth.ipf.commons.ihe.xds.iti18.Iti18PortType;

import dk.nsi.common.util.SLALogItemFactory;
import dk.nsi.common.util.SLALoggingDataHandler;
import dk.nsi.dds.projects.sxa.common.util.perflogger.PerformanceLogger;
import dk.nsi.dds.projects.sxa.common.util.perflogger.PerformanceLogger.Monitor;
import dk.nsi.dds.projects.sxa.documentmetadataprovider.impl.MetadataProviderQuery;
import dk.nsi.dds.projects.sxa.documentmetadataprovider.impl.MetadataProviderQueryImpl;

@WebService(serviceName = "iti18")
@BindingType(value = SOAP12HTTP_BINDING)
@Addressing(enabled = true, required = false)
public class MetadataProviderWS implements Iti18PortType {
	private static Logger logger = Logger.getLogger(MetadataProviderWS.class);

	@Resource
	private WebServiceContext webServiceContext;
	@Inject
	private SLALogItemFactory slaLogItemFactory;
	@EJB
	private DocumentMetadataProvider documentMetadataProvider;

	@Produces
	public SLALoggingDataHandler createSLALoggingDataHandler() {
		return new SLALoggingDataHandler(slaLogItemFactory, webServiceContext);
	}

	@Override
	@Interceptors({AccessValidationInterceptor.class})
	public AdhocQueryResponse documentRegistryRegistryStoredQuery(AdhocQueryRequest request) {
		try (Monitor monitor = new PerformanceLogger.Monitor("MetadataProviderWS.documentRegistryRegistryStoredQuery")) {
			logger.debug("Calling iti18 endpoint request:" + request);

			QueryResponse response = null;

			if (documentMetadataProvider != null) {
				try {
					QueryRegistry registry = toRegistry(request);
					MetadataProviderQuery queryProvider = new MetadataProviderQueryImpl(documentMetadataProvider);
					response = queryProvider.handleRequest(registry);
				} catch (XDSMetaDataException e) {
					response = MetadataProviderQueryImpl.createAndAddCustomError(e.getMessage(), documentMetadataProvider.getCodeContext(), Severity.ERROR);
				}
			} else {
				logger.error("MetadataProviderWS: DocumentMetadataProvider is null");
				response = MetadataProviderQueryImpl.createAndAddStandarError(ErrorCode.REGISTRY_ERROR, null, Severity.WARNING);
			}

			response = ensureStatus(response);
			AdhocQueryResponse adhocQueryResponse = convert(response);
			logger.debug("Done iti18 response:" + adhocQueryResponse);
			return adhocQueryResponse;
		}
	}

	public static QueryRegistry toRegistry(AdhocQueryRequest request) {
		EbXMLAdhocQueryRequest30 ebXMLAdhocQueryRequest30 = new EbXMLAdhocQueryRequest30(request);
		QueryRegistry queryRegistry = new QueryRegistryTransformer().fromEbXML(ebXMLAdhocQueryRequest30);
		return queryRegistry;
	}

	static QueryResponse ensureStatus(QueryResponse response) {
		if (response.getStatus() == null) {
			final long metadataCount = response.getDocumentEntries().stream().count();
			final long errorCount = response.getErrors().stream().filter(e -> e.getSeverity() == Severity.ERROR).count();

			if (errorCount == 0) {
				response.setStatus(Status.SUCCESS);
			} else if (metadataCount > 0) {
				response.setStatus(Status.PARTIAL_SUCCESS);
			} else {
				response.setStatus(Status.FAILURE);
			}
		}
		return response;
	}

	public static AdhocQueryResponse convert(QueryResponse response) {
		final long metadataCount = response.getDocumentEntries().stream().count();

		EbXMLFactory30 ebXMLFactory30 = new EbXMLFactory30();
		EbXMLQueryResponse ebxmlQueryResponse = new QueryResponseTransformer(ebXMLFactory30).toEbXML(response);
		AdhocQueryResponse adhocQueryResponse = (AdhocQueryResponse) ebxmlQueryResponse.getInternal();

		adhocQueryResponse.setRequestId(QueryType.FIND_DOCUMENTS.getId());
		adhocQueryResponse.setTotalResultCount(BigInteger.valueOf(metadataCount));

		return adhocQueryResponse;
	}
}
