/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentmetadataprovider;

import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.runners.MockitoJUnitRunner;
import org.openehealth.ipf.commons.ihe.xds.core.ebxml.ebxml30.EbXMLFactory30;
import org.openehealth.ipf.commons.ihe.xds.core.ebxml.ebxml30.EbXMLQueryResponse30;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.DocumentEntry;
import org.openehealth.ipf.commons.ihe.xds.core.requests.query.FindDocumentsQuery;
import org.openehealth.ipf.commons.ihe.xds.core.responses.QueryResponse;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.query.AdhocQueryResponse;
import org.openehealth.ipf.commons.ihe.xds.core.transform.responses.QueryResponseTransformer;

import dk.nsi.dds.projects.sxa.common.PracticeSettingCodesConfiguration;
import dk.nsi.dds.projects.sxa.common.PracticeSettingCodesMappingConfiguration;
import dk.nsi.dds.projects.sxa.common.TestDataHelper;
import dk.nsi.dds.projects.sxa.common.util.configloading.Log4jConfigurationLoader;
import dk.nsi.dds.projects.sxa.documentmetadataprovider.configuration.MetadataConfiguration;
import dk.nsi.dds.projects.sxa.documentmetadataprovider.impl.DocumentMetadataProviderImpl;
import dk.nsi.dds.projects.sxa.documentmetadataprovider.ws.DocumentMetadataProvider;
import org.junit.Assert;

@RunWith(MockitoJUnitRunner.class)
public class DocumentMetadataProviderImplTest {

	private TestDataHelper configTestHelper = new TestDataHelper();
	private DocumentMetadataProviderTestDataHelper metadataTestHelper = new DocumentMetadataProviderTestDataHelper(configTestHelper);
	private DocumentMetadataProviderHelper helper = new DocumentMetadataProviderHelper();

	@Mock
	private MetadataConfiguration metadataConfiguration;

	@Mock
	private PracticeSettingCodesMappingConfiguration practiceSettingCodesMappingConfiguration;

	@InjectMocks
	private DocumentMetadataProvider sxaDocumentMetadataProvider = new DocumentMetadataProviderImpl();

	@Before
	public void setup() {
		Log4jConfigurationLoader.readLog4jConfigurationFile("log4j-sxa-documentmetadataprovider-ws.xml");
	}

	@Test
	public void queryForSxaMetadata() throws Exception {

		// Arrange
		MetadataConfiguration sxaConfiguration = setupMetadataConfigurationMock();
		FindDocumentsQuery query = metadataTestHelper.createSxaQuery();

		// Act
		AdhocQueryResponse result = helper.doRequest(query, sxaDocumentMetadataProvider);

		// Assert
		checkSuccesfullQueryResult(result, query, sxaConfiguration);
	}

	@Test
	public void queryForSxaMetadataWithPracticeSettingCode() throws Exception {
		// Arrange
		MetadataConfiguration sxaConfiguration = setupMetadataConfigurationMock();
		FindDocumentsQuery query = metadataTestHelper.createSxaQueryWithPracticeSettingCode();

		// Act
		AdhocQueryResponse result = helper.doRequest(query, sxaDocumentMetadataProvider);

		// Assert
		checkSuccesfullQueryResult(result, query, sxaConfiguration);
	}

	@Test
	public void queryForSxaMetadataWithMultiplePracticeSettingCode() throws Exception {

		// Arrange
		MetadataConfiguration sxaConfiguration = setupMetadataConfigurationMock();
		FindDocumentsQuery query = metadataTestHelper.createSxaQueryWithMultiplePracticeSettingCode();

		// Act
		AdhocQueryResponse result = helper.doRequest(query, sxaDocumentMetadataProvider);

		// Assert
		checkSuccesfullQueryResult(result, query, sxaConfiguration);
	}

	@Test
	public void queryForSxaMetadataWithInvalidPracticeSettingCode() throws Exception {

		// Arrange
		setupMetadataConfigurationMock();
		FindDocumentsQuery query = metadataTestHelper.createSxaQueryWithInvalidPracticeSettingCode();

		// Act
		AdhocQueryResponse result = helper.doRequest(query, sxaDocumentMetadataProvider);

		// Assert
		Assert.assertNotNull(result);
		Assert.assertTrue(result.getRegistryObjectList().getIdentifiable().isEmpty());
	}

	private MetadataConfiguration setupMetadataConfigurationMock() {
		MetadataConfiguration sxaConfiguration = metadataTestHelper.createDefaultConfiguration();
		Mockito.when(metadataConfiguration.getMetadataConfigurationFilter()).thenReturn(sxaConfiguration.getMetadataConfigurationFilter());
		Mockito.when(metadataConfiguration.getMetadataContentConfiguration()).thenReturn(sxaConfiguration.getMetadataContentConfiguration());

		PracticeSettingCodesConfiguration practiceSettingCodesConfigurationMock = new PracticeSettingCodesConfiguration();
		practiceSettingCodesConfigurationMock.setPracticeSettingCodes(configTestHelper.getPracticeSettingCodes());

		Mockito.when(practiceSettingCodesMappingConfiguration.getPracticeSettingCodesConfiguration()).thenReturn(practiceSettingCodesConfigurationMock);

		return sxaConfiguration;
	}

	@Test
	public void queryForMoreSpecificThanSxaMetadata() throws Exception {

		// Arrange
		setupMetadataConfigurationMock();
		FindDocumentsQuery query = metadataTestHelper.createQueryBeyondSxaMetadata();

		// Act
		AdhocQueryResponse result = helper.doRequest(query, sxaDocumentMetadataProvider);

		// Assert
		checkDocumentMetadataList(result, false);
	}

	@Test
	public void queryForLessSpecificThanSxaMetadata() throws Exception {

		// Arrange
		setupMetadataConfigurationMock();
		FindDocumentsQuery query = metadataTestHelper.createQueryLessSpecificThanSxaMetadata();

		// Act
		AdhocQueryResponse result = helper.doRequest(query, sxaDocumentMetadataProvider);

		// Assert
		checkDocumentMetadataList(result, false);
	}

	@Test
	public void queryForOtherThanSxaMetadata() throws Exception {

		// Arrange
		setupMetadataConfigurationMock();
		FindDocumentsQuery query = metadataTestHelper.createNonSxaQuery();

		// Act
		AdhocQueryResponse result = helper.doRequest(query, sxaDocumentMetadataProvider);

		// Assert
		checkDocumentMetadataList(result, true);
	}

	private void checkSuccesfullQueryResult(AdhocQueryResponse result, FindDocumentsQuery query, MetadataConfiguration sxaConfiguration) throws Exception {
		QueryResponse queryResponse = convert(result);
		
		// Assert
		Assert.assertNotNull(queryResponse);
		Assert.assertNotNull(queryResponse.getDocumentEntries());
		Assert.assertFalse(queryResponse.getDocumentEntries().isEmpty());
		Assert.assertEquals(1, queryResponse.getDocumentEntries().size());

		DocumentEntry metadataResult = queryResponse.getDocumentEntries().get(0);

		Assert.assertNotNull(metadataResult);
		metadataTestHelper.assertBasicMetadata(query, metadataResult);
		metadataTestHelper.compareConfigurationWithDocumentMetadata(query, sxaConfiguration, metadataResult);
	}

	private void checkDocumentMetadataList(AdhocQueryResponse result, boolean expectedResult) throws Exception {
		QueryResponse queryResponse = convert(result);
		Assert.assertNotNull(queryResponse);
		Assert.assertNotNull(queryResponse.getDocumentEntries());
		Assert.assertEquals(expectedResult, queryResponse.getDocumentEntries().isEmpty());
	}

	public static QueryResponse convert(AdhocQueryResponse response) {
		return new QueryResponseTransformer(new EbXMLFactory30()).fromEbXML(new EbXMLQueryResponse30(response));
	}
}
