/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentmetadataprovider;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import org.junit.Assert;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.AssigningAuthority;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.AvailabilityStatus;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Code;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.DocumentEntry;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.DocumentEntryType;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Identifiable;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Name;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Organization;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Person;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.XcnName;
import org.openehealth.ipf.commons.ihe.xds.core.requests.query.FindDocumentsQuery;
import org.openehealth.ipf.commons.ihe.xds.core.requests.query.QueryList;

import ca.uhn.hl7v2.model.v25.datatype.XCN;
import dk.nsi.dds.projects.sxa.common.TestDataHelper;
import dk.nsi.dds.projects.sxa.common.configuration.SXAMappedValue;
import dk.nsi.dds.projects.sxa.common.testutil.AssertUtil;
import dk.nsi.dds.projects.sxa.common.util.ConverterUtil;
import dk.nsi.dds.projects.sxa.common.util.QueryUtil;
import dk.nsi.dds.projects.sxa.documentmetadataprovider.configuration.MetadataConfiguration;
import dk.nsi.dds.projects.sxa.documentmetadataprovider.configuration.MetadataContentConfiguration;
import dk.nsi.dds.projects.sxa.documentmetadataprovider.configuration.MetadataFilterConfiguration;

public class DocumentMetadataProviderTestDataHelper {

	private TestDataHelper configTestHelper;

	private SXAMappedValue classCodeSxa = new SXAMappedValue("summary", "TBD", "summary");
	private SXAMappedValue confidentialityCodeSxa = new SXAMappedValue("confidentialityCode", "confidentialityCodeSystem");
	public SXAMappedValue eventCodeSxa = new SXAMappedValue("BOHA0", "TBD", "Antikoagulationsbehandling");
	public SXAMappedValue eventCodeNonSxa = new SXAMappedValue("BOB", "TBD", "Behandling");
	private SXAMappedValue formatCodeSxa = new SXAMappedValue("urn:ad:dk:medcom:labreports:svareksponeringsservice", "1.2.208.184.100.10", "Laboratoriesvar (samling af svar)");
	private SXAMappedValue typeCodeSxa = new SXAMappedValue("11502-2", "2.16.840.1.113883.6.1", "LABORATORY REPORT.TOTAL");
	private SXAMappedValue typeCodeNonSxa = new SXAMappedValue("53576-5", "2.16.840.1.113883.6.1", "Personal Health Monitoring Report");

	public DocumentMetadataProviderTestDataHelper(TestDataHelper configTestHelper) {
		this.configTestHelper = configTestHelper;
	}

	public FindDocumentsQuery createSxaQuery() {
		return createQuery(typeCodeSxa, null, null, null);
	}

	public FindDocumentsQuery createSxaQueryWithPracticeSettingCode() {
		return createQuery(typeCodeSxa, null, null, new ArrayList<>(Arrays.asList(configTestHelper.practiceSettingCodeForBiochemistry)));
	}

	public FindDocumentsQuery createSxaQueryWithMultiplePracticeSettingCode() {
		return createQuery(typeCodeSxa, null, null, new ArrayList<>(Arrays.asList(configTestHelper.practiceSettingCodeForMicrobiology,
				configTestHelper.practiceSettingCodeForBiochemistry, configTestHelper.practiceSettingCodeForPathology)));
	}

	public FindDocumentsQuery createSxaQueryWithInvalidPracticeSettingCode() {
		return createQuery(typeCodeSxa, null, null, new ArrayList<>(Arrays.asList(configTestHelper.practiceSettingCodeNonSxa)));
	}

	public FindDocumentsQuery createNonSxaQuery() {
		return createQuery(typeCodeNonSxa, null, null, null);
	}

	public FindDocumentsQuery createQueryBeyondSxaMetadata() {
		FindDocumentsQuery query = createSxaQuery();
		QueryUtil.addHealtcareFacilityTypeCode(query, new SXAMappedValue("code6", "codesystem6"));
		return query;
	}

	public FindDocumentsQuery createQueryLessSpecificThanSxaMetadata() {
		FindDocumentsQuery result = createSxaQuery();
		result.getFormatCodes().clear();
		return result;
	}

	public FindDocumentsQuery createQuery(SXAMappedValue typeCode, SXAMappedValue classCode, SXAMappedValue eventCode, List<SXAMappedValue> practiceSettingCodes) {
		FindDocumentsQuery query = new FindDocumentsQuery();
		query.setPatientId(createDummyPerson().getId());

		List<AvailabilityStatus> statuses = query.getStatus();
		if (statuses == null) {
			statuses = new ArrayList<>();
			query.setStatus(statuses);
		}
		statuses.add(AvailabilityStatus.APPROVED);

		List<DocumentEntryType> documentEntryTypes = query.getDocumentEntryTypes();
		if (documentEntryTypes == null) {
			documentEntryTypes = new ArrayList<>();
			query.setDocumentEntryTypes(documentEntryTypes);
		}
		documentEntryTypes.add(DocumentEntryType.ON_DEMAND);

		if (classCode != null) {
			QueryUtil.addClassCode(query, classCode);
		}

		List<Code> typeCodes = query.getTypeCodes();
		if (typeCodes == null) {
			typeCodes = new ArrayList<>();
			query.setTypeCodes(typeCodes);
		}
		typeCodes.add(ConverterUtil.convert(typeCode));

		QueryList<Code> confidentialityCodes = query.getConfidentialityCodes();
		if (confidentialityCodes == null) {
			confidentialityCodes = new QueryList<>();
			query.setConfidentialityCodes(confidentialityCodes);
		}
		confidentialityCodes.getOuterList().add(Collections.singletonList(ConverterUtil.convert(confidentialityCodeSxa)));

		List<Code> formatCodes = query.getFormatCodes();
		if (formatCodes == null) {
			formatCodes = new ArrayList<>();
			query.setFormatCodes(formatCodes);
		}
		formatCodes.add(ConverterUtil.convert(formatCodeSxa));

		if (eventCode != null) {
			QueryList<Code> eventCodes = query.getEventCodes();
			if (eventCodes == null) {
				eventCodes = new QueryList<>();
				query.setEventCodes(eventCodes);
			}
			eventCodes.getOuterList().add(Collections.singletonList(ConverterUtil.convert(eventCode)));
		}

		if (practiceSettingCodes != null) {
			query.setPracticeSettingCodes(ConverterUtil.convert(practiceSettingCodes));
		}

		return query;
	}

	public Person createDummyPerson() {
		Identifiable id = new Identifiable("2512484916", new AssigningAuthority("1.2.3.4", null));
		Name<XCN> name = new XcnName("surname", "givenName", "middleName", "suffix", "prefix", "degree");
		return new Person(id, name);
	}

	public MetadataConfiguration createDefaultConfiguration() {
		String expectedHomeValue = "homevalue";
		String expectedRepoValue = "repovalue";
		String expectedUniqueIdRoot = "1.2.3.4";
		MetadataContentConfiguration metadataContentConfig = new MetadataContentConfiguration();
		metadataContentConfig.setClassCode(classCodeSxa);
		metadataContentConfig.setConfidentialityCode(confidentialityCodeSxa);
		metadataContentConfig.setTypeCode(typeCodeSxa);
		metadataContentConfig.setLanguageCode("languageCode");
		metadataContentConfig.setMimeType("mimeType");
		metadataContentConfig.setPracticeSettingCode(ConverterUtil.convert(configTestHelper.practiceSettingCodeForMicrobiology));
		metadataContentConfig.setHomeCommunityId(expectedHomeValue);
		metadataContentConfig.setHealthcareFacilityTypeCode(new Code(null, null, "2.16.840.1.113883.6.96"));
		metadataContentConfig.setAuthorInstitution(new Organization("institution", "1234", new AssigningAuthority("2.16.15.16")));
		metadataContentConfig.setFormatCode(formatCodeSxa);
		metadataContentConfig.setRepositoryUniqueId(expectedRepoValue);
		metadataContentConfig.setUniqueIdRoot(expectedUniqueIdRoot);

		MetadataFilterConfiguration filterConfig = new MetadataFilterConfiguration();
		filterConfig.addClassCode(classCodeSxa);
		filterConfig.addTypeCode(typeCodeSxa);
		filterConfig.addConfidentialityCodes(confidentialityCodeSxa);
		filterConfig.addFormatCodes(formatCodeSxa);
		filterConfig.addPracticeSettingCodes(configTestHelper.practiceSettingCodeForMicrobiology);
		filterConfig.addPracticeSettingCodes(configTestHelper.practiceSettingCodeForBiochemistry);
		filterConfig.addPracticeSettingCodes(configTestHelper.practiceSettingCodeForPathology);

		MetadataConfiguration metadataConfig = new MetadataConfiguration();
		metadataConfig.setMetadataConfigurationFilter(filterConfig);
		metadataConfig.setMetadataContentConfiguration(metadataContentConfig);

		return metadataConfig;
	}

	public void assertBasicMetadata(FindDocumentsQuery query, DocumentEntry documentEntry) {
		Assert.assertEquals("Lab Reports for " + query.getPatientId().getId(), documentEntry.getTitle().getValue());
	}

	public void compareConfigurationWithDocumentMetadata(FindDocumentsQuery query, MetadataConfiguration sxaConfiguration, DocumentEntry documentEntry) {
		MetadataContentConfiguration sxaMetadataConfiguration = sxaConfiguration.getMetadataContentConfiguration();

		if (query.getClassCodes() != null && !query.getClassCodes().isEmpty()) {
			AssertUtil.assertTrue(compareLists(query.getClassCodes(), documentEntry.getClassCode()));
		}
		if (query.getEventCodes() != null && !query.getEventCodes().getOuterList().isEmpty()) {
			AssertUtil.assertTrue(comparePartialLists(query.getEventCodes(), documentEntry.getEventCodeList()));
		}

		AssertUtil.assertTrue(compareLists(query.getTypeCodes(), documentEntry.getTypeCode()));
		AssertUtil.assertTrue(compareLists(query.getStatus(), documentEntry.getAvailabilityStatus()));
		AssertUtil.assertEquals(query.getPatientId(), documentEntry.getPatientId());
		AssertUtil.assertEquals(query.getPatientId(), documentEntry.getSourcePatientId());

		AssertUtil.assertEquals(sxaMetadataConfiguration.getAuthorInstitution(), documentEntry.getAuthors().get(0).getAuthorInstitution().get(0));
		AssertUtil.assertEquals(sxaMetadataConfiguration.getAuthorPerson(), documentEntry.getAuthors().get(0).getAuthorPerson());
		AssertUtil.assertEquals(sxaMetadataConfiguration.getConfidentialityCode(), documentEntry.getConfidentialityCodes().get(0));
		AssertUtil.assertEquals(sxaMetadataConfiguration.getFormatCode(), documentEntry.getFormatCode());
		AssertUtil.assertEquals(sxaMetadataConfiguration.getHealthcareFacilityTypeCode(), documentEntry.getHealthcareFacilityTypeCode(), true);
		AssertUtil.assertEquals(sxaMetadataConfiguration.getPracticeSettingCode(), documentEntry.getPracticeSettingCode());

		AssertUtil.assertEquals(sxaMetadataConfiguration.getHomeCommunityId(), documentEntry.getHomeCommunityId());
		AssertUtil.assertEquals(sxaMetadataConfiguration.getLanguageCode(), documentEntry.getLanguageCode());
		AssertUtil.assertEquals(sxaMetadataConfiguration.getMimeType(), documentEntry.getMimeType());
		AssertUtil.assertEquals(sxaMetadataConfiguration.getRepositoryUniqueId(), documentEntry.getRepositoryUniqueId());
		AssertUtil.assertTrue(documentEntry.getUniqueId().startsWith(sxaMetadataConfiguration.getUniqueIdRoot()));
	}

//	@SuppressWarnings({ "serial" })
//	private Code getFixedCodeComparator(Code in) {
//		return new Code(in.getHapiObject()) {
//			@Override
//			public boolean equals(Object o) {
//				boolean equal = false;
//				if (o instanceof Code) {
//					Code other = (Code) o;
//					equal = getCode().equals(other.getCode());
//					equal = equal && getSchemeName().equals(other.getSchemeName());
//					equal = equal && (getDisplayName() != null ? other.getDisplayName() != null : true);
//					equal = equal && getDisplayName().getValue().equals(other.getDisplayName().getValue());
//				}
//				return equal;
//			}
//		};
//	}

	private <T> boolean compareLists(List<T> expectedList, T actual) {
		List<T> actualList = new ArrayList<T>();
		actualList.add(actual);
		return compareCompleteLists(expectedList, actualList);
	}

	private <T> boolean compareCompleteLists(List<T> expectedList, List<T> actualList) {
		boolean found = false;

		if (expectedList == null || actualList == null) {
			return false;
		}
		if (expectedList.size() != actualList.size()) {
			return false;
		}
		for (T actualValue : actualList) {
			found = false;
			for (T expectedValue : expectedList) {
				if (actualValue.equals(expectedValue)) {
					found = true;
					break;
				}
			}
			if (!found) {
				return false;
			}
		}
		return found;
	}

	private <T> boolean comparePartialLists(QueryList<T> queryList, List<T> actualList) {
		List<List<T>> outerList = queryList.getOuterList();
		for (List<T> expectedList : outerList) {
			if (comparePartialLists(expectedList, actualList) == false) {
				return false;
			}
		}
		return true;
	}

	private <T> boolean comparePartialLists(List<T> expectedList, List<T> actualList) {
		if (expectedList == null || actualList == null) {
			return false;
		}
		if (expectedList.size() != actualList.size()) {
			return false;
		}
		for (T expectedValue : expectedList) {
			for (T actualValue : actualList) {
				if (actualValue.equals(expectedValue)) {
					return true;
				}
			}
		}
		return false;
	}

}
