/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentmetadataprovider;

import java.util.List;

import org.joda.time.DateTime;
import org.joda.time.Interval;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.AssigningAuthority;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.AvailabilityStatus;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.DocumentEntryType;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Identifiable;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Person;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.XcnName;
import org.openehealth.ipf.commons.ihe.xds.core.requests.query.FindDocumentsQuery;

import dk.nsi.dds.projects.sxa.common.configuration.SXAMappedValue;
import dk.nsi.dds.projects.sxa.common.util.QueryUtil;
import dk.nsi.dds.projects.sxa.documentmetadataprovider.configuration.MetadataFilterConfiguration;
import dk.nsi.dds.projects.sxa.documentmetadataprovider.impl.MetadataQueryFilter;

public class MetadataQueryFilterTest {

  private MetadataFilterConfiguration filter;
  private MetadataFilterConfiguration valuesForQuery;
  private boolean isApproved;
  private boolean onDemandDocument;

  @Before
  public void setup() {
    filter = createDefaultFilter();
    valuesForQuery = createDefaultFilter();
    isApproved = true;
    onDemandDocument = true;
  }

  @Test
  public void queryMatchesFilter() {
    valuesForQuery = createDefaultFilter();
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
  }

  @Test
  public void classCodeQueryAndFilter() {
    // Act / Assert
    valuesForQuery.getClassCodes().remove(0);
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
    filter.getClassCodes().remove(0);
    valuesForQuery = createDefaultFilter();
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
  }

  @Test
  public void confidentialityCodesQueryAndFilter() {
    valuesForQuery.getConfidentialityCodes().remove(0);
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
    filter.getConfidentialityCodes().remove(0);
    valuesForQuery = createDefaultFilter();
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
  }

  @Test
  public void eventCodesQueryAndFilter() {
    valuesForQuery.getEventCodes().remove(0);
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
    filter.getEventCodes().remove(0);
    valuesForQuery = createDefaultFilter();
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
  }

  @Test
  public void formatCodesQueryAndFilter() {
    valuesForQuery.getFormatCodes().remove(0);
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
    filter.getFormatCodes().remove(0);
    valuesForQuery = createDefaultFilter();
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
  }

  @Test
  public void typeCodesQueryAndFilter() {
    valuesForQuery.getTypeCodes().remove(0);
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
    filter.getTypeCodes().remove(0);
    valuesForQuery = createDefaultFilter();
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
  }
  
  @Test
  public void practiceSettingCodesQueryAndFilter() {
    valuesForQuery.getPracticeSettingCodes().remove(0);
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
    filter.getPracticeSettingCodes().remove(0);
    valuesForQuery = createDefaultFilter();
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
  }

  @Test
  public void invalidQueryTest() {
    // Arrange

    MetadataFilterConfiguration CodesHolder = createDefaultFilter();
    MetadataFilterConfiguration filter = createDefaultFilter();

    boolean isApproved = true;
    boolean onDemandDocument = true;
    Interval creationTime = new Interval(new DateTime(), new DateTime());
    Person authorPerson = new Person(new Identifiable("1234", new AssigningAuthority("1.2.3.4")), new XcnName("surname", "givenName", "middleName", "suffix", "prefix", "degree"));

    // Act / Assert
    Assert.assertFalse(generalTest(CodesHolder, filter, authorPerson, null, isApproved, onDemandDocument));
    Assert.assertFalse(generalTest(CodesHolder, filter, null, creationTime, isApproved, onDemandDocument));
    Assert.assertFalse(generalTest(CodesHolder, filter, null, null, isApproved, !onDemandDocument));

    //new Code($1, new LocalizedString($3, null, null), )
    removeAllAndAddValue(CodesHolder.getClassCodes(), new SXAMappedValue("asd", "asd", "asdf"));
    Assert.assertFalse(generalTest(CodesHolder, filter, null, null, isApproved, !onDemandDocument));

    CodesHolder = createDefaultFilter();
    removeAllAndAddValue(CodesHolder.getConfidentialityCodes(), new SXAMappedValue("asd", "asd", "asdf"));
    Assert.assertFalse(generalTest(CodesHolder, filter, null, null, isApproved, !onDemandDocument));

    CodesHolder = createDefaultFilter();
    removeAllAndAddValue(CodesHolder.getEventCodes(), new SXAMappedValue("asd", "asd", "asdf"));
    Assert.assertFalse(generalTest(CodesHolder, filter, null, null, isApproved, !onDemandDocument));

    CodesHolder = createDefaultFilter();
    removeAllAndAddValue(CodesHolder.getFormatCodes(), new SXAMappedValue("asd", "asd", "asdf"));
    Assert.assertFalse(generalTest(CodesHolder, filter, null, null, isApproved, !onDemandDocument));

    CodesHolder = createDefaultFilter();
    removeAllAndAddValue(CodesHolder.getTypeCodes(), new SXAMappedValue("asd", "asd", "asdf"));
    Assert.assertFalse(generalTest(CodesHolder, filter, null, null, isApproved, !onDemandDocument));
    
    CodesHolder = createDefaultFilter();
    removeAllAndAddValue(CodesHolder.getPracticeSettingCodes(), new SXAMappedValue("asd", "asd", "asdf"));
    Assert.assertFalse(generalTest(CodesHolder, filter, authorPerson, creationTime, isApproved, onDemandDocument));
  }

  private MetadataFilterConfiguration createDefaultFilter() {
    MetadataFilterConfiguration filter = new MetadataFilterConfiguration();
    filter.addClassCode(new SXAMappedValue("classCode", "classCodeSystem", "classCodeDesc"));
    filter.addTypeCode(new SXAMappedValue("typeCode", "typeCodeSystem", "typeCodeDesc"));
    filter.addEventCodes(new SXAMappedValue("eventCode", "eventCodeSystem", "eventCodeDesc"));
    filter.addConfidentialityCodes(new SXAMappedValue("confidentialityCode", "confidentialityCodeSystem", "confidentialityCodeDesc"));
    filter.addFormatCodes(new SXAMappedValue("formatCode", "formatCodeSystem", "formatCodeDesc"));
    filter.addPracticeSettingCodes(new SXAMappedValue("practiceSettingCode1", "practiceSettingCodeSystem", "practiceSettingCodeDesc"));
    filter.addPracticeSettingCodes(new SXAMappedValue("practiceSettingCode2", "practiceSettingCodeSystem", "practiceSettingCodeDesc"));
    filter.addPracticeSettingCodes(new SXAMappedValue("practiceSettingCode3", "practiceSettingCodeSystem", "practiceSettingCodeDesc"));
    return filter;
  }

  public boolean generalTest(MetadataFilterConfiguration queryCodesHolder, MetadataFilterConfiguration filter, 
    Person authorPerson, Interval creationTime, boolean isApproved, boolean onDemandDocument) {
    FindDocumentsQuery query = new FindDocumentsQuery();

    for (SXAMappedValue code : queryCodesHolder.getClassCodes()) {
      QueryUtil.addClassCode(query, code);
    }
    for (SXAMappedValue code : queryCodesHolder.getEventCodes()) {
      QueryUtil.addEventCode(query, code);
    }
    for (SXAMappedValue code : queryCodesHolder.getTypeCodes()) {
      QueryUtil.addTypeCode(query, code);
    }
    for (SXAMappedValue code : queryCodesHolder.getConfidentialityCodes()) {
      QueryUtil.addConfidentialityCode(query, code);
    }
    for (SXAMappedValue code : queryCodesHolder.getFormatCodes()) {
      QueryUtil.addFormatCode(query, code);
    }
    for (SXAMappedValue code : queryCodesHolder.getPracticeSettingCodes()) {
      QueryUtil.addPracticeSettingCode(query, code);
    }
    
    if (isApproved) {
    	QueryUtil.addStatus(query, AvailabilityStatus.APPROVED);
    } else {
    	QueryUtil.addStatus(query, AvailabilityStatus.DEPRECATED);
    }
    if (onDemandDocument) {
    	QueryUtil.addDocumentEntryType(query, DocumentEntryType.ON_DEMAND);
    } else {
    	QueryUtil.addDocumentEntryType(query, DocumentEntryType.STABLE);
    }

    
    QueryUtil.setCreationTime(query, creationTime);
    
    if (authorPerson != null) {
      QueryUtil.addAuthorPerson(query, authorPerson);
    }

    MetadataQueryFilter queryFilter = new MetadataQueryFilter(filter);

    return queryFilter.isValidQuery(query);
  }

  private void removeAllAndAddValue(List<SXAMappedValue> list, SXAMappedValue value) {
    list.clear();
    list.add(value);
  }
}
