/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (SDS), Denmark
 * (http://sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (SDS), Denmark (http://sundhedsdatastyrelsen.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentmetadataprovider.ws;

import java.lang.reflect.Field;
import java.util.Collections;

import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.runners.MockitoJUnitRunner;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.DocumentEntry;
import org.openehealth.ipf.commons.ihe.xds.core.responses.ErrorInfo;
import org.openehealth.ipf.commons.ihe.xds.core.responses.QueryResponse;
import org.openehealth.ipf.commons.ihe.xds.core.responses.Severity;
import org.openehealth.ipf.commons.ihe.xds.core.responses.Status;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.query.AdhocQueryRequest;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.query.AdhocQueryResponse;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.query.ResponseOptionType;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.rim.AdhocQueryType;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.rim.RegistryObjectListType;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.rs.RegistryError;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.rs.RegistryErrorList;
import org.openehealth.ipf.commons.ihe.xds.core.validate.ValidationMessage;
import org.openehealth.ipf.commons.ihe.xds.core.validate.XDSMetaDataException;

@RunWith(MockitoJUnitRunner.class)
public class MetadataProviderWSTest {
	
	@Mock
	public DocumentMetadataProvider documentMetadataProvider;
	@Mock
	public AdhocQueryRequest adhocQueryRequest;
	@Mock
	public AdhocQueryType adhocQueryType;
	@Mock
	public ResponseOptionType responseOptionType;
	
	@InjectMocks
	public MetadataProviderWS ws;
	
	@Before
	public void setup() {
		Mockito.when(adhocQueryRequest.getAdhocQuery()).thenReturn(adhocQueryType);
		Mockito.when(adhocQueryRequest.getResponseOption()).thenReturn(responseOptionType);
		
		Mockito.when(adhocQueryType.getId()).thenReturn("urn:uuid:14d4debf-8f97-4251-9a74-a90016b0af0d");
		
		//LeafClass eller ObjectRef
		Mockito.when(responseOptionType.getReturnType()).thenReturn("LeafClass");
	}
	
	@Test(expected=NullPointerException.class)
	public void testDocumentRegistryRegistryStoredQueryWithNull() {
		ws.documentRegistryRegistryStoredQuery(null);
	}
	
	@Test
	public void testDocumentRegistryRegistryStoredQueryWithDocumentMetadataProviderNotSet() throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException, SecurityException {
		final Field documentMetadataProviderField = MetadataProviderWS.class.getDeclaredField("documentMetadataProvider");
		documentMetadataProviderField.setAccessible(true);
		documentMetadataProviderField.set(ws, null);
		
		AdhocQueryResponse response = ws.documentRegistryRegistryStoredQuery(null);
		Assert.assertNotNull(response);
		
		Assert.assertEquals("urn:oasis:names:tc:ebxml-regrep:ResponseStatusType:Success", response.getStatus());
		Assert.assertEquals(0, getMetadataCount(response.getRegistryObjectList()));
		Assert.assertEquals(1, getErrorCount(response.getRegistryErrorList()));
		
		RegistryError error = response.getRegistryErrorList().getRegistryError().get(0);
		Assert.assertEquals("XDSRegistryError", error.getErrorCode());
		Assert.assertEquals("urn:oasis:names:tc:ebxml-regrep:ErrorSeverityType:Warning", error.getSeverity());
	}
	
	@Test
	public void testDocumentRegistryRegistryStoredQueryGetEmptyQueryResponseFromDocumentMetadataProvider() {
		Mockito.when(documentMetadataProvider.doQuery(Mockito.any())).thenReturn(new QueryResponse());
		
		AdhocQueryRequest request = adhocQueryRequest;
		AdhocQueryResponse response = ws.documentRegistryRegistryStoredQuery(request);
		Assert.assertNotNull(response);
		
		Assert.assertEquals("urn:oasis:names:tc:ebxml-regrep:ResponseStatusType:Success", response.getStatus());
		Assert.assertEquals(0, getMetadataCount(response.getRegistryObjectList()));
		Assert.assertEquals(0, getErrorCount(response.getRegistryErrorList()));
	}
	
	@Test
	public void testDocumentRegistryRegistryStoredQueryThrowExceptionFrommDocumentMetadataProvider() {
		XDSMetaDataException exception = new XDSMetaDataException(ValidationMessage.UNKNOWN_PATIENT_ID);
		Mockito.when(documentMetadataProvider.doQuery(Mockito.any())).thenThrow(exception);
		
		AdhocQueryRequest request = adhocQueryRequest;
		AdhocQueryResponse response = ws.documentRegistryRegistryStoredQuery(request);
		Assert.assertNotNull(response);
		
		Assert.assertEquals("urn:oasis:names:tc:ebxml-regrep:ResponseStatusType:Failure", response.getStatus());
		Assert.assertEquals(0, getMetadataCount(response.getRegistryObjectList()));
		Assert.assertEquals(1, getErrorCount(response.getRegistryErrorList()));
		
		RegistryError error = response.getRegistryErrorList().getRegistryError().get(0);
		Assert.assertEquals("Patient Id is unknown", error.getErrorCode());
		Assert.assertEquals("urn:oasis:names:tc:ebxml-regrep:ErrorSeverityType:Error", error.getSeverity());

	}
	
	@Test
	public void testDocumentRegistryRegistryStoredQueryGetEmptyQueryResponseFromDocumentMetadataProviderStatusFailure() {
		Mockito.when(documentMetadataProvider.doQuery(Mockito.any())).thenReturn(new QueryResponse(Status.FAILURE));
		
		AdhocQueryRequest request = adhocQueryRequest;
		AdhocQueryResponse response = ws.documentRegistryRegistryStoredQuery(request);
		Assert.assertNotNull(response);
		
		Assert.assertEquals("urn:oasis:names:tc:ebxml-regrep:ResponseStatusType:Failure", response.getStatus());
		Assert.assertEquals(0, getMetadataCount(response.getRegistryObjectList()));
		Assert.assertEquals(0, getErrorCount(response.getRegistryErrorList()));
	}
	
	@Test
	public void testDocumentRegistryRegistryStoredQueryGetQueryResponseWithDocumentFromDocumentMetadataProvider() {
		final QueryResponse queryReponse = new QueryResponse();
		queryReponse.setDocumentEntries(Collections.singletonList(new DocumentEntry()));
		Mockito.when(documentMetadataProvider.doQuery(Mockito.any())).thenReturn(queryReponse);
		
		AdhocQueryRequest request = adhocQueryRequest;
		AdhocQueryResponse response = ws.documentRegistryRegistryStoredQuery(request);
		Assert.assertNotNull(response);
		
		Assert.assertEquals("urn:oasis:names:tc:ebxml-regrep:ResponseStatusType:Success", response.getStatus());
		Assert.assertEquals(1, getMetadataCount(response.getRegistryObjectList()));
		Assert.assertEquals(0, getErrorCount(response.getRegistryErrorList()));
	}
	
	@Test
	public void testDocumentRegistryRegistryStoredQueryGetQueryResponseWithErrorFromDocumentMetadataProvider() {
		final QueryResponse queryReponse = new QueryResponse();
		final ErrorInfo ErrorInfo = new ErrorInfo();
		ErrorInfo.setSeverity(Severity.ERROR);
		queryReponse.setErrors(Collections.singletonList(ErrorInfo));
		Mockito.when(documentMetadataProvider.doQuery(Mockito.any())).thenReturn(queryReponse);
		
		AdhocQueryRequest request = adhocQueryRequest;
		AdhocQueryResponse response = ws.documentRegistryRegistryStoredQuery(request);
		Assert.assertNotNull(response);
		
		Assert.assertEquals("urn:oasis:names:tc:ebxml-regrep:ResponseStatusType:Failure", response.getStatus());
		Assert.assertEquals(0, getMetadataCount(response.getRegistryObjectList()));
		Assert.assertEquals(1, getErrorCount(response.getRegistryErrorList()));
	}
	
	@Test
	public void testDocumentRegistryRegistryStoredQueryGetQueryResponseWithErrorAndDocumentFromDocumentMetadataProvider() {
		final QueryResponse queryReponse = new QueryResponse();
		queryReponse.setDocumentEntries(Collections.singletonList(new DocumentEntry()));
		final ErrorInfo ErrorInfo = new ErrorInfo();
		ErrorInfo.setSeverity(Severity.ERROR);
		queryReponse.setErrors(Collections.singletonList(ErrorInfo));
		Mockito.when(documentMetadataProvider.doQuery(Mockito.any())).thenReturn(queryReponse);
		
		AdhocQueryRequest request = adhocQueryRequest;
		AdhocQueryResponse response = ws.documentRegistryRegistryStoredQuery(request);
		Assert.assertNotNull(response);
		
		Assert.assertEquals("urn:ihe:iti:2007:ResponseStatusType:PartialSuccess", response.getStatus());
		Assert.assertEquals(1, getMetadataCount(response.getRegistryObjectList()));
		Assert.assertEquals(1, getErrorCount(response.getRegistryErrorList()));
	}
	
	private static int getMetadataCount(final RegistryObjectListType registryObjectList) {
		if(registryObjectList == null ||  registryObjectList.getIdentifiable() == null) {
			return 0;
		}
		return registryObjectList.getIdentifiable().size();
	}

	private static int getErrorCount(final RegistryErrorList registryErrorList) {
		if(registryErrorList == null ||  registryErrorList.getRegistryError() == null) {
			return 0;
		}
		return registryErrorList.getRegistryError().size();
	}

}
