/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentprovider;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringReader;
import java.nio.charset.StandardCharsets;
import java.util.Base64;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.activation.DataSource;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.annotation.XmlAccessType;
import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlType;
import javax.xml.bind.annotation.XmlValue;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;
import javax.xml.ws.soap.SOAPFaultException;

import org.apache.commons.lang3.builder.ToStringBuilder;
import org.apache.commons.lang3.builder.ToStringStyle;
import org.joda.time.DateTime;
import org.junit.Assert;
import org.junit.BeforeClass;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.openehealth.ipf.commons.ihe.xds.core.responses.ErrorCode;
import org.openehealth.ipf.commons.ihe.xds.core.responses.ErrorInfo;
import org.openehealth.ipf.commons.ihe.xds.core.responses.RetrievedDocument;
import org.openehealth.ipf.commons.ihe.xds.core.responses.Severity;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.SpringRunner;
import org.springframework.test.context.support.AnnotationConfigContextLoader;

import com.cgi.svareksponering.DocumentWrapper;
import com.cgi.svareksponering.LeverandoerSvar;
import com.cgi.svareksponering.ResultatPakkePIH;
import com.cgi.svareksponering.ResultatStatus;

import dk.nsi.dds.projects.sxa.common.TestDataHelper;
import dk.nsi.dds.projects.sxa.common.configuration.IntegrationTestConfiguration;
import dk.nsi.dds.projects.sxa.common.testutil.AssertUtil;
import dk.nsi.dds.projects.sxa.common.util.ConverterUtil;
import dk.nsi.dds.projects.sxa.common.util.xmlhandling.XStreamFileException;
import dk.nsi.dds.projects.sxa.common.xds.impl.ValueEncoderDecoder;
import dk.nsi.dds.projects.sxa.documentprovider.impl.ConfigurationLoader;
import dk.nsi.dds.projects.sxa.documentprovider.impl.ProviderConfiguration;
import dk.nsi.dds.projects.sxa.documentprovider.impl.ProviderConfigurationHandler;
import dk.nsi.schema.privacy.pih._1.AttributeType;
import dk.nsi.schema.privacy.pih._1.DeclarationType;
import dk.nsi.schema.privacy.pih._1.PrivacyInformationHeaderType;
import dk.nsi.schema.privacy.piw._1.PrivacyInformationWrapperType;
import dk.nsi.util.Resource;

/**
 * 
 * These tests works for both CGI test service and provided stub
 * 
 * Note that the CGI test service returns the same data response despite choice of parameters
 *
 */
@RunWith(SpringRunner.class)
@ContextConfiguration(classes = { IntegrationTestConfiguration.class, LocalIntegrationTestConfiguration.class }, loader = AnnotationConfigContextLoader.class)
public class DocumentProviderWSIT {
	private JAXBContext jaxbContext = createJAXBContext();
	private TestDataHelper configTestData = new TestDataHelper();
	
	@Autowired
	private RequestPerformer requestPerformer;

	@BeforeClass
	public static void setupTestClientProperties() {
		// Denne opsætning bruges, hvis man vil afvikle testen af udviklingsvækrtøj og man ikke har kaldt 'mvn verify -P external-test'
		// Ved kald af 'mvn verify -P external-test' så kopieres testclient.properties fra integrationtest-modulet til classpath for denne test.
		if (System.getProperty(IntegrationTestConfiguration.clientPropertiesFileName) == null) {
			System.setProperty(IntegrationTestConfiguration.clientPropertiesFileName, new File("../../../integrationtest/src/test/resources/testclient.properties").getAbsolutePath());
		}
	}

	@Test
	public void requestEmptySpeciale() throws Exception {
		// No practiceSettingsCodes added
		// Arrange
		ResponseContext context = requestPerformer.patientId(TestDataHelper.DEFAULT_PATIENT_ID).from(new DateTime()).to(new DateTime().plusMinutes(1)).performRequest();
		// Assert
		Assert.assertNotNull(unmarshalPIW(context.getResult().get(0)));
	}

	@XmlAccessorType(XmlAccessType.FIELD)
	@XmlType(name = "", propOrder = {
	    "proeveTagningsTid",
	})
	@XmlRootElement(name = "Rekvisition")
	static class Rekvisition{
		@XmlElement(name = "ProeveTagningsTid")
		protected ProeveTagningsTid proeveTagningsTid;
		
		public ProeveTagningsTid getProeveTagningsTid() {
			return proeveTagningsTid;
		}
		
		public void setProeveTagningsTid(ProeveTagningsTid proeveTagningsTid) {
			this.proeveTagningsTid = proeveTagningsTid;
		}
		
		@Override
		public String toString() {
			return ToStringBuilder.reflectionToString(this, ToStringStyle.SHORT_PREFIX_STYLE);
		}
	}
	
	@XmlAccessorType(XmlAccessType.FIELD)
	@XmlType(name = "ProeveTagningsTid", propOrder = {
	    "value"
	})
	static class ProeveTagningsTid{
		@XmlValue
	    protected String value;
		
		public String getValue() {
			return value;
		}
		
		public void setValue(String value) {
			this.value = value;
		}
		
		@Override
		public String toString() {
			return ToStringBuilder.reflectionToString(this, ToStringStyle.SHORT_PREFIX_STYLE);
		}
	}
	
	@Test
	public void requestSuccessSingleSpeciale() throws Exception {
		// Arrange
		ResponseContext context = requestPerformer //
				.patientId(TestDataHelper.DEFAULT_PATIENT_ID) //
				.from(new DateTime()) //
				.to(new DateTime().plusMinutes(1)) //
				.addRequestPracticeSettingsCodes(ConverterUtil.convert(configTestData.practiceSettingCodeForMicrobiology)) //
				.expectOnePIWDocument() //
				.performRequest();
		// Assert
		List<RetrievedDocument> result = context.getResult();
		AssertUtil.assertEquals(1, result.size());
		
		
		final PrivacyInformationWrapperType unmarshalPIW = unmarshalPIW(context.getResult().get(0));
		Assert.assertNotNull(unmarshalPIW);
		AssertUtil.assertEquals(1, unmarshalPIW.getHeader().getDeclaration().size());
		DeclarationType declarationType = unmarshalPIW.getHeader().getDeclaration().get(0);
		List<AttributeType> attribute = declarationType.getSource().getAttribute();
		AssertUtil.assertEquals(3, attribute.size());
		for (AttributeType attributeType : attribute) {
			if("1.2.208.176.1.2".equals(attributeType.getCodeSystem())) {
				AssertUtil.assertEquals("patientId", attributeType.getName());
				AssertUtil.assertEquals("CPR", attributeType.getCodeSystemName());
				AssertUtil.assertEquals("1212121212", attributeType.getValue());
			} else if("2.16.840.1.113883.3.4208.100.5".equals(attributeType.getCodeSystem())) {
				AssertUtil.assertEquals("requestingOrganisationId", attributeType.getName());
				AssertUtil.assertEquals("SKS/SHAK", attributeType.getCodeSystemName());
				AssertUtil.assertEquals("4201050", attributeType.getValue());				
			} else if("1.0.8601".equals(attributeType.getCodeSystem())) {
				AssertUtil.assertEquals("timestamp", attributeType.getName());
				AssertUtil.assertEquals("", attributeType.getCodeSystemName());
				AssertUtil.assertEquals("2011-01-29T06:30:00.000+0000", attributeType.getValue());
			}
		}
		
		List<String> element = declarationType.getLocation().getElement();
		AssertUtil.assertEquals(1, element.size());
		AssertUtil.assertEquals("DocumentWrapper[@id='0']", element.get(0));
		
		String id = unmarshalPIW.getDocument().getId();
		AssertUtil.assertTrue(id.startsWith("1.2.3.4.55.66.12172240491026."));
		
	
		ResultatPakkePIH resultatPakke = unmarshalPIH(jaxbContext, ResultatPakkePIH.class, unmarshalPIW.getDocument().getValue());	
		
		Map<String, String> decodePIHMappedeSvar = decodePIHMappedeSvar(resultatPakke);	
		Map<String, String> decodePIHRaaSvar = decodePIHRaaSvar(resultatPakke);
		
		AssertUtil.assertEquals(1, decodePIHMappedeSvar.size());
		AssertUtil.assertTrue(decodePIHMappedeSvar.containsKey("0"));
		AssertUtil.assertEquals(1961653657, decodePIHMappedeSvar.get("0").hashCode());
		//Assert that code for Microbiology is in mappedeSvar.
		AssertUtil.assertContains("<ProeveTagningsTid>2011-01-29T06:30:00</ProeveTagningsTid>", decodePIHMappedeSvar.get("0"));
		AssertUtil.assertContains("<LaboratorieSpeciale>Mikrobiologi</LaboratorieSpeciale>", decodePIHMappedeSvar.get("0"));
		
		
		AssertUtil.assertEquals(1, decodePIHRaaSvar.size());
		AssertUtil.assertTrue(decodePIHRaaSvar.containsKey("0"));
		AssertUtil.assertEquals(150192645, decodePIHRaaSvar.get("0").hashCode());

		AssertUtil.assertContains("<Letter>\r\n\t\t<Identifier>80901082504866</Identifier>", decodePIHRaaSvar.get("0"));
		AssertUtil.assertContains("<Sender>\r\n\t\t<EANIdentifier>5790000120284</EANIdentifier>", decodePIHRaaSvar.get("0"));
		AssertUtil.assertContains("<Receiver>\r\n\t\t<EANIdentifier>5790000125012</EANIdentifier>", decodePIHRaaSvar.get("0"));
		AssertUtil.assertContains("<Patient>\r\n\t\t<CivilRegistrationNumber>0805721057</CivilRegistrationNumber>", decodePIHRaaSvar.get("0"));
		AssertUtil.assertContains("<LaboratoryInternalSampleIdentifier>Utst6SampleId</LaboratoryInternalSampleIdentifier>", decodePIHRaaSvar.get("0"));
		AssertUtil.assertContains("<LaboratoryInternalProductionIdentifier>Utst71ProdId</LaboratoryInternalProductionIdentifier>", decodePIHRaaSvar.get("0"));
		AssertUtil.assertContains("<Analysis>\r\n\t\t\t\t<AnalysisCode>NPU01807</AnalysisCode>", decodePIHRaaSvar.get("0"));
	}

	@Test
	public void requestSuccessTwoSpeciale() throws Exception {
		// Arrange
		ResponseContext context = requestPerformer //
				.patientId(TestDataHelper.DEFAULT_PATIENT_ID) //
				.from(new DateTime()) //
				.to(new DateTime().plusMinutes(1)) //
				.addRequestPracticeSettingsCodes(ConverterUtil.convert(configTestData.practiceSettingCodeForMicrobiology), ConverterUtil.convert(configTestData.practiceSettingCodeForBiochemistry)) //
				.expectOnePIWDocument() //
				.performRequest();
		// Assert
		List<RetrievedDocument> result = context.getResult();
		AssertUtil.assertEquals(1, result.size());
		
		final PrivacyInformationWrapperType unmarshalPIW = unmarshalPIW(context.getResult().get(0));
		Assert.assertNotNull(unmarshalPIW);
		
		AssertUtil.assertEquals(2, unmarshalPIW.getHeader().getDeclaration().size());
		System.out.println(unmarshalPIW.getMimeType());
		System.out.println(unmarshalPIW.getDocument().getId());
		
		/*First Header*/
		DeclarationType declarationType1 = unmarshalPIW.getHeader().getDeclaration().get(0);
		List<AttributeType> attributeList1 = declarationType1.getSource().getAttribute();
		AssertUtil.assertEquals(3, attributeList1.size());
		for (AttributeType attributeType : attributeList1) {
			if("1.2.208.176.1.2".equals(attributeType.getCodeSystem())) {
				AssertUtil.assertEquals("patientId", attributeType.getName());
				AssertUtil.assertEquals("CPR", attributeType.getCodeSystemName());
				AssertUtil.assertEquals("1212121212", attributeType.getValue());
			} else if("2.16.840.1.113883.3.4208.100.5".equals(attributeType.getCodeSystem())) {
				AssertUtil.assertEquals("requestingOrganisationId", attributeType.getName());
				AssertUtil.assertEquals("SKS/SHAK", attributeType.getCodeSystemName());
				AssertUtil.assertEquals("6008319", attributeType.getValue());				
			} else if("1.0.8601".equals(attributeType.getCodeSystem())) {
				AssertUtil.assertEquals("timestamp", attributeType.getName());
				AssertUtil.assertEquals("", attributeType.getCodeSystemName());
				AssertUtil.assertEquals("2014-05-12T11:38:00.000+0000", attributeType.getValue());
			}
		}
		
		List<String> location1 = declarationType1.getLocation().getElement();
		AssertUtil.assertEquals(1, location1.size());
		AssertUtil.assertEquals("DocumentWrapper[@id='0']", location1.get(0));
		
		/*Second Header*/
		DeclarationType declarationType2 = unmarshalPIW.getHeader().getDeclaration().get(1);
		List<AttributeType> attributeList2 = declarationType2.getSource().getAttribute();
		AssertUtil.assertEquals(3, attributeList2.size());
		for (AttributeType attribute : attributeList2) {
			if("1.2.208.176.1.2".equals(attribute.getCodeSystem())) {
				AssertUtil.assertEquals("patientId", attribute.getName());
				AssertUtil.assertEquals("CPR", attribute.getCodeSystemName());
				AssertUtil.assertEquals("1212121212", attribute.getValue());
			} else if("2.16.840.1.113883.3.4208.100.5".equals(attribute.getCodeSystem())) {
				AssertUtil.assertEquals("requestingOrganisationId", attribute.getName());
				AssertUtil.assertEquals("SKS/SHAK", attribute.getCodeSystemName());
				AssertUtil.assertEquals("4201050", attribute.getValue());				
			} else if("1.0.8601".equals(attribute.getCodeSystem())) {
				AssertUtil.assertEquals("timestamp", attribute.getName());
				AssertUtil.assertEquals("", attribute.getCodeSystemName());
				AssertUtil.assertEquals("2011-01-29T06:30:00.000+0000", attribute.getValue());
			}
		}
		
		List<String> location2 = declarationType2.getLocation().getElement();
		AssertUtil.assertEquals(1, location2.size());
		AssertUtil.assertEquals("DocumentWrapper[@id='1']", location2.get(0));
		
		/*Document*/	
		String id = unmarshalPIW.getDocument().getId();
		AssertUtil.assertTrue(id.startsWith("1.2.3.4.55.66.12172240491026."));
		
		ResultatPakkePIH resultatPakke = unmarshalPIH(jaxbContext, ResultatPakkePIH.class, unmarshalPIW.getDocument().getValue());	
		
		Map<String, String> decodePIHMappedeSvar = decodePIHMappedeSvar(resultatPakke);	
		Map<String, String> decodePIHRaaSvar = decodePIHRaaSvar(resultatPakke);
		
		AssertUtil.assertEquals(2, decodePIHMappedeSvar.size());
		AssertUtil.assertEquals(2, decodePIHRaaSvar.size());

		//Firts
		String key = "0";
		//MappedeSvar
		AssertUtil.assertTrue(decodePIHMappedeSvar.containsKey(key));
		final String mappedeSvar1 = decodePIHMappedeSvar.get(key);
		AssertUtil.assertEquals(1088579494, mappedeSvar1.hashCode());
		//Assert that code for Microbiology is in mappedeSvar.
		AssertUtil.assertContains("<ProeveTagningsTid>2014-05-12T11:38:00</ProeveTagningsTid>", mappedeSvar1);
		AssertUtil.assertContains("<LaboratorieSpeciale>KliniskBiokemi</LaboratorieSpeciale>", mappedeSvar1);
		AssertUtil.assertAntal(29, "<AnalyseVaerdi>", mappedeSvar1);
		//RaaSvar
		AssertUtil.assertTrue(decodePIHRaaSvar.containsKey(key));
		final String raaSvar1 = decodePIHRaaSvar.get(key);
		AssertUtil.assertEquals(721201669, raaSvar1.hashCode());
		//Assert values in RaaSvar
		AssertUtil.assertContains("<Letter>\r\n\t\t<Identifier>80901082504866</Identifier>", raaSvar1);
		AssertUtil.assertContains("<Sender>\r\n\t\t<EANIdentifier>5790000120284</EANIdentifier>", raaSvar1);
		AssertUtil.assertContains("<Receiver>\r\n\t\t<EANIdentifier>5790000125012</EANIdentifier>", raaSvar1);
		AssertUtil.assertContains("<Patient>\r\n\t\t<CivilRegistrationNumber>0805721057</CivilRegistrationNumber>", raaSvar1);
		AssertUtil.assertContains("<LaboratoryInternalSampleIdentifier>Utst6SampleId</LaboratoryInternalSampleIdentifier>", raaSvar1);
		AssertUtil.assertContains("<LaboratoryInternalProductionIdentifier>Utst71ProdId</LaboratoryInternalProductionIdentifier>", raaSvar1);
		AssertUtil.assertContains("<Analysis>\r\n\t\t\t\t<AnalysisCode>NPU01807</AnalysisCode>", raaSvar1);
		
		//Second
		key = "1";
		//MappedeSvar
		AssertUtil.assertTrue(decodePIHMappedeSvar.containsKey(key));
		final String mappedeSvar2 = decodePIHMappedeSvar.get(key);
		AssertUtil.assertEquals(1961653657, mappedeSvar2.hashCode());
		//Assert that code for Microbiology is in mappedeSvar.
		AssertUtil.assertContains("<ProeveTagningsTid>2011-01-29T06:30:00</ProeveTagningsTid>", mappedeSvar2);
		AssertUtil.assertContains("<LaboratorieSpeciale>Mikrobiologi</LaboratorieSpeciale>", mappedeSvar2);
		//RaaSvar
		AssertUtil.assertTrue(decodePIHRaaSvar.containsKey(key));
		final String raaSvar2 = decodePIHRaaSvar.get(key);
		AssertUtil.assertEquals(-1696860527, raaSvar2.hashCode());
		//Assert values in RaaSvar
		AssertUtil.assertContains("<Letter>\r\n\t\t<Identifier>80901082504866</Identifier>", raaSvar2);
		AssertUtil.assertContains("<Sender>\r\n\t\t<EANIdentifier>5790000120284</EANIdentifier>", raaSvar2);
		AssertUtil.assertContains("<Receiver>\r\n\t\t<EANIdentifier>5790000125012</EANIdentifier>", raaSvar2);
		AssertUtil.assertContains("<Patient>\r\n\t\t<CivilRegistrationNumber>0805721057</CivilRegistrationNumber>", raaSvar2);
		AssertUtil.assertContains("<LaboratoryInternalSampleIdentifier>Utst6SampleId</LaboratoryInternalSampleIdentifier>", raaSvar2);
		AssertUtil.assertContains("<LaboratoryInternalProductionIdentifier>Utst71ProdId</LaboratoryInternalProductionIdentifier>", raaSvar2);
		AssertUtil.assertContains("<Analysis>\r\n\t\t\t\t<AnalysisCode>NPU01807</AnalysisCode>", raaSvar2);
		
	}

	@Test
	public void requestFromTimeIsNull() throws Exception {
		// Arrange
		ResponseContext context = requestPerformer //
				.patientId(TestDataHelper.DEFAULT_PATIENT_ID) //
				.from(null) //
				.to(new DateTime().plusMinutes(1)) //
				.addRequestPracticeSettingsCodes(ConverterUtil.convert(configTestData.practiceSettingCodeForMicrobiology)) //
				.performRequest();
		// Assert
		Assert.assertNotNull(context.getResult());
	}

	@Test
	public void requestToTimeIsNull() throws Exception {
		// Arrange
		ResponseContext context = requestPerformer //
				.patientId(TestDataHelper.DEFAULT_PATIENT_ID) //
				.from(new DateTime()) //
				.to(null) //
				.addRequestPracticeSettingsCodes(ConverterUtil.convert(configTestData.practiceSettingCodeForMicrobiology)) //
				.performRequest();
		// Assert
		Assert.assertNotNull(context.getResult());
	}

	@Test
	public void requestFromAndToTimesAreNull() throws Exception {
		// Arrange
		ResponseContext context = requestPerformer //
				.patientId(TestDataHelper.DEFAULT_PATIENT_ID) //
				.from(null) //
				.to(null) //
				.addRequestPracticeSettingsCodes(ConverterUtil.convert(configTestData.practiceSettingCodeForMicrobiology)) //
				.performRequest();
		// Assert
		Assert.assertNotNull(context.getResult());
	}

	@Test
	public void requestFromTimeAfterToTime() throws Exception {
		// Arrange
		ResponseContext context = requestPerformer //
				.patientId(TestDataHelper.DEFAULT_PATIENT_ID) //
				.from(new DateTime().plusMinutes(1)) //
				.to(new DateTime()) //
				.addRequestPracticeSettingsCodes(ConverterUtil.convert(configTestData.practiceSettingCodeForMicrobiology)) //
				.performRequest();
		// Assert
		expectRepositoryError(context);
	}

	@Test(expected = SOAPFaultException.class)
	public void requestInvalidPatientId() throws Exception {
		// Arrange
		String invalidPatientIdWithTooManyDigits = "25124849161111111";
		ResponseContext context = requestPerformer //
				.patientId(invalidPatientIdWithTooManyDigits) //
				.from(new DateTime()) //
				.to(new DateTime().plusMinutes(1)) //
				.addRequestPracticeSettingsCodes(ConverterUtil.convert(configTestData.practiceSettingCodeForMicrobiology)) //
				.performRequest();
	}

	@Test
	public void requestWrongHomeCommunityId() throws Exception {
		// Arrange
		// Override home community for the request
		ProviderConfiguration providerConfiguration = loadConfiguration();
		String wrongHomeCommunityId = providerConfiguration.getHomeCommunityId() + ".3.1";
		providerConfiguration.setHomeCommunityId(wrongHomeCommunityId);
		// Act
		ResponseContext context = requestPerformer.patientId(TestDataHelper.DEFAULT_PATIENT_ID).from(null).to(null).providerConfiguration(providerConfiguration).performRequest();
		// Assert
		expectDocumentSharingErrorWithWrongCommunityId(wrongHomeCommunityId, context);
	}

	private void expectRepositoryError(ResponseContext context) {
		List<RetrievedDocument> result = context.getResult();
		Assert.assertNotNull(result);
		ErrorInfo error = context.getErrors().get(0);
		Assert.assertNotNull(error);
		Assert.assertNotNull(context.getGeneratedDocumentId());
		Assert.assertEquals(context.getGeneratedDocumentId().toString(), ValueEncoderDecoder.extractDocumentUniqueId(error.getCodeContext()));
		Assert.assertEquals(context.getRepositoryId(), ValueEncoderDecoder.extractRepositoryUniqueId(error.getCodeContext()));
		Assert.assertTrue(error.getSeverity() == Severity.ERROR);
		Assert.assertTrue(error.getErrorCode() != null);
		Assert.assertEquals(ErrorCode.REPOSITORY_ERROR, error.getErrorCode());
	}

	private void expectDocumentSharingErrorWithWrongCommunityId(String expectedHomeCommunityId, ResponseContext context) {
		List<RetrievedDocument> result = context.getResult();
		Assert.assertNotNull(result);
		Assert.assertEquals(0, result.size());

		ErrorInfo error = null;
		for (ErrorInfo err : context.getErrors()) {
			if(err.getCodeContext() != null) {
				error = err;
				break;
			}
		}
		
		Assert.assertNotNull(error);

		Assert.assertEquals(expectedHomeCommunityId, ValueEncoderDecoder.extractHomeCommunityId(error.getCodeContext()));
		Assert.assertEquals(context.getGeneratedDocumentId().toString(), ValueEncoderDecoder.extractDocumentUniqueId(error.getCodeContext()));
		Assert.assertEquals(context.getRepositoryId(), ValueEncoderDecoder.extractRepositoryUniqueId(error.getCodeContext()));
		Assert.assertTrue(error.getSeverity() == Severity.ERROR);
		Assert.assertTrue(error.getErrorCode() != null);
		Assert.assertEquals(ErrorCode.UNKNOWN_COMMUNITY, error.getErrorCode());
	}

	private ProviderConfiguration loadConfiguration() throws XStreamFileException, FileNotFoundException {
		File confFile = Resource.getFile(this.getClass(), ConfigurationLoader.PROVIDER_CONFIGURATION_FILE_NAME);
		ProviderConfigurationHandler handler = new ProviderConfigurationHandler(confFile);
		return handler.readConfiguration();
	}

	private PrivacyInformationWrapperType unmarshalPIW(RetrievedDocument retrievedDocument) throws JAXBException, IOException {
		final DataSource dataSource = retrievedDocument.getDataHandler().getDataSource();
		try(InputStream is = dataSource.getInputStream()){
			return unmarshalPIW(is);
		}
	}

	private PrivacyInformationWrapperType unmarshalPIW(InputStream is) throws JAXBException {
		StreamSource source = new StreamSource(new InputStreamReader(is, StandardCharsets.UTF_8));
		JAXBElement<PrivacyInformationWrapperType> piwElement = jaxbContext.createUnmarshaller().unmarshal(source, PrivacyInformationWrapperType.class);
		return piwElement.getValue();
	}

	private JAXBContext createJAXBContext() {
		try {
			return JAXBContext.newInstance(PrivacyInformationWrapperType.class, PrivacyInformationHeaderType.class, ResultatPakkePIH.class);
		} catch (JAXBException e) {
			throw new RuntimeException(e);
		}
	}
	
	public Map<String, String> decodePIHMappedeSvar(ResultatPakkePIH resultatPakke) {
		if(resultatPakke == null || resultatPakke.getMappedeSvar() == null || resultatPakke.getMappedeSvar().getDocumentWrapper() == null) {
			return null;
		}
		List<DocumentWrapper> documentWrapperList = resultatPakke.getMappedeSvar().getDocumentWrapper();
		Map<String, String> result = new HashMap<String, String>();
		for (DocumentWrapper documentWrapper : documentWrapperList) {
			String id = documentWrapper.getId();
			String decoded = decodePIH(documentWrapper.getValue());
			if(result.containsKey(id)) {
				throw new RuntimeException("Flere dokumenter på sammen id.");
			}
			result.put(id, decoded);
		}  
		return result;
	}
	
	public Map<String, String> decodePIHRaaSvar(ResultatPakkePIH resultatPakke) {
		if(resultatPakke == null || resultatPakke.getRaaSvar() == null || resultatPakke.getRaaSvar().getLeverandoerSvar() == null) {
			return null;
		}
		List<LeverandoerSvar> leverandoerSvarList = resultatPakke.getRaaSvar().getLeverandoerSvar();
		Map<String, String> result = new HashMap<String, String>();
		for (LeverandoerSvar leverandoerSvar : leverandoerSvarList) {
			@SuppressWarnings("unused")
			String leverandoerNavn = leverandoerSvar.getLeverandoerNavn();
			@SuppressWarnings("unused")
			String namespace = leverandoerSvar.getNamespace();
			@SuppressWarnings("unused")
			ResultatStatus status = leverandoerSvar.getStatus();
			List<DocumentWrapper> documentWrapperList = leverandoerSvar.getLeverandoerSvarXml().getDocumentWrapper();			
			for (DocumentWrapper documentWrapper : documentWrapperList) {
				String id = documentWrapper.getId();
				String decoded = decodePIH(documentWrapper.getValue());
				if(result.containsKey(id)) {
					throw new RuntimeException("Flere dokumenter på sammen id.");
				}
				result.put(id, decoded);
			}  			
		} 
		return result;
	}
	
	public String decodePIH(String endocedStr) {
		byte[] decodedStr = Base64.getDecoder().decode(endocedStr);
		return new String(decodedStr, StandardCharsets.UTF_8);
	}
	public static <E> E unmarshalPIH(JAXBContext ctx, Class<E> cl, String str) throws JAXBException {
		Reader r = new StringReader(str);
		Source src = new StreamSource(r);
		return ctx.createUnmarshaller().unmarshal(src, cl).getValue();
	}
	
	public static <E> E unmarshalPIH(JAXBContext ctx, Class<E> cl, byte[] data) throws JAXBException {
		return unmarshalPIH(ctx, cl, new String(data, StandardCharsets.UTF_8));
	}
}
