/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentprovider;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;

import dk.nsi.dds.projects.sxa.common.AddTestSoapHeadersInterceptor;
import org.apache.cxf.endpoint.Client;
import org.apache.cxf.frontend.ClientProxy;
import org.joda.time.DateTime;
import org.junit.Assert;
import org.openehealth.ipf.commons.ihe.xds.core.ebxml.ebxml30.RetrieveDocumentSetRequestType;
import org.openehealth.ipf.commons.ihe.xds.core.ebxml.ebxml30.RetrieveDocumentSetResponseType;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.AssigningAuthority;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Code;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Identifiable;
import org.openehealth.ipf.commons.ihe.xds.core.requests.RetrieveDocument;
import org.openehealth.ipf.commons.ihe.xds.core.requests.RetrieveDocumentSet;
import org.openehealth.ipf.commons.ihe.xds.core.responses.RetrievedDocument;
import org.openehealth.ipf.commons.ihe.xds.core.responses.RetrievedDocumentSet;
import org.openehealth.ipf.commons.ihe.xds.iti43.Iti43PortType;
import org.openehealth.ipf.platform.camel.ihe.xds.core.converters.EbXML30Converters;
import org.springframework.beans.factory.annotation.Autowired;

import com.cgi.svareksponering.ResultatPakkePIH;

import dk.nsi.dds.projects.sxa.common.TestDataHelper;
import dk.nsi.dds.projects.sxa.common.configuration.SXAMappedValue;
import dk.nsi.dds.projects.sxa.common.util.DocumentUniqueIdUtil;
import dk.nsi.dds.projects.sxa.documentprovider.impl.ConfigurationLoader;
import dk.nsi.dds.projects.sxa.documentprovider.impl.ProviderConfiguration;
import dk.nsi.dds.projects.sxa.documentprovider.impl.ProviderConfigurationHandler;
import dk.nsi.schema.privacy.pih._1.PrivacyInformationHeaderType;
import dk.nsi.schema.privacy.piw._1.PrivacyInformationWrapperType;
import dk.nsi.util.Resource;

public class RequestPerformer {
	
	@Autowired
	private Iti43PortType iti43PortType;
	
	private List<Code> requestPracticeSettingCodes;
	private DateTime fromTime;
	private DateTime toTime;
	private String patientId;
	private String uniqueRootId;
	private TestDataHelper configTestData;
	private ProviderConfiguration providerConfiguration;
	private boolean validateResponse;

	public RequestPerformer() {
		this.requestPracticeSettingCodes = new ArrayList<Code>();
		this.validateResponse = false;
		this.uniqueRootId = "1.2.3.4.55.66";
		this.configTestData = new TestDataHelper();
		this.providerConfiguration = loadConfiguration();
	}

	public RequestPerformer addRequestPracticeSettingsCodes(Code... requestPracticeSettingsCodes) {
		Arrays.stream(requestPracticeSettingsCodes).forEach(code -> this.requestPracticeSettingCodes.add(code));
		return this;
	}

	public RequestPerformer providerConfiguration(ProviderConfiguration providerConfiguration) {
		this.providerConfiguration = providerConfiguration;
		return this;
	}

	public RequestPerformer from(DateTime fromTime) {
		this.fromTime = fromTime;
		return this;
	}

	public RequestPerformer to(DateTime toTime) {
		this.toTime = toTime;
		return this;
	}

	public RequestPerformer patientId(String patientId) {
		if(this.patientId != null) {
			throw new RuntimeException("Old Patienid="+this.patientId+" new patientid="+patientId);
		}
		this.patientId = patientId;
		return this;
	}

	public RequestPerformer rootId(String uniqueRootId) {
		this.uniqueRootId = uniqueRootId;
		return this;
	}

	public ResponseContext performRequest() throws Exception {
		List<SXAMappedValue> validPracticeSettingCodes = configTestData.getPracticeSettingCodes();
		Identifiable patient = new Identifiable(patientId, new AssigningAuthority("1.208.3.4"));
		String docid = DocumentUniqueIdUtil.createUniqueId(uniqueRootId, patient, fromTime, toTime, requestPracticeSettingCodes, validPracticeSettingCodes);

		List<RetrieveDocument> request = createDefaultRequest(providerConfiguration, patient, docid);
		
		RetrievedDocumentSet result =  validateResponse ? callWebserviceAndValidateResponse(request) : callWebservice(request);

		
		ResponseContext ctx = new ResponseContext();
		ctx.setGeneratedDocumentId(docid);
		ctx.setRepositoryId("1.2.3.4.55.66");
		ctx.setResult(result.getDocuments());
		ctx.setErrors(result.getErrors());
		ctx.setStatus(result.getStatus().getOpcode21());
		return ctx;
	}

	private ProviderConfiguration loadConfiguration() {
		try {
			File confFile = Resource.getFile(this.getClass(), ConfigurationLoader.PROVIDER_CONFIGURATION_FILE_NAME);
			ProviderConfigurationHandler handler = new ProviderConfigurationHandler(confFile);
			return handler.readConfiguration();
		} catch (Exception ex) {
			throw new RuntimeException(ex);
		}
	}

	public List<RetrieveDocument> createDefaultRequest(ProviderConfiguration providerConfiguration, Identifiable patientId, String uniqueDocumentId) {
		List<RetrieveDocument> toRetrieve = new ArrayList<>();
		toRetrieve.add(createMetadataDocument(providerConfiguration.getRepositoryUniqueId(), providerConfiguration.getHomeCommunityId(), uniqueDocumentId));
		return toRetrieve;
	}

	private RetrieveDocument createMetadataDocument(String repositoryUniqueId, String homeCommunityId, String documentUniqueId) {
		RetrieveDocument toRetrieve = new RetrieveDocument();
		toRetrieve.setDocumentUniqueId(documentUniqueId);
		toRetrieve.setHomeCommunityId(homeCommunityId);
		toRetrieve.setRepositoryUniqueId(repositoryUniqueId);
		return toRetrieve;
	}

	private RetrievedDocumentSet callWebservice(List<RetrieveDocument> toRetrieve) {
		RetrieveDocumentSet documentSet = new RetrieveDocumentSet();
		documentSet.getDocuments().addAll(toRetrieve);
		RetrieveDocumentSetRequestType documentSetRequestType = EbXML30Converters.convert(documentSet);

		Client proxy = ClientProxy.getClient(iti43PortType);
		proxy.getOutInterceptors().add(new AddTestSoapHeadersInterceptor());
		
		RetrieveDocumentSetResponseType documentRepositoryRetrieveDocumentSet = iti43PortType.documentRepositoryRetrieveDocumentSet(documentSetRequestType);
		RetrievedDocumentSet retrievedDocumentSet = EbXML30Converters.convert(documentRepositoryRetrieveDocumentSet);
		return retrievedDocumentSet;
	}

	private RetrievedDocumentSet callWebserviceAndValidateResponse(List<RetrieveDocument> request) throws Exception {
		RetrievedDocumentSet result = callWebservice(request);
		Assert.assertNotNull(result);
		Assert.assertNotNull(result.getDocuments());
		Assert.assertEquals(1, result.getDocuments().size());
		Assert.assertNotNull(result.getErrors());
		Assert.assertTrue(result.getErrors().isEmpty());
		RetrievedDocument retrievedDocument = result.getDocuments().get(0);	
		Assert.assertNotNull(retrievedDocument);
		Assert.assertEquals("text/xml", retrievedDocument.getMimeType());
		return result;
	}
	
	private JAXBContext createJAXBContext() {
		try {
			return JAXBContext.newInstance(PrivacyInformationWrapperType.class, PrivacyInformationHeaderType.class, ResultatPakkePIH.class);
		} catch (JAXBException e) {
			throw new RuntimeException(e);
		}
	}

	public RequestPerformer expectOnePIWDocument() {
		this.validateResponse = true;
		return this;
	}
}
