package dk.nsi.dds.projects.sxa.documentprovider.audit;

import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.nsi.hsuid.ConsentOverrideAttribute;
import dk.nsi.hsuid._2016._08.hsuid_1_1.Attribute;
import dk.nsi.hsuid._2016._08.hsuid_1_1.HsuidHeader;
import dk.nsi.hsuid._2016._08.hsuid_1_1.NameConstraint;
import dk.nsi.hsuid._2016._08.hsuid_1_1.SubjectIdentifierType;
import loginfo.healthcare._2017_03_01.*;
import org.openehealth.ipf.commons.ihe.xds.core.ebxml.ebxml30.RetrieveDocumentSetRequestType;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;
import java.math.BigInteger;
import java.util.UUID;
import java.util.stream.Collectors;

public class LogRecordBuilder {
    private DateTimeProvider dateTimeProvider;

    public LogRecordBuilder(DateTimeProvider dateTimeProvider) {
        this.dateTimeProvider = dateTimeProvider;
    }

    public LogRecord buildLogRecord(HsuidHeader hsuidHeader, Header header, LogRecord inboundLogRecord, RetrieveDocumentSetRequestType request) {
        LogRecord logRecord = new LogRecord();

        logRecord.setTraceInfo(buildTraceInfo(header, inboundLogRecord));
        logRecord.setRequestingClient(buildRequestingClientType(hsuidHeader));
        logRecord.setPatient(buildPatient(request));
        logRecord.setConsent(buildConsent(hsuidHeader));
        logRecord.setExtendedRequestInfo(buildExtendedRequestInfo(request));

        return logRecord;
    }

    private TraceInfoType buildTraceInfo(Header header, LogRecord inboundLogRecord) {
        TraceInfoType traceInfo = new TraceInfoType();

        traceInfo.setRequestTime(getRequestTime());
        traceInfo.setSessionId(getSessionId(header));
        traceInfo.setRequestId(getRequestId());
        traceInfo.setRoute(buildRoute(inboundLogRecord));

        return traceInfo;
    }

    private XMLGregorianCalendar getRequestTime() {
        try {
            return DatatypeFactory.newInstance().newXMLGregorianCalendar(dateTimeProvider.now().toGregorianCalendar());
        }
        catch(DatatypeConfigurationException e) {
            throw new IllegalStateException("Could not instantiate DatatypeFactory!", e);
        }
    }

    private String getSessionId(Header header) {
        if(header.getLinking() != null && header.getLinking().getFlowID() != null) {
            return header.getLinking().getFlowID();
        }
        return UUID.randomUUID().toString();
    }

    private String getRequestId() {
        return UUID.randomUUID().toString();
    }

    private RouteType buildRoute(LogRecord inboundLogRecord) {
        RouteType route = new RouteType();

        if(inboundLogRecord != null && inboundLogRecord.getTraceInfo() != null) {
            TraceInfoType traceInfo = inboundLogRecord.getTraceInfo();
            if(traceInfo.getRoute() != null) {
                RouteType forwardedRoute = traceInfo.getRoute();
                route.getViaSystem().addAll(forwardedRoute.getViaSystem());
            }
        }

        RouteType.ViaSystem viaSystem = new RouteType.ViaSystem();
        viaSystem.setId(route.getViaSystem().size() + 1);
        viaSystem.setValue("Dokumentdelingsservice");
        route.getViaSystem().add(viaSystem);

        return route;
    }

    private RequestingClientType buildRequestingClientType(HsuidHeader hsuidHeader) {
        RequestingClientType requestingClientType = new RequestingClientType();

        requestingClientType.setClientIdentification(buildClientIdentification(hsuidHeader));
        requestingClientType.setClinicName("");
        requestingClientType.setClinicDepartment("");
        requestingClientType.setUserCprNumber(getActingUserCivilRegistrationNumber(hsuidHeader));
        requestingClientType.setUserName("");
        requestingClientType.setUserInitials("");
        requestingClientType.setBehalfOfCprNumber(getResponsibleUserCivilRegistrationNumber(hsuidHeader));
        requestingClientType.setLookUpType(getLookupType(hsuidHeader));
        requestingClientType.setTaskType("Opslag Dokumentdelingsservice");

        return requestingClientType;
    }

    private RequestingClientType.ClientIdentification buildClientIdentification(HsuidHeader hsuidHeader) {
        Attribute userTypeAttribute = getRequiredAttribute(hsuidHeader, NameConstraint.NSI_USERTYPE);
        switch(userTypeAttribute.getAttributeValue()) {
            case "nsi:Citizen":
                return buildCitizenClientIdentification(hsuidHeader);
            case "nsi:HealthcareProfessional":
                return buildHealthcareProfessionalClientIdentification(hsuidHeader);
            default:
                throw new IllegalArgumentException(String.format("Unexpected nsi:UserType: %s", userTypeAttribute.getAttributeValue()));
        }
    }

    private RequestingClientType.ClientIdentification buildCitizenClientIdentification(HsuidHeader hsuidHeader) {
        RequestingClientType.ClientIdentification clientIdentification = new RequestingClientType.ClientIdentification();

        clientIdentification.setClientIdentificationType(ClientIdentificationTypeEnumeration.CPR_NUMMER);
        clientIdentification.setValue(getActingUserCivilRegistrationNumber(hsuidHeader));

        return clientIdentification;
    }

    private RequestingClientType.ClientIdentification buildHealthcareProfessionalClientIdentification(HsuidHeader hsuidHeader) {
        RequestingClientType.ClientIdentification clientIdentification = new RequestingClientType.ClientIdentification();

        Attribute orgUsingIdAttribute = getRequiredAttribute(hsuidHeader, NameConstraint.NSI_ORG_USING_ID);
        clientIdentification.setClientIdentificationType(getClientIdentificationTypeForOrganisation(orgUsingIdAttribute.getNameFormat()));
        clientIdentification.setValue(orgUsingIdAttribute.getAttributeValue());

        return clientIdentification;
    }

    private ClientIdentificationTypeEnumeration getClientIdentificationTypeForOrganisation(SubjectIdentifierType subjectIdentifier) {
        switch(subjectIdentifier) {
            case NSI_SKSCODE:
                return ClientIdentificationTypeEnumeration.SKS;
            case NSI_SORCODE:
                return ClientIdentificationTypeEnumeration.SOR;
            case NSI_YNUMBER:
                return ClientIdentificationTypeEnumeration.YDERNUMMER;
            default:
                throw new IllegalArgumentException(String.format("Unexpected SubjectIdentifierType: %s", subjectIdentifier));
        }
    }

    private String getActingUserCivilRegistrationNumber(HsuidHeader hsuidHeader) {
        Attribute actingUserCivilRegistrationNumberAttribute = getRequiredAttribute(hsuidHeader, NameConstraint.NSI_ACTING_USER_CIVIL_REGISTRATION_NUMBER);
        return normalizeCpr(actingUserCivilRegistrationNumberAttribute.getAttributeValue());
    }

    private String getResponsibleUserCivilRegistrationNumber(HsuidHeader hsuidHeader) {
        Attribute responsibleUserCivilRegistrationNumberAttribute = getOptionalAttribute(hsuidHeader, NameConstraint.NSI_RESPONSIBLE_USER_CIVIL_REGISTRATION_NUMBER);
        return responsibleUserCivilRegistrationNumberAttribute != null ? normalizeCpr(responsibleUserCivilRegistrationNumberAttribute.getAttributeValue()) : null;
    }

    private LookupTypeEnumeration getLookupType(HsuidHeader hsuidHeader) {
        Attribute userTypeAttribute = getRequiredAttribute(hsuidHeader, NameConstraint.NSI_USERTYPE);
        switch(userTypeAttribute.getAttributeValue()) {
            case "nsi:Citizen":
                return getCitizenLookupType(hsuidHeader);
            case "nsi:HealthcareProfessional":
                return getHealthcareProfessionalLookupType(hsuidHeader);
            default:
                throw new IllegalArgumentException(String.format("Unexpected nsi:UserType: %s", userTypeAttribute.getAttributeValue()));
        }
    }

    private LookupTypeEnumeration getCitizenLookupType(HsuidHeader hsuidHeader) {
        if(lookupOnOwnData(hsuidHeader)) {
            return LookupTypeEnumeration.BORGER;
        }

        Attribute citizenUserRelationAttribute = getRequiredAttribute(hsuidHeader, NameConstraint.NSI_CITIZEN_USER_RELATION);
        return getLookupTypeFromCitizenUserRelation(citizenUserRelationAttribute);
    }

    private LookupTypeEnumeration getLookupTypeFromCitizenUserRelation(Attribute citizenUserRelationAttribute) {
        switch(citizenUserRelationAttribute.getAttributeValue()) {
            case "nsi:Citizen":
                return LookupTypeEnumeration.BORGER;
            case "nsi:ChildCustodyHolder":
                return LookupTypeEnumeration.FORÆLDER;
            case "nsi:Guardian":
                return LookupTypeEnumeration.VÆRGE;
            case "nsi:ProxyHolder":
                return LookupTypeEnumeration.FULDMAGTSHAVER;
            default:
                throw new IllegalArgumentException(String.format("Unexpected nsi:CitizenUserRelation: %s", citizenUserRelationAttribute.getAttributeValue()));
        }
    }

    private boolean lookupOnOwnData(HsuidHeader hsuidHeader) {
        Attribute responsibleUserCivilRegistrationNumberAttribute = getOptionalAttribute(hsuidHeader, NameConstraint.NSI_RESPONSIBLE_USER_CIVIL_REGISTRATION_NUMBER);
        return responsibleUserCivilRegistrationNumberAttribute == null;
    }

    private LookupTypeEnumeration getHealthcareProfessionalLookupType(HsuidHeader hsuidHeader) {
        Attribute responsibleUserAuthorizationCode = getOptionalAttribute(hsuidHeader, NameConstraint.NSI_RESPONSIBLE_USER_AUTHORIZATION_CODE);
        if(responsibleUserAuthorizationCode != null) {
            return LookupTypeEnumeration.KLINIKER;
        }

        // Tentative - may need to change.
        return LookupTypeEnumeration.KLINIKER;
    }

    private Attribute getRequiredAttribute(HsuidHeader hsuidHeader, NameConstraint nameConstraint) {
        return getAttribute(hsuidHeader, nameConstraint, true);
    }

    private Attribute getOptionalAttribute(HsuidHeader hsuidHeader, NameConstraint nameConstraint) {
        return getAttribute(hsuidHeader, nameConstraint, false);
    }

    private Attribute getAttribute(HsuidHeader hsuidHeader, NameConstraint nameConstraint, boolean required) {
        for(Attribute a : hsuidHeader.getAssertion().getAttributeStatement().getAttribute()) {
            if(a.getName().equals(nameConstraint)) {
                return a;
            }
        }

        if(required) {
            throw new IllegalArgumentException(String.format("No attribute matched name constraint %s", nameConstraint.toString()));
        }

        return null;
    }

    private PatientType buildPatient(RetrieveDocumentSetRequestType request) {
        PatientType patient = new PatientType();

        patient.setCprNumber(getPatientCprNumber(request));
        patient.setName("");

        return patient;
    }

    private String getPatientCprNumber(RetrieveDocumentSetRequestType request) {
        if(request.getDocumentRequest().isEmpty()) {
            // Returning the empty string instead of failing, as empty requests can occur.
            return "";
        }
        String documentUniqueId = request.getDocumentRequest().get(0).getDocumentUniqueId();

        // The documentUniqueId is a sequence of integers, separated by dots. The fourth-last is the encoded cpr.
        String[] parts = documentUniqueId.split("\\.");
        if(parts.length < 4) {
            throw new IllegalArgumentException(String.format("Illegal documentUniqueId: %s", documentUniqueId));
        }

        String encodedCpr = parts[parts.length - 4];
        String decodedCpr = decodeId(encodedCpr);
        return normalizeCpr(decodedCpr);
    }

    // Taken from DocumentIdEncoderDecoder
    private static final String magic = "b";
    private String decodeId(String encoded) {
        String decoded = new BigInteger(encoded).toString(16);
        if (decoded == null || !decoded.toLowerCase().startsWith(magic.toLowerCase())) {
            throw new IllegalArgumentException();
        }
        return decoded.substring(magic.length());
    }

    private ConsentInfoType buildConsent(HsuidHeader hsuidHeader) {
        ConsentInfoType consentInfo = new ConsentInfoType();

        if(consentOverridden(hsuidHeader)) {
            consentInfo.setCode(LogRecordConstants.CONSENT_OVERRIDDEN_CODE);
            consentInfo.setText(LogRecordConstants.CONSENT_OVERRIDDEN_TEXT);
            consentInfo.setPrivacyOverrideCode(LogRecordConstants.CONSENT_OVERRIDDEN_PRIVACY_OVERRIDE_CODE);
        }
        else {
            consentInfo.setCode(LogRecordConstants.CONSENT_GRANTED_CODE);
            consentInfo.setText(LogRecordConstants.CONSENT_GRANTED_TEXT);
            consentInfo.setPrivacyOverrideCode(LogRecordConstants.CONSENT_GRANTED_PRIVACY_OVERRIDE_CODE);
        }

        return consentInfo;
    }

    private boolean consentOverridden(HsuidHeader hsuidHeader) {
        Attribute consentOverrideAttribute = getOptionalAttribute(hsuidHeader, NameConstraint.NSI_CONSENT_OVERRIDE);
        return consentOverrideAttribute != null && consentOverrideAttribute.getAttributeValue().equals(ConsentOverrideAttribute.LegalValues.TRUE.getValue());
    }

    private ExtendedRequestInfoType buildExtendedRequestInfo(RetrieveDocumentSetRequestType request) {
        ExtendedRequestInfoType extendedRequestInfo = new ExtendedRequestInfoType();

        extendedRequestInfo.setFormat("Text");
        String requestBody = request
            .getDocumentRequest()
            .stream()
            .map(dr -> String.format("{HomeCommunityId: %s, RepositoryUniqueId: %s, DocumentUniqueId: %s}", dr.getHomeCommunityId(), dr.getRepositoryUniqueId(), dr.getDocumentUniqueId()))
            .collect(Collectors.joining(", "));
        extendedRequestInfo.setValue(requestBody);

        return extendedRequestInfo;
    }

    private String normalizeCpr(String cpr) {
        if(cpr == null || cpr.length() < 10 || cpr.length() > 11) {
            throw new IllegalArgumentException(String.format("Malformed cpr: %s", cpr));
        }

        if(cpr.length() == 11) {
            return cpr;
        }
        return cpr.substring(0, 6) + "-" + cpr.substring(6, 10);
    }
}
