package dk.nsi.dds.projects.sxa.documentprovider.audit;

import dk.nsi.dds.projects.sxa.common.configuration.SXAConfigKeys;
import dk.nsi.dds.projects.sxa.documentprovider.context.HttpHeaderContext;
import dk.nsi.hsuid._2016._08.hsuid_1_1.HsuidHeader;
import dk.nsi.util.Config;
import loginfo.healthcare._2017_03_01.LogRecord;
import org.apache.cxf.binding.soap.SoapFault;
import org.apache.cxf.binding.soap.SoapHeader;
import org.apache.cxf.headers.Header;
import org.apache.log4j.Logger;
import org.openehealth.ipf.commons.ihe.xds.core.ebxml.ebxml30.RetrieveDocumentSetRequestType;
import org.w3c.dom.Node;

import javax.annotation.Priority;
import javax.ejb.Stateless;
import javax.inject.Inject;
import javax.inject.Named;
import javax.interceptor.AroundInvoke;
import javax.interceptor.Interceptor;
import javax.interceptor.InvocationContext;
import javax.servlet.http.HttpServletRequest;
import javax.xml.bind.JAXBContext;
import javax.xml.namespace.QName;
import javax.xml.ws.WebServiceContext;
import javax.xml.ws.handler.MessageContext;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.List;

@Interceptor
@BuildLogRecord
@Priority(Interceptor.Priority.APPLICATION)
public class LogRecordHeaderInterceptor {
    public static final String HSUID_HEADER_NAMESPACE = "http://www.nsi.dk/hsuid/2016/08/hsuid-1.1.xsd";
    public static final String HSUID_HEADER_NAME = "HsuidHeader";

    public static final String MEDCOM_HEADER_NAMESPACE = "http://www.medcom.dk/dgws/2006/04/dgws-1.0.xsd";
    public static final String MEDCOM_HEADER_NAME = "Header";

    private static final Logger logger = Logger.getLogger(LogRecordHeaderInterceptor.class);

    private WebServiceContext webServiceContext;

    private HttpHeaderContext httpHeaderContext;

    private Config config;

    private LogRecordBuilder logRecordBuilder;

    @Inject
    public LogRecordHeaderInterceptor(WebServiceContext webServiceContext, HttpHeaderContext httpHeaderContext, Config config, LogRecordBuilder logRecordBuilder) {
        this.webServiceContext = webServiceContext;
        this.httpHeaderContext = httpHeaderContext;
        this.config = config;
        this.logRecordBuilder = logRecordBuilder;
    }

    @AroundInvoke
    public Object generateLogRecordHeader(InvocationContext context) throws Exception {
        // Extract required information from headers, request body, etc.
        HsuidHeader hsuidHeader = extractHsuidHeader();
        dk.medcom.dgws._2006._04.dgws_1_0.Header header = extractMedcomHeader();
        LogRecord inboundLogRecord = extractInboundLogRecord();
        RetrieveDocumentSetRequestType request = extractRetrieveDocumentSetRequest(context);

        try {
            // Generate the LogRecord
            LogRecord logRecord = logRecordBuilder.buildLogRecord(hsuidHeader, header, inboundLogRecord, request);

            // Log the message id and the generated request id
            logMessageIdAndRequestId(header, logRecord);

            // Serialize the LogRecord, put it on the header context
            serializeAndAddToContext(logRecord);
        }
        catch(IllegalArgumentException e) {
            throw new SoapFault(String.format("Could not generate LogRecord: %s", e.getMessage()), e, new QName("http://www.w3.org/2003/05/soap-envelope", "Sender"));
        }

        return context.proceed();
    }

    private HsuidHeader extractHsuidHeader() throws Exception {
        List<SoapHeader> soapHeaders = (List<SoapHeader>) webServiceContext.getMessageContext().get(Header.HEADER_LIST);

        for(SoapHeader sh : soapHeaders) {
            if(sh.getName().getNamespaceURI().equals(HSUID_HEADER_NAMESPACE) && sh.getName().getLocalPart().equals(HSUID_HEADER_NAME)) {
                return (HsuidHeader) JAXBContext.newInstance(HsuidHeader.class).createUnmarshaller().unmarshal((Node) sh.getObject());
            }
        }

        throw new IllegalStateException("No HsuidHeader found in request!");
    }

    private dk.medcom.dgws._2006._04.dgws_1_0.Header extractMedcomHeader() throws Exception {
        List<SoapHeader> soapHeaders = (List<SoapHeader>) webServiceContext.getMessageContext().get(Header.HEADER_LIST);

        for(SoapHeader sh : soapHeaders) {
            if(sh.getName().getNamespaceURI().equals(MEDCOM_HEADER_NAMESPACE) && sh.getName().getLocalPart().equals(MEDCOM_HEADER_NAME)) {
                return (dk.medcom.dgws._2006._04.dgws_1_0.Header) JAXBContext.newInstance(dk.medcom.dgws._2006._04.dgws_1_0.Header.class).createUnmarshaller().unmarshal((Node) sh.getObject());
            }
        }

        throw new IllegalStateException("No HsuidHeader found in request!");
    }

    private LogRecord extractInboundLogRecord() throws Exception {
        MessageContext messageContext = webServiceContext.getMessageContext();
        if(!messageContext.containsKey("HTTP.REQUEST") || !(messageContext.get("HTTP.REQUEST") instanceof HttpServletRequest)) {
            throw new IllegalStateException("Could not extract http request!");
        }

        HttpServletRequest request = (HttpServletRequest) messageContext.get("HTTP.REQUEST");

        String logRecordHeader = request.getHeader(config.getProperty(SXAConfigKeys.SXA_LOGRECORD_INBOUND_HEADER_NAME));
        if(logRecordHeader == null || logRecordHeader.isEmpty()) {
            return null;
        }

        return (LogRecord) JAXBContext.newInstance(LogRecord.class).createUnmarshaller().unmarshal(new StringReader(logRecordHeader));
    }

    private RetrieveDocumentSetRequestType extractRetrieveDocumentSetRequest(InvocationContext context) {
        Object[] parameters = context.getParameters();
        if(parameters.length != 1 || !(parameters[0] instanceof RetrieveDocumentSetRequestType)) {
            throw new IllegalStateException("Expected exactly one argument of type RetrieveDocumentSetRequestType!");
        }
        return (RetrieveDocumentSetRequestType) parameters[0];
    }

    private void logMessageIdAndRequestId(dk.medcom.dgws._2006._04.dgws_1_0.Header header, LogRecord logRecord) {
        String messageID = getMessageID(header);
        String requestId = getRequestId(logRecord);

        logger.info(String.format("LogRecord generated. MessageID: %s, RequestId: %s", messageID, requestId));
    }

    private String getMessageID(dk.medcom.dgws._2006._04.dgws_1_0.Header header) {
        if(header.getLinking() != null && header.getLinking().getMessageID() != null) {
            return header.getLinking().getMessageID();
        }
        throw new IllegalArgumentException("No MessageID on medcom header, but that element is required!");
    }

    private String getRequestId(LogRecord logRecord) {
        if(logRecord.getTraceInfo() != null && logRecord.getTraceInfo().getRequestId() != null) {
            return logRecord.getTraceInfo().getRequestId();
        }
        throw new IllegalArgumentException("No RequestId on LogRecord!");
    }

    private void serializeAndAddToContext(LogRecord logRecord) throws Exception {
        try(StringWriter writer = new StringWriter()) {
            JAXBContext.newInstance(LogRecord.class).createMarshaller().marshal(logRecord, writer);
            httpHeaderContext.addHttpHeader(config.getProperty(SXAConfigKeys.SXA_LOGRECORD_OUTBOUND_HEADER_NAME), writer.toString());
        }
    }
}
