/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentprovider.document;

import java.util.regex.Pattern;

import org.joda.time.DateTime;
import org.joda.time.DateTimeZone;

import dk.nsi.dds.projects.sxa.common.InvalidDateException;
import dk.nsi.dds.projects.sxa.common.PatientIdDerivationException;

/**
 * Holds an instance of a patient id (CPR and eCPR).
 */
public class PatientId {
  private static final String DEFAULT_BIRTHDAY_FORMAT = "ddMMYYYY";
  private static final Pattern PATIENT_ID_PATTERN = Pattern.compile("^(\\d){10}$");
  public static final int NORMAL_DAY_RANGE_START = 1;
  public static final int NORMAL_DAY_RANGE_END = 31;
  public static final int ALTERNATIVE_DAY_RANGE_START = 61;
  public static final int ALTERNATIVE_DAY_RANGE_END = 91;
  public static final int YEAR_RANGE_START = 0;
  public static final int YEAR_RANGE_END = 99;
  public static final int MONTH_RANGE_START = 1;
  public static final int MONTH_RANGE_END = 12;
  public static final char GENDER_FEMALE = 'F';
  public static final char GENDER_MALE = 'M';
  private int day;
  private int month;
  private int year;
  private int controlCiffer;
  private boolean isAlternative;
  private String rawId;

  private PatientId(int day, int month, int year, int controlCiffer, boolean isAlternative, String rawId) {
    super();
    this.day = day;
    this.month = month;
    this.year = year;
    this.controlCiffer = controlCiffer;
    this.isAlternative = isAlternative;
    this.rawId = rawId;
  }

  public static PatientId createPatientId(String rawPatientId) throws PatientIdDerivationException {
    if (rawPatientId == null) {
      throw new PatientIdDerivationException("PatientId is null");
    }
    rawPatientId = rawPatientId.trim();
    if (!PATIENT_ID_PATTERN.matcher(rawPatientId).matches()) {
      throw new PatientIdDerivationException("PatientId format is invalid");
    }
    boolean isAlternative = isAlternative(rawPatientId);
    assertValidPatientDate(rawPatientId, isAlternative);
    return new PatientId(parseSubstring(rawPatientId, 0, 2), parseSubstring(rawPatientId, 2, 4),
      parseSubstring(rawPatientId, 4, 6), parseSubstring(rawPatientId, 6, 10), isAlternative, rawPatientId);
  }

  private static int parseSubstring(String text, int start, int end) {
    return Integer.parseInt(text.substring(start, end));
  }

  private static boolean isAlternative(String rawPatientId) throws PatientIdDerivationException {
    int day = Integer.parseInt(rawPatientId.substring(0, 2));
    if (day < NORMAL_DAY_RANGE_START || (day > NORMAL_DAY_RANGE_END && day < ALTERNATIVE_DAY_RANGE_START)
      || day > ALTERNATIVE_DAY_RANGE_END) {
      throw new PatientIdDerivationException("Day is invalid: " + day);
    }
    return day > NORMAL_DAY_RANGE_END;
  }

  /**
   * Verifies that given patient id has CPR number form and contains date information
   * which is sensible. The method must not be used as validation whether the patient
   * id identifies a real person or not.
   * 
   * @param patientId
   * @throws PatientIdDerivationException
   */
  private static void assertValidPatientDate(String patientId, boolean isAlternative)
    throws PatientIdDerivationException {
    int day = Integer.parseInt(patientId.substring(0, 2));
    int month = Integer.parseInt(patientId.substring(2, 4));
    Integer.parseInt(patientId.substring(4, 6)); // year sensible?
    if (isAlternative) {
      day = day - ALTERNATIVE_DAY_RANGE_START + 1;
    }
    if (day < NORMAL_DAY_RANGE_START || day > NORMAL_DAY_RANGE_END || month < MONTH_RANGE_START
      || month > MONTH_RANGE_END) {
      throw new PatientIdDerivationException("Day and/or month not of sensible values in patient id");
    }
  }

  public char determineGender() throws PatientIdDerivationException {
    if (controlCiffer % 2 == 0) {
      return GENDER_FEMALE;
    }
    return GENDER_MALE;
  }

  /**
   * Calculates person birthday from person id. Algorithm for determining birth year
   * takes two first digits from current year, adds two digits from person id, and
   * compares the resulting year with current year. If resulting year is greater than
   * current year, 100 years are subtracted from the resulting year.
   * 
   * Example: CurrentYear: 2015 PersonId: 010116 Output: 1916
   * 
   * CurrentYear: 2015 personId: 010115 Output: 2015
   * 
   * @param patientId
   * @return
   * @throws InvalidDateException
   */
  public DateTime deriveDateOfBirth() throws InvalidDateException {
    try {
      DateTime now = new DateTime(DateTimeZone.UTC);
      int currentYear = now.getYear();
      String currentYearText = currentYear + "";
      int proposedYear = Integer
        .parseInt(currentYearText.substring(0, currentYearText.length() - 2) + String.format("%02d", year));
      int correctedDay = isAlternative ? day - ALTERNATIVE_DAY_RANGE_START + 1 : day;
      DateTime result = new DateTime(proposedYear, month, correctedDay, 0, 0, DateTimeZone.UTC);
      if (result.isAfter(now)) {
        result = result.minusYears(100);
      }
      return result;
    } catch (Exception ex) {
      throw new InvalidDateException(ex);
    }
  }

  public String determineDateofBirthText() throws InvalidDateException {
    return determineDateofBirthText(DEFAULT_BIRTHDAY_FORMAT);
  }

  public String determineDateofBirthText(String format) throws InvalidDateException {
    return deriveDateOfBirth().toString(format);
  }

  public boolean isAlternative() {
    return isAlternative;
  }

  public String getRawId() {
    return rawId;
  }

  /**
   * Converts year to a two digit string e.g: 1980 -> 80
   * 
   * @param year    Complete year as a number
   * @return        Truncated year as String
   */
  @SuppressWarnings("unused")
private static String yearToCprText(int year) {
    String textYear = Integer.toString(year);
    if (textYear.length() > 2) {
      return textYear.substring(textYear.length() - 2, textYear.length());
    }
    return textYear;
  }
}
