/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentprovider.impl;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import javax.ejb.ConcurrencyManagement;
import javax.ejb.ConcurrencyManagementType;
import javax.ejb.Singleton;
import javax.ejb.Startup;
import javax.enterprise.inject.Produces;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.ws.WebServiceContext;

import dk.nsi.dds.projects.sxa.common.configuration.SXAConfigKeys;
import dk.nsi.dds.projects.sxa.common.configuration.SXADocumentProviderConfig;
import dk.nsi.dds.projects.sxa.documentprovider.audit.DateTimeProvider;
import dk.nsi.dds.projects.sxa.documentprovider.audit.LogRecordBuilder;
import org.apache.cxf.endpoint.Client;
import org.apache.cxf.frontend.ClientProxy;
import org.apache.log4j.Logger;

import com.cgi.svareksponering.ISvarEksponering;
import com.cgi.svareksponering.ResultatPakkePIH;

import dk.nsi.common.servicestatuscheck.ServiceStatusCheckBean;
import dk.nsi.common.util.SLALogItemFactory;
import dk.nsi.common.util.SLALoggingDataHandler;
import dk.nsi.dds.projects.sxa.common.PracticeSettingCodesHandler;
import dk.nsi.dds.projects.sxa.common.PracticeSettingCodesMappingConfiguration;
import dk.nsi.dds.projects.sxa.common.util.configloading.ConfigFileFinder;
import dk.nsi.dds.projects.sxa.common.util.configloading.Log4jConfigurationLoader;
import dk.nsi.dds.projects.sxa.common.util.xmlhandling.XStreamFileException;
import dk.nsi.dds.projects.sxa.documentprovider.context.HttpHeaderContext;
import dk.nsi.dds.projects.sxa.documentprovider.interceptor.HttpMessageInterceptor;
import dk.nsi.dds.projects.sxa.documentprovider.invoker.GetDataFromPeriodParametersFactory;
import dk.nsi.dds.projects.sxa.documentprovider.invoker.SXAServiceInvokerFactory;
import dk.nsi.dds.projects.sxa.documentprovider.ws.DocumentProviderWS;
import dk.nsi.schema.privacy.pih._1.PrivacyInformationHeaderType;
import dk.nsi.schema.privacy.piw._1.PrivacyInformationWrapperType;
import dk.nsi.util.Config;
import dk.sdsd.nsp.slalog.api.SLALogConfig;
import dk.sdsd.nsp.slalog.api.SLALogger;

@Singleton
@Startup
@ConcurrencyManagement(ConcurrencyManagementType.BEAN)
public class ConfigurationLoader {
	private static final Logger log = Logger.getLogger(ConfigurationLoader.class);

	private SLALogItemFactory slaLogItemFactory;
	public static final String PROVIDER_CONFIGURATION_FILE_NAME = "sxaproviderconfiguration.xml";
	public static final String PRACTICESETTINGCODES_CONFIGURATION_FILE_NAME = "sxapracticesettingcodes.xml";
	private ProviderConfigurationHandler providerConfigurationHandler;
	private PracticeSettingCodesHandler practiceSettingCodesHandler;
	private SXAServiceInvokerFactory sxaDispatchFactory;
	private GetDataFromPeriodParametersFactory getDataFromPeriodParametersFactory;
	private JAXBContext jaxbContext;
	private Config config;
	private ServiceStatusCheckBean serviceStatusCheckBean;
	private HttpHeaderContext httpHeaderContext;
	private LogRecordBuilder logRecordBuilder;

	@Resource
	private WebServiceContext webServiceContext;

	@PostConstruct
	public void initialize() {
		try {
			Log4jConfigurationLoader.readLog4jConfigurationFile("log4j-sxa-documentprovider-ws.xml");
			setupSLALogging();

			config = new SXADocumentProviderConfig("sxadocumentprovider");
			createProductionConfigurationHandler();
			createPracticeSettingCodesConfigurationHandler();
			sxaDispatchFactory = new SXAServiceInvokerFactory();
			sxaDispatchFactory.initialize(config);
			getDataFromPeriodParametersFactory = new GetDataFromPeriodParametersFactory(DatatypeFactory.newInstance());
			jaxbContext = createJAXBContext();
			serviceStatusCheckBean = createServiceStatusCheckBean();
			httpHeaderContext = createHttpHeaderContext();
			logRecordBuilder = new LogRecordBuilder(new DateTimeProvider());
		} catch (Exception ex) {
			throw new RuntimeException("Exception during service initialization: " + ex.getMessage(), ex);
		}
	}

	private JAXBContext createJAXBContext() {
		try {
			return JAXBContext.newInstance(ResultatPakkePIH.class, PrivacyInformationHeaderType.class, PrivacyInformationWrapperType.class);
		} catch (JAXBException e) {
			throw new RuntimeException(e);
		}
	}

	private void createProductionConfigurationHandler() throws XStreamFileException {
		providerConfigurationHandler = new ProviderConfigurationHandler(ConfigFileFinder.loadFile(PROVIDER_CONFIGURATION_FILE_NAME));
		providerConfigurationHandler.readConfiguration();
	}

	private void createPracticeSettingCodesConfigurationHandler() throws XStreamFileException {
		practiceSettingCodesHandler = new PracticeSettingCodesHandler(ConfigFileFinder.loadFile(PRACTICESETTINGCODES_CONFIGURATION_FILE_NAME));
		practiceSettingCodesHandler.readConfiguration();
	}

	private void setupSLALogging() {
		SLALogger slaLogger = new SLALogConfig(DocumentProviderWS.class.getName(), "sxa-documentprovider").getSLALogger();
		slaLogItemFactory = new SLALogItemFactory(slaLogger, DocumentProviderWS.class.getName());
	}

	@Produces
	public SLALoggingDataHandler createSLALoggingDataHandler() {
		return new SLALoggingDataHandler(slaLogItemFactory, webServiceContext);
	}

	private ServiceStatusCheckBean createServiceStatusCheckBean() {
		ServiceStatusCheckBean bean = new ServiceStatusCheckBean();
		bean.registerCounter("SXAInvoker", Integer.parseInt(config.getProperty(SXAConfigKeys.SXA_SERVICECHECK_FAILURE_THRESHOLD)));
		return bean;
	}

	private HttpHeaderContext createHttpHeaderContext() {
		return new HttpHeaderContext();
	}

	@Produces
	public ProviderConfigurationHandler getProviderConfigurationHandler() {
		return providerConfigurationHandler;
	}

	@Produces
	public PracticeSettingCodesMappingConfiguration getPracticeSettingCodesConfigurationHandler() throws XStreamFileException {
		return practiceSettingCodesHandler.readConfiguration();
	}

	@Produces
	public Config getConfig() {
		return config;
	}

	@Produces
	public ISvarEksponering getISvarEksponering() {
		try {
			ISvarEksponering service = sxaDispatchFactory.getPort();
			Client proxy = ClientProxy.getClient(service);
			proxy.getOutInterceptors().add(new HttpMessageInterceptor(getHttpHeaderContext()));
			return service;
		} catch (ClassCastException cce) {
			throw new RuntimeException(cce.getMessage());
		} finally {
			log.info("OUT - getISvarEksponering");
		}
	}

	@Produces
	public GetDataFromPeriodParametersFactory getGetDataFromPeriodParametersFactory() {
		return getDataFromPeriodParametersFactory;
	}

	@Produces
	public JAXBContext getJAXBContext() {
		return jaxbContext;
	}

	@Produces
	public SLALogItemFactory createSLALogItemFactory() {
		return slaLogItemFactory;
	}

	@Produces
	public ServiceStatusCheckBean getServiceStatusCheckBean() {
		return serviceStatusCheckBean;
	}

	@Produces
	public HttpHeaderContext getHttpHeaderContext() {
		return httpHeaderContext;
	}

	@Produces
	public WebServiceContext getWebServiceContext() {
		return webServiceContext;
	}

	@Produces
	public LogRecordBuilder getLogRecordBuilder() {
		return logRecordBuilder;
	}
}
