/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentprovider.impl;

import java.util.List;

import javax.activation.DataHandler;
import javax.ejb.Stateless;
import javax.inject.Inject;
import javax.interceptor.Interceptors;

import org.apache.log4j.Logger;
import org.openehealth.ipf.commons.ihe.xds.core.requests.RetrieveDocument;
import org.openehealth.ipf.commons.ihe.xds.core.responses.ErrorCode;
import org.openehealth.ipf.commons.ihe.xds.core.responses.ErrorInfo;
import org.openehealth.ipf.commons.ihe.xds.core.responses.RetrievedDocument;
import org.openehealth.ipf.commons.ihe.xds.core.responses.Severity;

import dk.nsi.dds.projects.sxa.common.InvalidSXAQueryParameters;
import dk.nsi.dds.projects.sxa.common.PracticeSettingCodesConfiguration;
import dk.nsi.dds.projects.sxa.common.PracticeSettingCodesMappingConfiguration;
import dk.nsi.dds.projects.sxa.common.QueryDetails;
import dk.nsi.dds.projects.sxa.common.configuration.SXAMappedValue;
import dk.nsi.dds.projects.sxa.common.util.DocumentUniqueIdUtil;
import dk.nsi.dds.projects.sxa.common.util.perflogger.PerformanceLogger;
import dk.nsi.dds.projects.sxa.common.util.perflogger.PerformanceLogger.Monitor;
import dk.nsi.dds.projects.sxa.common.util.xmlhandling.XStreamFileException;
import dk.nsi.dds.projects.sxa.documentprovider.DocumentRetriever;
import dk.nsi.dds.projects.sxa.documentprovider.DocumentsToRetrieve;
import dk.nsi.dds.projects.sxa.documentprovider.ErrorInfoBuilder;
import dk.nsi.dds.projects.sxa.documentprovider.RetrievedDocumentsResult;
import dk.nsi.dds.projects.sxa.documentprovider.context.HttpHeaderContext;
import dk.nsi.dds.projects.sxa.documentprovider.exceptions.InvokerCallException;
import dk.nsi.dds.projects.sxa.documentprovider.exceptions.InvokerCallTimedoutException;
import dk.nsi.dds.projects.sxa.documentprovider.interceptor.HttpHeaderReaderInterceptor;
import dk.nsi.dds.projects.sxa.documentprovider.invoker.SXADataSource;
import dk.nsi.dds.projects.sxa.documentprovider.invoker.SXAInvokeHandler;
import dk.nsi.dds.projects.sxa.documentprovider.util.NoMatchingPracticeSettingCodesException;
import dk.nsi.dds.projects.sxa.documentprovider.util.SXAInvocationAttributes;
import dk.nsi.dds.projects.sxa.documentprovider.util.SXAInvocationAttributesFactory;
import dk.nsi.privacy.piw.Constants;
import dk.nsi.util.Config;

@Stateless
public class DocumentRetrieverImpl implements DocumentRetriever {
	private static final Logger logger = Logger.getLogger(DocumentRetrieverImpl.class);
	@Inject
	private SXAInvokeHandler sxaInvokeHandler;
	@Inject
	private ProviderConfigurationHandler providerConfigurationHandler;
	@Inject
	private PracticeSettingCodesMappingConfiguration practiceSettingCodesMappingConfiguration;
	@Inject
	private Config config;
    @Inject
    private HttpHeaderContext httpHeaderContext;

	@Override
	@Interceptors(HttpHeaderReaderInterceptor.class)
	public RetrievedDocumentsResult retrieveDocuments(DocumentsToRetrieve documentsToRetrieve) {
		RetrievedDocumentsResultImpl result = new RetrievedDocumentsResultImpl();
		try (Monitor monitor = new PerformanceLogger.Monitor("DocumentRetrieverImpl.retrieveDocuments")) {
			if (documentsToRetrieve == null) {
				logger.error("DocumentsToRetrieve is null");
				result.addError(createError(ErrorCode.REPOSITORY_ERROR));
				return result;
			} else if (documentsToRetrieve.getDocumentReferences().isEmpty()) {
				logger.error("DocumentsToRetrieve is empty");
				result.addError(createError(ErrorCode.REPOSITORY_ERROR));
				return result;
			}
			
			
			PracticeSettingCodesConfiguration practiceSettingCodesConfiguration = practiceSettingCodesMappingConfiguration.getPracticeSettingCodesConfiguration();
			List<SXAMappedValue> validPracticeSettingCodesListPracticeSettingCodesList = practiceSettingCodesConfiguration.getPracticeSettingCodes();
			
			String patientId = null;
			for (RetrieveDocument documentReference : documentsToRetrieve.getDocumentReferences()) {
				try {
					String uniqueIdWithoutRoot = DocumentUniqueIdUtil.getUniqueIdWithoutRoot(getProviderConfiguration().getUniqueIdRoot(), documentReference.getDocumentUniqueId());
					QueryDetails query = new QueryDetails(uniqueIdWithoutRoot, validPracticeSettingCodesListPracticeSettingCodesList);
					if(patientId == null) {
						//Holder ved første patient id - den kan være null
						patientId = query.getPatientId();
					}
					
					if (patientId == null || !patientId.equals(query.getPatientId())) {
						result.addError(createError(documentReference, ErrorCode.REPOSITORY_ERROR));
						continue;
					}
					
					SXAInvocationAttributes attributes = SXAInvocationAttributesFactory.createInvocationAttributes(query, config, httpHeaderContext);
					result.addDocument(getDocument(documentReference, query, attributes));
				} catch (NoMatchingPracticeSettingCodesException e) {
					logger.warn(e.getMessage(), e);
					result.addError(createError(documentReference, ErrorCode.REPOSITORY_ERROR, "Forespørgslens søgeparametre for practiceSettingCode kunne ikke omsættes til laboratorie-specialer"));
				} catch (InvalidSXAQueryParameters e) {
					logger.warn(e.getMessage(), e);
					result.addError(createError(documentReference, ErrorCode.REPOSITORY_ERROR, "Der var en fejl i forespørgslens parametre: " + e.getMessage()));
				} catch (InvokerCallTimedoutException e) {
					logger.warn(e.getMessage(), e);
					//Unreachable
					result.addError(createError(documentReference, ErrorCode.REPOSITORY_BUSY));
				} catch (InvokerCallException e) {
					logger.warn(e.getMessage());
					result.addError(createError(documentReference, ErrorCode.REPOSITORY_ERROR, "Der opstod en fejl under invokeringen af laboratoriesvarsservicen (" + e.getMessage() + ")"));
				} catch (Exception e) {
					logger.warn(e.getMessage(), e);
					result.addError(createError(documentReference, ErrorCode.REPOSITORY_ERROR, "Intern fejl i Svareksponeringsservice XDS-adapter"));
				}
			}
		} catch (Exception ex) {
			logger.error(ex.getMessage(), ex);
			result.addError(createError(ErrorCode.REPOSITORY_ERROR));
		}
		return result;
	}
	
	private ErrorInfo createError(ErrorCode errorCode) {
		return createError(null, errorCode, null);
	}
	
	private ErrorInfo createError(RetrieveDocument retrieveDocument, ErrorCode errorCode) {
		return createError(retrieveDocument, errorCode, null);
	}

	private ErrorInfo createError(RetrieveDocument retrieveDocument, ErrorCode errorCode, String customErrorCode) {
		return new ErrorInfoBuilder().errorCode(errorCode).severity(Severity.ERROR).customErrorCode(customErrorCode).codeContext(retrieveDocument).build();
	}

	@Override
	public ProviderConfiguration getProviderConfiguration() {
		try {
			return providerConfigurationHandler.readConfiguration();
		} catch (XStreamFileException e) {
			logger.error(e);
			return null;
		}
	}

	private RetrievedDocument getDocument(RetrieveDocument documentReference, QueryDetails query, SXAInvocationAttributes attributes) throws InvokerCallTimedoutException, InvokerCallException {
		final byte[] documentBytes = sxaInvokeHandler.getDocument(documentReference.getDocumentUniqueId(), attributes);
		SXADataSource dataSource = new SXADataSource(documentBytes, "text/xml; charset=UTF-8");
		DataHandler data = new DataHandler(dataSource);

		RetrievedDocument result = new RetrievedDocument();
		result.setDataHandler(data);
		result.setRequestData(documentReference);
		result.setMimeType(Constants.MIME_TYPE);

		return result;
	}
}
