/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (SDS), Denmark
 * (http://sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (SDS), Denmark (http://sundhedsdatastyrelsen.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentprovider.impl;

import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.Logger;
import org.openehealth.ipf.commons.ihe.xds.core.requests.RetrieveDocument;
import org.openehealth.ipf.commons.ihe.xds.core.responses.ErrorCode;
import org.openehealth.ipf.commons.ihe.xds.core.responses.ErrorInfo;
import org.openehealth.ipf.commons.ihe.xds.core.responses.Severity;

import dk.nsi.dds.projects.sxa.documentprovider.ErrorInfoBuilder;

public class ProviderFilter {
	private static final Logger LOGGER = Logger.getLogger(ProviderFilter.class);

	private static final String DIFFERING_HOME_COMMUNITY_ID_MSG = "This document provider has home community id which was not specified in the request";
	private static final String DIFFERING_REPOSITORY_UNIQUE_ID_MSG = "This document provider has repository unique id which was not specified in the request";

	private ProviderConfiguration configuration;

	public ProviderFilter(ProviderConfiguration configuration) {
		super();
		this.configuration = configuration;
	}

	public ErrorInfo filter(RetrieveDocument retrieveDocument) {

		if (StringUtils.isEmpty(retrieveDocument.getHomeCommunityId()) || StringUtils.isEmpty(configuration.getHomeCommunityId())) {
			return createError(retrieveDocument, ErrorCode.MISSING_HOME_COMMUNITY_ID, DIFFERING_HOME_COMMUNITY_ID_MSG);
		}
		if (!StringUtils.equals(retrieveDocument.getHomeCommunityId(), configuration.getHomeCommunityId())) {
			return createError(retrieveDocument, ErrorCode.UNKNOWN_COMMUNITY, DIFFERING_HOME_COMMUNITY_ID_MSG);
		}
		if (!StringUtils.equals(retrieveDocument.getRepositoryUniqueId(), configuration.getRepositoryUniqueId())) {
			return createError(retrieveDocument, ErrorCode.UNKNOWN_REPOSITORY_ID, DIFFERING_REPOSITORY_UNIQUE_ID_MSG);
		}
		return null;
	}

	private ErrorInfo createError(RetrieveDocument retrieveDocument, /* ProviderConfiguration providerConfiguration, */ ErrorCode errorCode, String errorText) {
		LOGGER.info("RetriveDocument contains error: " + errorCode.getOpcode());

		final String documentUniqueId = retrieveDocument.getDocumentUniqueId();
		final String homeCommunityId = retrieveDocument.getHomeCommunityId();
		final String repositoryUniqueId = retrieveDocument.getRepositoryUniqueId();
		return new ErrorInfoBuilder().errorCode(errorCode).severity(Severity.ERROR).codeContext(documentUniqueId, homeCommunityId, repositoryUniqueId).build();

	}
}
