/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (SDS), Denmark
 * (http://sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (SDS), Denmark (http://sundhedsdatastyrelsen.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentprovider.impl;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;
import org.openehealth.ipf.commons.ihe.xds.core.requests.RetrieveDocument;
import org.openehealth.ipf.commons.ihe.xds.core.requests.RetrieveDocumentSet;
import org.openehealth.ipf.commons.ihe.xds.core.responses.ErrorInfo;
import org.openehealth.ipf.commons.ihe.xds.core.responses.RetrievedDocumentSet;
import org.openehealth.ipf.commons.ihe.xds.core.responses.Status;

import dk.nsi.dds.projects.sxa.common.util.perflogger.PerformanceLogger;
import dk.nsi.dds.projects.sxa.common.util.perflogger.PerformanceLogger.Monitor;
import dk.nsi.dds.projects.sxa.documentprovider.DocumentRetriever;
import dk.nsi.dds.projects.sxa.documentprovider.DocumentsToRetrieve;
import dk.nsi.dds.projects.sxa.documentprovider.FilteredDocumentsToRetrieve;
import dk.nsi.dds.projects.sxa.documentprovider.RetrievedDocumentsResult;

public class ProviderRequestHandlerImpl implements ProviderRequestHandler {
	private static final Logger logger = Logger.getLogger(ProviderRequestHandlerImpl.class);
	private DocumentRetriever documentRetriever;

	public ProviderRequestHandlerImpl(DocumentRetriever documentRetriever) {
		super();
		this.documentRetriever = documentRetriever;
	}

	@Override
	public RetrievedDocumentSet handleRequest(RetrieveDocumentSet retrieveDocumentSet) {
		RetrievedDocumentSet result = new RetrievedDocumentSet();
		if (retrieveDocumentSet == null || retrieveDocumentSet.getDocuments() == null || retrieveDocumentSet.getDocuments().size() == 0) {
			// Vi spørger ikke efter noget - returnér med det samme
			result.setStatus(Status.SUCCESS);
			return result;
		}

		DocumentsToRetrieve documentsToRetrieve = new DocumentsToRetrieveImpl(retrieveDocumentSet.getDocuments());
		RetrievedDocumentsResult retrievedDocuments = getDocuments(documentsToRetrieve);

		if(retrievedDocuments.getDocuments().size() == 0) {
			result.setStatus(Status.FAILURE);
		} else if(result.getErrors().size() > 0) {
			result.setStatus(Status.PARTIAL_SUCCESS);
		} else {
			result.setStatus(Status.SUCCESS);
		}

		result.getDocuments().addAll(retrievedDocuments.getDocuments());
		result.getErrors().addAll(retrievedDocuments.getErrors());
		
		return result;

	}
	
	private RetrievedDocumentsResult getDocuments(DocumentsToRetrieve documentsToRetrieve) {		
		FilteredDocumentsToRetrieve verifiedToRetrieve = filterErroneousMetadata(documentsToRetrieve);

		RetrievedDocumentsResult retrievedDocumentsResult;	
		try (Monitor monitor = new PerformanceLogger.Monitor("ProviderRequestHandlerImpl.getDocuments documentRetriever")) {
			retrievedDocumentsResult = documentRetriever.retrieveDocuments(verifiedToRetrieve);
			retrievedDocumentsResult.getErrors().addAll(verifiedToRetrieve.getErrors());
		}

		return retrievedDocumentsResult;
	}
	
	private FilteredDocumentsToRetrieve filterErroneousMetadata(DocumentsToRetrieve documentsToRetrieve) {
		
		List<RetrieveDocument> documentReferences = new ArrayList<>();
		List<ErrorInfo> errors = new ArrayList<>();

		ProviderFilter filter = new ProviderFilter(documentRetriever.getProviderConfiguration());
		for (RetrieveDocument documentReference : documentsToRetrieve.getDocumentReferences()) {
			ErrorInfo error = filter.filter(documentReference);
			if (error != null) {
				errors.add(error);
			} else {
				documentReferences.add(documentReference);
			}
		}

		logger.debug("filterErroneousMetadata: List size in="+documentsToRetrieve.size()+", List size out="+documentReferences.size());

		return new FilteredDocumentsToRetrieveImpl(documentReferences, errors);
	}
}
