/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentprovider.invoker;

import java.io.IOException;
import java.io.StringWriter;
import java.nio.charset.StandardCharsets;

import javax.enterprise.context.ApplicationScoped;
import javax.inject.Inject;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;

import org.apache.log4j.Logger;

import com.cgi.svareksponering.ResultatPakkePIH;

import dk.nsi.dds.projects.sxa.common.util.perflogger.PerformanceLogger;
import dk.nsi.dds.projects.sxa.common.util.perflogger.PerformanceLogger.Monitor;
import dk.nsi.dds.projects.sxa.documentprovider.exceptions.InvokerCallException;
import dk.nsi.dds.projects.sxa.documentprovider.exceptions.InvokerCallTimedoutException;
import dk.nsi.dds.projects.sxa.documentprovider.util.SXAInvocationAttributes;
import dk.nsi.privacy.piw.PIHDocumentWrapperBuilder;
import dk.nsi.schema.privacy.pih._1.PrivacyInformationHeaderType;
import dk.nsi.schema.privacy.piw._1.ObjectFactory;
import dk.nsi.schema.privacy.piw._1.PrivacyInformationWrapperType;

@ApplicationScoped
public class SXAInvokeHandler {
  private static final Logger logger = Logger.getLogger(SXAInvokeHandler.class);
  @Inject
  private SXAInvoker sxaInvoker;
  @Inject
  private JAXBContext jaxbContext;

  public byte[] getDocument(String documentId, SXAInvocationAttributes attributes)
    throws InvokerCallTimedoutException, InvokerCallException {
    try (Monitor monitor = new PerformanceLogger.Monitor("SvarEksponeringInvoker.getDocument")) {
      return toByteArray(wrapInPIW(documentId, sxaInvoker.invoke(attributes)));
    } catch (InvokerCallException | InvokerCallTimedoutException e) {
      throw e;
    } catch (Exception e) {
      logger.debug("Retrieve document failed", e);
      throw new InvokerCallException(e);
    }
  }

  private JAXBElement<PrivacyInformationWrapperType> wrapInPIW(
    String documentId, SXAResponseWrapper svarEksponeringResponseWrapper) {
    return new ObjectFactory().createWrapper(
      new PIHDocumentWrapperBuilder().addPrivacyInformationHeader(extractPIH(svarEksponeringResponseWrapper))
        .setDocument(documentId, toByteArray(extractResultatPakkePIH(svarEksponeringResponseWrapper)))
        .setOriginalMimeType(svarEksponeringResponseWrapper.getMimeType())
        .build());
  }

  private PrivacyInformationHeaderType extractPIH(SXAResponseWrapper svarEksponeringResponseWrapper) {
    PrivacyInformationHeaderType pih = svarEksponeringResponseWrapper.getPrivacyInformationHeaderType();
    if (pih == null) {
      throw new InvokerCallException("PIH is missing from SvarEksponeringsservice response");
    }
    return pih;
  }

  private ResultatPakkePIH extractResultatPakkePIH(SXAResponseWrapper svarEksponeringResponseWrapper) {
    ResultatPakkePIH resultatPakkePIH = svarEksponeringResponseWrapper.getResultatPakkePIH();
    if (resultatPakkePIH == null) {
      throw new InvokerCallException("ResultatPakkePIH is missing from SvarEksponeringsservice response");
    }
    return resultatPakkePIH;
  }

  private byte[] toByteArray(Object object) {
    try (StringWriter writer = new StringWriter()) {
      jaxbContext.createMarshaller().marshal(object, writer);
      return writer.toString().getBytes(StandardCharsets.UTF_8);
    } catch (IOException | JAXBException e) {
      throw new InvokerCallException(e);
    }
  }
}
