/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentprovider.invoker;

import javax.enterprise.context.RequestScoped;
import javax.inject.Inject;
import javax.interceptor.Interceptors;
import javax.xml.ws.Holder;
import javax.xml.ws.WebServiceException;

import com.cgi.svareksponering.ISvarEksponering;
import com.cgi.svareksponering.ResultatPakkePIH;

import dk.nsi.common.servicestatuscheck.ServiceStatusCheckInterceptor;
import dk.nsi.common.util.SLALoggingDataHandler;
import dk.nsi.dds.projects.sxa.common.configuration.SXAConfigKeys;
import dk.nsi.dds.projects.sxa.documentprovider.exceptions.InvokerCallException;
import dk.nsi.dds.projects.sxa.documentprovider.exceptions.InvokerCallTimedoutException;
import dk.nsi.dds.projects.sxa.documentprovider.util.SXAInvocationAttributes;
import dk.nsi.schema.privacy.pih._1.PrivacyInformationHeaderType;
import dk.nsi.util.Config;
import dk.sdsd.nsp.slalog.api.SLALogItem;

@RequestScoped
public class SXAInvoker {
  @Inject
  private ISvarEksponering service;
  @Inject
  private GetDataFromPeriodParametersFactory getDataFromPeriodParametersFactory;
  @Inject
  private SLALoggingDataHandler slaLoggingDataHandler;
  @Inject
  private Config config;

  @Interceptors(ServiceStatusCheckInterceptor.class)
  public SXAResponseWrapper invoke(SXAInvocationAttributes attributes) {
    SLALogItem slaLogItem = slaLoggingDataHandler.createSLALogItemFromTemplate("GetDataFromPeriod");
    String endpoint = "No endpoint found";
    try {    
      if (service == null) {
        throw new InvokerCallException("Der er ingen forbindelse til Svareksponeringsservicen.");
      }

      slaLoggingDataHandler.setMessageIdAndFlowId(slaLogItem, null);
      endpoint = config.getProperty(SXAConfigKeys.ENDPOINT_URL);
      slaLoggingDataHandler.setTarget(slaLogItem, "http://cgi.com/SvarEksponering/ISvarEksponering/GetDataFromPeriod", endpoint);
      return invokeService(attributes);
    } catch (WebServiceException we) {
      if (ServiceResponseChecker.isTimeoutOrServiceUnavailable(we)) {
        String message = "Timeout/unavailable service under invokering af svareksponeringsservicen";
        slaLogItem.setCallResultError(message);
        throw new InvokerCallTimedoutException("Timed out");
      } else if (ServiceResponseChecker.isServiceNotFound(we)) {
        String message = "Fejl i invokering af svareksponeringsservicen: Ingen service tilgængelig.";
        slaLogItem.setCallResultError(message);
        throw new InvokerCallException(message);
      }
      String message = "Fejl i invokering af svareksponeringsservicen";
      slaLogItem.setCallResultError(message);
      throw new InvokerCallException(we);
    } catch (InvokerCallException ex) {
      String message = "Fejl i invokering af svareksponeringsservicen: " + ex.getMessage();
      slaLogItem.setCallResultError(message);
      throw ex;
    } catch (Exception ex) {
      String message = "Fejl i invokering af svareksponeringsservicen";
      slaLogItem.setCallResultError(message);
      throw new InvokerCallException(ex);
    } finally {
      slaLoggingDataHandler.finalizeAndStore(slaLogItem);
    }
  }

  private SXAResponseWrapper invokeService(SXAInvocationAttributes attributes) {
    Holder<PrivacyInformationHeaderType> pihHeader = new Holder<>();
    ResultatPakkePIH resultatPakkePIH = service.getDataFromPeriod(getDataFromPeriodParametersFactory.buildRequest(attributes), pihHeader);
    return new SXAResponseWrapper(resultatPakkePIH, pihHeader.value);
  }
}
