/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentprovider.util;

import java.util.List;

import dk.nsi.dds.projects.sxa.common.configuration.SXAConfigKeys;
import org.apache.log4j.Logger;
import org.joda.time.DateTime;
import org.joda.time.DateTimeZone;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Code;

import com.cgi.svareksponering.Relationship;
import com.cgi.svareksponering.UserRole;

import dk.nsi.dds.projects.sxa.common.PatientIdDerivationException;
import dk.nsi.dds.projects.sxa.common.PersonIdInfoDerivation;
import dk.nsi.dds.projects.sxa.common.QueryDetails;
import dk.nsi.dds.projects.sxa.common.configuration.SXAMappedValue;
import dk.nsi.dds.projects.sxa.documentprovider.context.HttpHeaderContext;
import dk.nsi.util.Config;

public class SXAInvocationAttributesFactory {
	
	private static Logger log = Logger.getLogger(SXAInvocationAttributesFactory.class);

	private static final DateTime DEFAULT_START_TIME = new DateTime(1900, 1, 1, 0, 0, DateTimeZone.UTC);
	private static final DateTime DEFAULT_STOP_TIME = new DateTime(DateTimeZone.UTC).plusYears(1);

	public static SXAInvocationAttributes createInvocationAttributes(QueryDetails query, Config config, HttpHeaderContext httpHeaderContext) {
		SXAInvocationAttributes attributes = new SXAInvocationAttributes();
		setAttributesFromProperties(config, attributes);
		setAttributesFromContext(httpHeaderContext, attributes);
		setAttributesFromQuery(query, attributes);
		return attributes;
	}

	private static void setAttributesFromProperties(Config config, SXAInvocationAttributes attributes) {
		// Credentials
		attributes.setLoginUsername(config.getProperty(SXAConfigKeys.USERNAME));
		attributes.setLoginPassword(config.getProperty(SXAConfigKeys.PASSWORD));
		// Slutbruger
		attributes.setSlutBrugerIdentifikation(config.getProperty(SXAConfigKeys.SLUTBRUGER_IDENTIFIKATION));
		attributes.setSlutBrugerOrganisation(config.getProperty(SXAConfigKeys.SLUTBRUGER_ORGANISATION));
		attributes.setSlutBrugerCpr(config.getProperty(SXAConfigKeys.SLUTBRUGER_CPR));
		attributes.setSlutBrugerFornavn(config.getProperty(SXAConfigKeys.SLUTBRUGER_FORNAVN));
		attributes.setSlutBrugerEfternavn(config.getProperty(SXAConfigKeys.SLUTBRUGER_EFTERNAVN));
		attributes.setSlutBrugerStilling(config.getProperty(SXAConfigKeys.SLUTBRUGER_STILLING));
		attributes.setSlutBrugerEmail(config.getProperty(SXAConfigKeys.SLUTBRUGER_EMAIL));
		attributes.setSlutBrugerRolle(config.getProperty(SXAConfigKeys.SLUTBRUGER_ROLLE));
		attributes.setSlutBrugerAutorisationsKode(config.getProperty(SXAConfigKeys.SLUTBRUGER_AUTORISATIONSKODE));
		// Samtykker
		attributes.setSamtykkeKode(config.getProperty(SXAConfigKeys.SAMTYKKE_KODE));
		attributes.setSamtykkeTekst(config.getProperty(SXAConfigKeys.SAMTYKKE_TEKST));
	}
	
	private static void setAttributesFromQuery(QueryDetails query, SXAInvocationAttributes attributes) {
		attributes.setFraDato((query.getFrom() != null) ? query.getFrom() : getDateOfBirthOrDefaultStartTime(query.getPatientId()));
		attributes.setTilDato((query.getTo() != null) ? query.getTo() : getDefaultStopTime());
		attributes.setPatientCpr(query.getPatientId());
		setSpecialer(query.getPracticeSettingCodes(), query.getMappedPracticeSettingCodesList(), attributes);
	}

	private static DateTime getDateOfBirthOrDefaultStartTime(String patientId) {
		try {
			return PersonIdInfoDerivation.deriveDateOfBirth(patientId);
		} catch (PatientIdDerivationException e) {
			return DEFAULT_START_TIME;
		}
	}

	private static DateTime getDefaultStopTime() {
		return DEFAULT_STOP_TIME;
	}

	private static void setSpecialer(List<Code> practiceSettingCodes, List<SXAMappedValue> mappedPracticeSettingCodes, SXAInvocationAttributes attributes) {
		attributes.setSpecialer(SpecialtyMapper.mapToSpecialty(practiceSettingCodes, mappedPracticeSettingCodes));
	}
	
	private static void setAttributesFromContext(HttpHeaderContext ctx, SXAInvocationAttributes attributes) {
		if(ctx != null && ctx.getHttpHeaders() != null) {
			List<String> userRoles = ctx.getHttpHeaders().get("userrole");
			if(userRoles != null && userRoles.size() > 0) {
				if(userRoles.size() > 1) {
					log.warn("More than one userrole in HTTP header will only use first in list. userroles="+userRoles.toString());
				}
				String userRoleFromHeader = userRoles.get(0);
				try {
					UserRole userRole = UserRole.fromValue(userRoleFromHeader);
					attributes.setBrugerType(userRole);
				} catch(IllegalArgumentException e) {
					log.warn("Kunne ikke veksle HTTP header userrole til kendt rolle.", e);
				}
			}
			
			List<String> relationships = ctx.getHttpHeaders().get("relationship");
			if(relationships != null && relationships.size() > 0) {
				if(relationships.size() > 1) {
					log.warn("More than one relationship in HTTP header will only use first in list. relationships="+relationships.toString());
				}
				String relationshipFromHeader = relationships.get(0);
				try {
					Relationship relation = Relationship.fromValue(relationshipFromHeader);
					attributes.setRelation(relation);
				} catch(IllegalArgumentException e) {
					log.warn("Kunne ikke veksle HTTP header relationship til kendt relation.", e);
				}
			}
		}
	}
}
