package dk.nsi.dds.projects.sxa.documentprovider.audit;

import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.medcom.dgws._2006._04.dgws_1_0.Linking;
import dk.nsi.dds.projects.sxa.common.HsuidHeaderUtil;
import dk.nsi.hsuid._2016._08.hsuid_1_1.HsuidHeader;
import loginfo.healthcare._2017_03_01.*;
import org.joda.time.DateTime;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.runners.MockitoJUnitRunner;
import org.openehealth.ipf.commons.ihe.xds.core.ebxml.ebxml30.RetrieveDocumentSetRequestType;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;

import java.util.UUID;

import static org.junit.Assert.*;

@RunWith(MockitoJUnitRunner.class)
public class LogRecordBuilderTest {
    private LogRecordBuilder subject;

    @Mock
    private DateTimeProvider dateTimeProvider;

    @Before
    public void setup() {
        Mockito.when(dateTimeProvider.now()).thenReturn(DateTime.parse("2020-10-15T10:14:22.854+02:00"));

        subject = new LogRecordBuilder(dateTimeProvider);
    }

    @Test
    public void buildLogRecord_traceinfo_requestTime() throws DatatypeConfigurationException {
        // Arrange
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getCitizenHsuidHeader();
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        assertEquals(DatatypeFactory.newInstance().newXMLGregorianCalendar("2020-10-15T10:14:22.854+02:00"), logRecord.getTraceInfo().getRequestTime());
    }

    @Test
    public void buildLogRecord_traceinfo_sessionId() {
        // Arrange
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getCitizenHsuidHeader();
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // "11111111-1111-1111-1111-111111111111"
        // Assert
        assertEquals("00000000-0000-0000-0000-000000000000", logRecord.getTraceInfo().getSessionId());
    }

    @Test
    public void buildLogRecord_traceinfo_requestId() {
        // Arrange
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getCitizenHsuidHeader();
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        // The RequestId is generated - just check that it is a guid.
        try {
            UUID requestId = UUID.fromString(logRecord.getTraceInfo().getRequestId());
        }
        catch(IllegalArgumentException e) {
            fail();
        }
    }

    @Test
    public void buildLogRecord_traceinfo_route() {
        // Arrange
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getCitizenHsuidHeader();
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = getForwardedLogRecord();
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        assertEquals(4, logRecord.getTraceInfo().getRoute().getViaSystem().size());
        assertEquals(4, logRecord.getTraceInfo().getRoute().getViaSystem().get(3).getId());
        assertEquals("Dokumentdelingsservice", logRecord.getTraceInfo().getRoute().getViaSystem().get(3).getValue());
    }

    @Test
    public void buildLogRecord_clientIdentificationType_citizen() {
        // Arrange
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getCitizenHsuidHeader();
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        assertEquals(ClientIdentificationTypeEnumeration.CPR_NUMMER, logRecord.getRequestingClient().getClientIdentification().getClientIdentificationType());
        assertEquals("098765-4321", logRecord.getRequestingClient().getClientIdentification().getValue());
    }

    @Test
    public void buildLogRecord_clientIdentificationType_healthcareProfessional_sks() {
        // Arrange
        String organisationIdentifier = "987123";
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getSksHealthcareProfessionalHsuidHeader(organisationIdentifier);
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        assertEquals(ClientIdentificationTypeEnumeration.SKS, logRecord.getRequestingClient().getClientIdentification().getClientIdentificationType());
        assertEquals(organisationIdentifier, logRecord.getRequestingClient().getClientIdentification().getValue());
    }

    @Test
    public void buildLogRecord_clientIdentificationType_healthcareProfessional_sor() {
        // Arrange
        String organisationIdentifier = "440081000016006";
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getSorHealthcareProfessionalHsuidHeader(organisationIdentifier);
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        assertEquals(ClientIdentificationTypeEnumeration.SOR, logRecord.getRequestingClient().getClientIdentification().getClientIdentificationType());
        assertEquals(organisationIdentifier, logRecord.getRequestingClient().getClientIdentification().getValue());
    }

    @Test
    public void buildLogRecord_clientIdentificationType_healthcareProfessional_yder() {
        // Arrange
        String organisationIdentifier = "445566";
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getYderHealthcareProfessionalHsuidHeader(organisationIdentifier);
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        assertEquals(ClientIdentificationTypeEnumeration.YDERNUMMER, logRecord.getRequestingClient().getClientIdentification().getClientIdentificationType());
        assertEquals(organisationIdentifier, logRecord.getRequestingClient().getClientIdentification().getValue());
    }

    @Test
    public void buildLogRecord_requestingClient_clinicName() {
        // Arrange
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getCitizenHsuidHeader();
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        assertEquals("", logRecord.getRequestingClient().getClinicName());
    }

    @Test
    public void buildLogRecord_requestingClient_clinicDepartment() {
        // Arrange
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getCitizenHsuidHeader();
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        assertEquals("", logRecord.getRequestingClient().getClinicDepartment());
    }

    @Test
    public void buildLogRecord_requestingClient_userCprNumber() {
        // Arrange
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getCitizenHsuidHeader();
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        assertEquals("098765-4321", logRecord.getRequestingClient().getUserCprNumber());
    }

    @Test
    public void buildLogRecord_requestingClient_username() {
        // Arrange
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getCitizenHsuidHeader();
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        assertEquals("", logRecord.getRequestingClient().getUserName());
    }

    @Test
    public void buildLogRecord_requestingClient_userInitials() {
        // Arrange
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getCitizenHsuidHeader();
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        assertEquals("", logRecord.getRequestingClient().getUserInitials());
    }

    @Test
    public void buildLogRecord_requestingClient_behalfOfCprNumber() {
        // Arrange
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getBehalfOfCitizenHsuidHeader();
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        assertEquals("987654-3210", logRecord.getRequestingClient().getBehalfOfCprNumber());
    }

    @Test
    public void buildLogRecord_requestingClient_behalfOfCprNumber_none() {
        // Arrange
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getCitizenHsuidHeader();
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        assertNull(logRecord.getRequestingClient().getBehalfOfCprNumber());
    }

    @Test
    public void buildLogRecord_requestingClient_lookupType_citizen_lookupOwnData() {
        // Arrange
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getCitizenHsuidHeader();
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        assertEquals(LookupTypeEnumeration.BORGER ,logRecord.getRequestingClient().getLookUpType());
    }

    @Test
    public void buildLogRecord_requestingClient_lookupType_citizen_lookupChildData() {
        // Arrange
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getBehalfOfCitizenHsuidHeader();
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        assertEquals(LookupTypeEnumeration.FORÆLDER ,logRecord.getRequestingClient().getLookUpType());
    }

    @Test
    public void buildLogRecord_requestingClient_lookupType_healthcareProfessional_authorizationCodeSupplied() {
        // Arrange
        String organisationIdentifier = "123456789";
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getSksHealthcareProfessionalHsuidHeaderWithResponsibleAuthorizationCode(organisationIdentifier);
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        assertEquals(LookupTypeEnumeration.KLINIKER ,logRecord.getRequestingClient().getLookUpType());
    }

    @Test
    public void buildLogRecord_requestingClient_lookupType_healthcareProfessional_noAuthorizationCode() {
        // Arrange
        String organisationIdentifier = "123456789";
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getSksHealthcareProfessionalHsuidHeaderWithResponsibleAuthorizationCode(organisationIdentifier);
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        // NB - this might need to change - see the spec
        assertEquals(LookupTypeEnumeration.KLINIKER ,logRecord.getRequestingClient().getLookUpType());
    }

    @Test
    public void buildLogRecord_requestingClient_taskType() {
        // Arrange
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getCitizenHsuidHeader();
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        assertEquals("Opslag Dokumentdelingsservice", logRecord.getRequestingClient().getTaskType());
    }

    @Test
    public void buildLogRecord_patient_cprNumber() {
        // Arrange
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getCitizenHsuidHeader();
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        assertEquals("121212-1212", logRecord.getPatient().getCprNumber());
    }

    @Test
    public void buildLogRecord_patient_name() {
        // Arrange
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getCitizenHsuidHeader();
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        assertEquals("", logRecord.getPatient().getName());
    }

    @Test
    public void buildLogRecord_consent_code_consentOverrideSupplied() {
        // Arrange
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getConsentOverrideCitizenHsuidHeader();
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        assertEquals(LogRecordConstants.CONSENT_OVERRIDDEN_CODE, logRecord.getConsent().getCode());
        assertEquals(LogRecordConstants.CONSENT_OVERRIDDEN_TEXT, logRecord.getConsent().getText());
        assertEquals(LogRecordConstants.CONSENT_OVERRIDDEN_PRIVACY_OVERRIDE_CODE, logRecord.getConsent().getPrivacyOverrideCode());
    }

    @Test
    public void buildLogRecord_consent_code_consentOverrideMissing() {
        // Arrange
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getCitizenHsuidHeader();
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        assertEquals(LogRecordConstants.CONSENT_GRANTED_CODE, logRecord.getConsent().getCode());
        assertEquals(LogRecordConstants.CONSENT_GRANTED_TEXT, logRecord.getConsent().getText());
        assertEquals(LogRecordConstants.CONSENT_GRANTED_PRIVACY_OVERRIDE_CODE, logRecord.getConsent().getPrivacyOverrideCode());
    }

    @Test
    public void buildLogRecord_extendedRequestInfo_containsRequestBody() {
        // Arrange
        HsuidHeader hsuidHeader = HsuidHeaderUtil.getCitizenHsuidHeader();
        Header header = getMedcomHeader();
        LogRecord forwardedLogRecord = null;
        RetrieveDocumentSetRequestType request = getRequest();

        // Act
        LogRecord logRecord = subject.buildLogRecord(hsuidHeader, header, forwardedLogRecord, request);

        // Assert
        assertEquals("Text", logRecord.getExtendedRequestInfo().getFormat());
        // Verify that the value looks somewhat like a RetrieveDocumentSetRequest
        assertTrue(logRecord.getExtendedRequestInfo().getValue().contains("HomeCommunityId"));
        assertTrue(logRecord.getExtendedRequestInfo().getValue().contains("RepositoryUniqueId"));
        assertTrue(logRecord.getExtendedRequestInfo().getValue().contains("DocumentUniqueId"));
    }

    private Header getMedcomHeader() {
        Header header = new Header();

        Linking linking = new Linking();
        linking.setFlowID("00000000-0000-0000-0000-000000000000");
        linking.setMessageID("11111111-1111-1111-1111-111111111111");

        header.setLinking(linking);

        return header;
    }

    private LogRecord getForwardedLogRecord() {
        LogRecord logRecord = new LogRecord();

        RouteType route = new RouteType();
        route.getViaSystem().add(getViaSystem(1, "sundhed.dk"));
        route.getViaSystem().add(getViaSystem(2, "DokumentService"));
        route.getViaSystem().add(getViaSystem(3, "SvarEksponering PIH Webservice"));

        TraceInfoType traceInfo = new TraceInfoType();
        traceInfo.setRoute(route);

        logRecord.setTraceInfo(traceInfo);

        return logRecord;
    }

    private RouteType.ViaSystem getViaSystem(int id, String value) {
        RouteType.ViaSystem viaSystem = new RouteType.ViaSystem();

        viaSystem.setId(id);
        viaSystem.setValue(value);

        return viaSystem;
    }

    private RetrieveDocumentSetRequestType getRequest() {
        RetrieveDocumentSetRequestType request = new RetrieveDocumentSetRequestType();

        RetrieveDocumentSetRequestType.DocumentRequest documentRequest = new RetrieveDocumentSetRequestType.DocumentRequest();
        documentRequest.setHomeCommunityId("1.2.3.5.6");
        documentRequest.setRepositoryUniqueId("1.2.3.4.55.66");
        documentRequest.setDocumentUniqueId("1.2.3.4.55.66.12172240491026.20201015115813.20201015115913.0");
        request.getDocumentRequest().add(documentRequest);

        return request;
    }
}