/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentprovider.document;

import org.junit.Test;

import dk.nsi.dds.projects.sxa.common.InvalidDateException;
import dk.nsi.dds.projects.sxa.common.PatientIdDerivationException;
import org.junit.Assert;

public class PatientIdTest {
  @Test(expected = PatientIdDerivationException.class)
  public void testBogusPatientId() throws Exception {
    // Act
    PatientId.createPatientId("sfvsfvfv");
  }

  @Test(expected = PatientIdDerivationException.class)
  public void testEmptyPatientId() throws Exception {
    // Act
    PatientId.createPatientId("");
  }

  @Test
  public void testSpacesRemovalPatientId() throws Exception {
    // Arrange
    PatientId patientId = PatientId.createPatientId(" 1212121234 ");
    // Act
    assertCompletePatient(patientId, "12122012", PatientId.GENDER_FEMALE, false);
  }

  @Test(expected = PatientIdDerivationException.class)
  public void testNullPatientId() throws Exception {
    // Act
    PatientId.createPatientId(null);
  }

  @Test
  public void testGender() throws Exception {
    // Arrange
    int day = 12, month = 12, year = 15, controlciffer = 1234;
    // Act
    PatientId patientFemaleId = createPatientId(day, month, year, controlciffer);
    PatientId patientMaleId = createPatientId(day, month, year, ++controlciffer);
    // Assert
    assertCompletePatient(patientFemaleId, "12122015", PatientId.GENDER_FEMALE, false);
    assertCompletePatient(patientMaleId, "12122015", PatientId.GENDER_MALE, false);
  }

  /**
   * Test valid days for normal and alternative cprs
   * 
   * @throws Exception
   */
  @Test
  public void testValidDays() throws Exception {
    // Arrange
    int month = 12, year = 15, controlciffer = 1234;
    int[] validDays = new int[] { PatientId.NORMAL_DAY_RANGE_START, PatientId.NORMAL_DAY_RANGE_END,
      PatientId.ALTERNATIVE_DAY_RANGE_START, PatientId.ALTERNATIVE_DAY_RANGE_END };
    // Act
    for (int i = 0; i < validDays.length; i++) {
      PatientId paitentId = createPatientId(validDays[i], month, year, controlciffer);
      assertCompletePatient(paitentId, String.format("%02d122015", i > 1 ? validDays[i] - 60 : validDays[i]),
        PatientId.GENDER_FEMALE, i > 1);
    }
  }

  /**
   * Test invalid days for normal and alternative cprs
   * 
   * @throws Exception
   */
  @Test
  public void testInvalidDays() throws Exception {
    // Arrange
    int month = 12, year = 15, controlciffer = 1234;
    int[] invalidDays = new int[] { PatientId.NORMAL_DAY_RANGE_START - 1, PatientId.NORMAL_DAY_RANGE_END + 1,
      PatientId.ALTERNATIVE_DAY_RANGE_START - 1, PatientId.ALTERNATIVE_DAY_RANGE_END + 1 };
    // Act
    for (int i = 0; i < invalidDays.length; i++) {
      try {
        createPatientId(invalidDays[i], month, year, controlciffer);
        Assert.fail("No exception on invalid day: " + invalidDays[i]);
      } catch (PatientIdDerivationException pe) {
      }
    }
  }

  @Test
  public void testValidMonths() throws Exception {
    // Arrange
    int day = 12, year = 15, controlciffer = 1234;
    int[] validMonths = new int[] { PatientId.MONTH_RANGE_START, PatientId.MONTH_RANGE_END };
    // Act
    for (int i = 0; i < validMonths.length; i++) {
      PatientId paitentId = createPatientId(day, validMonths[i], year, controlciffer);
      Assert.assertEquals(validMonths[i], paitentId.deriveDateOfBirth().getMonthOfYear());
      assertCompletePatient(paitentId, String.format("12%02d2015", validMonths[i]), PatientId.GENDER_FEMALE, false);
    }
  }

  @Test
  public void testInvalidMonths() throws Exception {
    // Arrange
    int day = 12, year = 15, controlciffer = 1234;
    int[] validMonths = new int[] { PatientId.MONTH_RANGE_START - 1, PatientId.MONTH_RANGE_END + 1 };
    // Act
    for (int i = 0; i < validMonths.length; i++) {
      try {
        createPatientId(day, validMonths[i], year, controlciffer);
        Assert.fail("No exception on invalid month: " + validMonths[i]);
      } catch (PatientIdDerivationException pe) {
      }
    }
  }

  @Test
  public void testValidYears() throws Exception {
    // Arrange
    int day = 12, month = 6, controlciffer = 1234;
    int[] validYears = new int[] { PatientId.YEAR_RANGE_START, PatientId.YEAR_RANGE_END };
    int[] expectedYears = new int[] { 2000, 1999 };
    // Act
    for (int i = 0; i < validYears.length; i++) {
      PatientId paitentId = createPatientId(day, month, validYears[i], controlciffer);
      Assert.assertEquals(expectedYears[i], paitentId.deriveDateOfBirth().getYear());
      assertCompletePatient(paitentId, String.format("1206%d", expectedYears[i]), PatientId.GENDER_FEMALE, false);
    }
  }

  @Test
  public void testInvalidYears() throws Exception {
    // Arrange
    int day = 12, month = 6, controlciffer = 1234;
    int[] invalidYears = new int[] { PatientId.YEAR_RANGE_START - 1, PatientId.YEAR_RANGE_END + 1 };
    // Act
    for (int i = 0; i < invalidYears.length; i++) {
      try {
        createPatientId(day, month, invalidYears[i], controlciffer);
        Assert.fail("No exception on invalid year: " + invalidYears[i]);
      } catch (PatientIdDerivationException pe) {
      }
    }
  }

  private static PatientId createPatientId(int day, int month, int year, int controlciffer)
    throws PatientIdDerivationException {
    return PatientId.createPatientId(String.format("%02d", day) + String.format("%02d", month)
      + String.format("%02d", year) + Integer.toString(controlciffer));
  }

  private static void assertCompletePatient(PatientId patientId, String dateOfBirth, char gender, boolean isAlternative)
    throws InvalidDateException, PatientIdDerivationException {
    Assert.assertEquals(dateOfBirth, patientId.determineDateofBirthText());
    Assert.assertEquals(gender, patientId.determineGender());
    Assert.assertEquals(isAlternative, patientId.isAlternative());
  }
}
