/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentprovider.impl;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.enterprise.inject.Produces;
import javax.inject.Inject;
import javax.servlet.http.HttpServletRequest;
import javax.xml.ws.WebServiceContext;
import javax.xml.ws.handler.MessageContext;

import dk.nsi.dds.projects.sxa.common.configuration.SXAConfigKeys;
import org.apache.commons.io.IOUtils;
import org.jglue.cdiunit.CdiRunner;
import org.joda.time.DateTime;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.AssigningAuthority;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Code;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Identifiable;
import org.openehealth.ipf.commons.ihe.xds.core.requests.RetrieveDocument;
import org.openehealth.ipf.commons.ihe.xds.core.responses.ErrorCode;
import org.openehealth.ipf.commons.ihe.xds.core.responses.ErrorInfo;
import org.openehealth.ipf.commons.ihe.xds.core.responses.RetrievedDocument;

import dk.nsi.common.util.SLALogItemFactory;
import dk.nsi.dds.projects.sxa.common.PracticeSettingCodesConfiguration;
import dk.nsi.dds.projects.sxa.common.PracticeSettingCodesMappingConfiguration;
import dk.nsi.dds.projects.sxa.common.TestDataHelper;
import dk.nsi.dds.projects.sxa.common.util.DocumentUniqueIdUtil;
import dk.nsi.dds.projects.sxa.common.util.configloading.Log4jConfigurationLoader;
import dk.nsi.dds.projects.sxa.common.util.xmlhandling.XStreamFileException;
import dk.nsi.dds.projects.sxa.common.xds.impl.ValueEncoderDecoder;
import dk.nsi.dds.projects.sxa.documentprovider.DocumentsToRetrieve;
import dk.nsi.dds.projects.sxa.documentprovider.ErrorInfoBuilder;
import dk.nsi.dds.projects.sxa.documentprovider.RetrievedDocumentsResult;
import dk.nsi.dds.projects.sxa.documentprovider.context.HttpHeaderContext;
import dk.nsi.dds.projects.sxa.documentprovider.exceptions.InvokerCallException;
import dk.nsi.dds.projects.sxa.documentprovider.invoker.SXAInvokeHandler;
import dk.nsi.dds.projects.sxa.documentprovider.util.SXAInvocationAttributes;
import dk.nsi.dds.projects.sxa.documentprovider.util.TestConfig;
import dk.nsi.util.Config;

import org.junit.Assert;

@RunWith(CdiRunner.class)
public class DocumentRetrieverImplTest {
	@Produces
	@Mock
	private SXAInvokeHandler sxaInvokeHandler;
	@Produces
	@Mock
	private ProviderConfigurationHandler providerConfigurationHandler;
	@Produces
	@Mock
	private PracticeSettingCodesMappingConfiguration practiceSettingCodesMappingConfiguration;
	@Inject
	private DocumentRetrieverImpl documentRetriever;
//	@Mock
//	private DSSInvocationContext invocationContext;
	@Produces
	@Mock
	private SLALogItemFactory slaLogItemFactory;

	@Produces
	@Mock
	private WebServiceContext webServiceContext;

	@Produces
	@Mock
	private HttpHeaderContext httpHeaderContext;
	
	@Before
	public void setup() {
		Log4jConfigurationLoader.readLog4jConfigurationFile("log4j-sxa-documentprovider-ws.xml");
	}
	

	// Decoded patient id (from DocumentUniqueId) and patient id simulated from HSUID header
	// (patient id in RetrieveDocument) differ
	@Test
	public void retrieveDocuments_PatientIdDifferingFails() throws Exception {
		
		final String uniqueIdRoot = "1.2.3";
		final String homeCommunityId = "2.3";
		final String repositoryUniqueId = "3.4";
		
		// Arrange
		Identifiable patientId1 = new Identifiable("3333", new AssigningAuthority("patientIdAssigningAuthority"));
		Identifiable patientId2 = new Identifiable("4444", new AssigningAuthority("patientIdAssigningAuthority"));
				
		final List<String> uniqueId1 = createDocumentUniqueIds(1, uniqueIdRoot, patientId1, new ArrayList<>());
		final List<String> uniqueId2 = createDocumentUniqueIds(1, uniqueIdRoot, patientId2, new ArrayList<>());
		
		List<String> ids = new ArrayList<>();
		ids.addAll(uniqueId1);
		ids.addAll(uniqueId2);
		

		final List<String> errorCodeContextList = new ArrayList<>();
		errorCodeContextList.add(ErrorInfoBuilder.createCodeContextFormat(uniqueId1.get(0), homeCommunityId, repositoryUniqueId));
		errorCodeContextList.add(ErrorInfoBuilder.createCodeContextFormat(uniqueId2.get(0), homeCommunityId, repositoryUniqueId));
		
		
		DocumentsToRetrieve documentsToRetrieve = createDocumentMetadataList(homeCommunityId, repositoryUniqueId, ids);

		setupMocks(homeCommunityId, repositoryUniqueId, uniqueIdRoot);
		setupInterceptorMocks();

		// Act
		RetrievedDocumentsResult result = documentRetriever.retrieveDocuments(documentsToRetrieve);
		// Assert
		Assert.assertNotNull(result);
		Assert.assertTrue(result.getErrors().size() > 0);
		Assert.assertTrue(result.getErrors().stream().anyMatch(e -> e.getErrorCode().equals(ErrorCode.REPOSITORY_ERROR) && errorCodeContextList.contains(e.getCodeContext())));
	}
	
	private List<String> createDocumentUniqueIds(int idCount, String uniqueIdRoot, Identifiable patientId, List<Code> selectedPracticeCodes) {
		List<String> ids = new ArrayList<>();
		DateTime now = new DateTime();
		for (int i = 0; i < idCount; i++) {
			String id = DocumentUniqueIdUtil.createUniqueId(uniqueIdRoot, patientId, now.plusMinutes(i), now.plusMinutes(i + 1), selectedPracticeCodes,
					new TestDataHelper().getPracticeSettingCodes());
			ids.add(id);
		}
		return ids;
	}
	
	@Test
	public void retrieveDocuments_Null() throws Exception {
		setupInterceptorMocks();
		// Act
		RetrievedDocumentsResult result = documentRetriever.retrieveDocuments(null);
		
		// Assert
		Assert.assertNotNull(result);
		Assert.assertEquals(0, result.getDocuments().size());
		Assert.assertNotNull(result.getErrors());
		Assert.assertEquals(1, result.getErrors().size());
		
		ErrorInfo errorInfo = result.getErrors().get(0);
		Assert.assertEquals(ErrorCode.REPOSITORY_ERROR, errorInfo.getErrorCode());
	}
	
	@Test
	public void retrieveDocuments_EmptyListOfDocumentsToRetrieve() throws Exception {
		setupInterceptorMocks();
		// Act
		DocumentsToRetrieve documentsToRetrieve = new DocumentsToRetrieveImpl(Collections.emptyList());
		RetrievedDocumentsResult result = documentRetriever.retrieveDocuments(documentsToRetrieve);
		
		// Assert
		Assert.assertNotNull(result);
		Assert.assertEquals(0, result.getDocuments().size());
		Assert.assertNotNull(result.getErrors());
		Assert.assertEquals(1, result.getErrors().size());
		
		ErrorInfo errorInfo = result.getErrors().get(0);
		Assert.assertEquals(ErrorCode.REPOSITORY_ERROR, errorInfo.getErrorCode());
	}

	@Test
	public void retrieveDocuments_OnlyReachableDocumentsTest() throws Exception {
		// Arrange
		final int documentsCount = 5;
		String homeCommunityId = "home";
		String repositoryUniqueId = "repo";
		String uniqueIdRoot = "1.2.3";
		String id = "1.2.3.4";
		String patientIdText = "1234";
		byte[] data = new byte[0];
		Identifiable patientId = new Identifiable(patientIdText, new AssigningAuthority(id));
		setupMocks(homeCommunityId, repositoryUniqueId, uniqueIdRoot);
		setupInterceptorMocks();
		// Arrange Mocks
		List<String> ids = createDocumentUniqueIds(documentsCount, uniqueIdRoot, patientId, new ArrayList<>());
		DocumentsToRetrieveImpl documentReferenceList = createDocumentMetadataList(homeCommunityId, repositoryUniqueId, ids);
		setupLabInvokerMock(documentReferenceList, DocumentsToRetrieveImpl.EMPTY_LIST, data);
		
		// Act
		RetrievedDocumentsResult result = documentRetriever.retrieveDocuments(documentReferenceList);
		// Assert
		Assert.assertNotNull(result);
		Assert.assertEquals(documentReferenceList.getDocumentReferences().size(), result.getDocuments().size());
		Assert.assertNotNull(result.getErrors());
		Assert.assertTrue(result.getErrors().isEmpty());
		assertReachableDocuments(result, documentReferenceList.getDocumentReferences().size(), providerConfigurationHandler.readConfiguration(), ids, data);
	}

	private void setupMocks(String homeCommunityId, String repositoryUniqueId, String uniqueIdRoot) throws XStreamFileException {
		ProviderConfiguration configuration = new ProviderConfiguration(homeCommunityId, repositoryUniqueId, uniqueIdRoot);
		Mockito.reset(providerConfigurationHandler);
		Mockito.when(providerConfigurationHandler.readConfiguration()).thenReturn(configuration);
		PracticeSettingCodesConfiguration practiceSettingCodesConfiguration = new PracticeSettingCodesConfiguration();
		practiceSettingCodesConfiguration.setPracticeSettingCodes(new TestDataHelper().getPracticeSettingCodes());
		Mockito.when(practiceSettingCodesMappingConfiguration.getPracticeSettingCodesConfiguration()).thenReturn(practiceSettingCodesConfiguration);
	}

	private void setupInterceptorMocks() {
		HttpServletRequest request = Mockito.mock(HttpServletRequest.class);

		MessageContext messageContext = Mockito.mock(MessageContext.class);
		Mockito.when(messageContext.containsKey(Mockito.anyString())).thenReturn(true);
		Mockito.when(messageContext.get(Mockito.anyString())).thenReturn(request);

		Mockito.when(webServiceContext.getMessageContext()).thenReturn(messageContext);
	}



//	@SuppressWarnings("unused")
//	private List<String> createDocumentUniqueIdsWithPracticeSettingCodes(int idCount, String uniqueIdRoot, String patientId) {
//		List<String> ids = new ArrayList<String>();
//		DateTime now = new DateTime();
//		Code psc = new Code("408454008", new LocalizedString("Test"), "SOR Test");
//		for (int i = 0; i < idCount; i++) {
//			ids.add(new RootedId(uniqueIdRoot, new DocumentIdEncoderDecoder(patientId, now.plusMinutes(i), null, Arrays.asList(psc),
//					practiceSettingCodesMappingConfiguration.getPracticeSettingCodesConfiguration().getPracticeSettingCodes()).encode()).getId());
//		}
//		return ids;
//	}

	private DocumentsToRetrieveImpl createDocumentMetadataList(String homeCommunityId, String repositoryUniqueId, List<String> ids) {
		List<RetrieveDocument> DocumentReferenceList = new ArrayList<>();
		for (int i = 0; i < ids.size(); i++) {
			DocumentReferenceList.add(createMetadataDocument(repositoryUniqueId, homeCommunityId, ids.get(i)));
		}
		return new DocumentsToRetrieveImpl(DocumentReferenceList);
	}

	private RetrieveDocument createMetadataDocument(String repositoryUniqueId, String homeCommunityId, String documentUniqueId) {
		RetrieveDocument metadata = new RetrieveDocument();
		metadata.setDocumentUniqueId(documentUniqueId);
		metadata.setHomeCommunityId(homeCommunityId);
		metadata.setRepositoryUniqueId(repositoryUniqueId);
		return metadata;
	}

	private void setupLabInvokerMock(DocumentsToRetrieve documentReferenceList, DocumentsToRetrieve unreachableList, byte[] documentData) throws Exception {
		for (RetrieveDocument documentReference : documentReferenceList.getDocumentReferences()) {
			String documentId = documentReference.getDocumentUniqueId();
			Mockito.when(sxaInvokeHandler.getDocument(Mockito.eq(documentId), Mockito.any(SXAInvocationAttributes.class))).thenReturn(documentData);
		}
		for (RetrieveDocument documentReference : unreachableList.getDocumentReferences()) {
			String documentId = documentReference.getDocumentUniqueId();
			Mockito.when(sxaInvokeHandler.getDocument(Mockito.eq(documentId), Mockito.any(SXAInvocationAttributes.class))).thenThrow(new InvokerCallException(new Exception()));
		}
	}

	private void assertReachableDocuments(RetrievedDocumentsResult result, int expectedReachableDocumentsCount, ProviderConfiguration providerConfiguration, List<String> ids,
			byte[] documentData) throws IOException {
		Assert.assertNotNull(result);
		Assert.assertEquals(expectedReachableDocumentsCount, result.getDocuments().size());

		for (RetrievedDocument retrievedDocument : result.getDocuments()) {
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			int read = IOUtils.copy(retrievedDocument.getDataHandler().getDataSource().getInputStream(), baos);
			Assert.assertEquals(read, documentData.length);
			Assert.assertEquals(new String(documentData), new String(baos.toByteArray(), 0, read));
			Assert.assertTrue(ids.contains(retrievedDocument.getRequestData().getDocumentUniqueId()));
			Assert.assertEquals(providerConfiguration.getHomeCommunityId(), retrievedDocument.getRequestData().getHomeCommunityId());
			Assert.assertEquals("text/xml", retrievedDocument.getMimeType());
			Assert.assertEquals(providerConfiguration.getRepositoryUniqueId(), retrievedDocument.getRequestData().getRepositoryUniqueId());
		}
	}

	@Test
	public void retrieveDocuments_OnlyErrorsForDocumentsTest() throws Exception {
		// Arrange
		final int documentsCount = 5;
		String homeCommunityId = "1.2";
		String repositoryUniqueId = "2.3";
		String uniqueIdRoot = "1.2.3";
		byte[] data = new byte[0];
		Identifiable patientId = new Identifiable("1234", new AssigningAuthority("1.2.3.4"));
		setupMocks(homeCommunityId, repositoryUniqueId, uniqueIdRoot);
		setupInterceptorMocks();
		// Arrange Mocks
		List<String> ids = createDocumentUniqueIds(documentsCount, uniqueIdRoot, patientId, new ArrayList<>());
		DocumentsToRetrieveImpl documentReferenceList = createDocumentMetadataList(homeCommunityId, repositoryUniqueId, ids);
		setupLabInvokerMock(DocumentsToRetrieveImpl.EMPTY_LIST, documentReferenceList, data);

		// Act
		RetrievedDocumentsResult result = documentRetriever.retrieveDocuments(documentReferenceList);
		// Assert
		Assert.assertNotNull(result);
		Assert.assertNotNull(result.getErrors());
		Assert.assertEquals(documentReferenceList.getDocumentReferences().size(), result.getErrors().size());
		ProviderConfiguration providerConfiguration = providerConfigurationHandler.readConfiguration();
		for (ErrorInfo error : result.getErrors()) {
			assertDocumentIdInList(ids, ValueEncoderDecoder.extractDocumentUniqueId(error.getCodeContext()));
			Assert.assertEquals(providerConfiguration.getHomeCommunityId(), ValueEncoderDecoder.extractHomeCommunityId(error.getCodeContext()));
			Assert.assertEquals(providerConfiguration.getRepositoryUniqueId(), ValueEncoderDecoder.extractRepositoryUniqueId(error.getCodeContext()));
		}
	}

	private void assertDocumentIdInList(List<String> ids, String documentUniqueIdFromContext) {
		Assert.assertTrue(ids.stream().anyMatch(id -> id.compareTo(documentUniqueIdFromContext) == 0));
	}

	private void assertErrorsForDocuments(RetrievedDocumentsResult result, int expectedUnreachableDocumentsCount, ProviderConfiguration providerConfiguration, List<String> ids) {
		Assert.assertNotNull(result);
		Assert.assertNotNull(result.getErrors());
		Assert.assertEquals(expectedUnreachableDocumentsCount, result.getErrors().size());
		for (ErrorInfo error : result.getErrors()) {
			assertDocumentIdInList(ids, ValueEncoderDecoder.extractDocumentUniqueId(error.getCodeContext()));
			Assert.assertEquals(providerConfiguration.getHomeCommunityId(), ValueEncoderDecoder.extractHomeCommunityId(error.getCodeContext()));
			Assert.assertEquals(providerConfiguration.getRepositoryUniqueId(), ValueEncoderDecoder.extractRepositoryUniqueId(error.getCodeContext()));
		}
	}

	@Test
	public void retrieveDocuments_BothDocumentsTest() throws Exception {
		// Arrange
		final int documentsCount = 5;
		String homeCommunityId = "1.2";
		String repositoryUniqueId = "2.3";
		String uniqueIdRoot = "1.2.3";
		Identifiable patientId = new Identifiable("1234", new AssigningAuthority("1.2.3.4"));
		
		byte[] data = new byte[0];
		// Arrange Mocks
		// Half of ids are used for reachable.
		// Other half unreachable
		setupMocks(homeCommunityId, repositoryUniqueId, uniqueIdRoot);
		setupInterceptorMocks();
		List<String> ids = createDocumentUniqueIds(documentsCount * 2, uniqueIdRoot, patientId, new ArrayList<>());
		DocumentsToRetrieveImpl reachableDocumentReferenceList = createDocumentMetadataList(homeCommunityId, repositoryUniqueId, ids.subList(0, ids.size() / 2));
		DocumentsToRetrieveImpl unreachableDocumentReferenceList = createDocumentMetadataList(homeCommunityId, repositoryUniqueId, ids.subList(ids.size() / 2, ids.size()));
		setupLabInvokerMock(unreachableDocumentReferenceList, reachableDocumentReferenceList, data);
		
		List<RetrieveDocument> documentReferences  = new ArrayList<>();
		documentReferences.addAll(reachableDocumentReferenceList.getDocumentReferences());
		documentReferences.addAll(unreachableDocumentReferenceList.getDocumentReferences());
		DocumentsToRetrieveImpl documentsToRetrieve = new DocumentsToRetrieveImpl(documentReferences);
		
		// Act
		
		RetrievedDocumentsResult result = documentRetriever.retrieveDocuments(documentsToRetrieve);
		// Assert
		Assert.assertNotNull(result);
		ProviderConfiguration providerConfiguration = providerConfigurationHandler.readConfiguration();
		assertErrorsForDocuments(result, unreachableDocumentReferenceList.getDocumentReferences().size(), providerConfiguration, ids);
		assertReachableDocuments(result, reachableDocumentReferenceList.getDocumentReferences().size(), providerConfiguration, ids, data);
	}

//	/**
//	 * @param patientId
//	 * @return dummy DSSInvocationContext for testing
//	 */
//	private DSSInvocationContext createTestInvocationContext() {
//		return new DSSInvocationContext(getCitizenHsuidHeader(), new Security(), new Header());
//	}

//	/**
//	 * @return A valid {@link HsuidHeader} with default information, where the user is a citizen
//	 */
//	private static HsuidHeader getCitizenHsuidHeader() {
//		return getValidCitizenHsuidHeader("issuer", "0123456789", "System owner", "System name", "0.0", "Responsible org");
//	}

//	/**
//	 * @return A valid {@link HsuidHeader} filled in with the supplied information, where the user is a citizen
//	 */
//	private static HsuidHeader getValidCitizenHsuidHeader(String issuer, String cpr, String systemOwner, String systemName, String systemVersion, String responsibleOrg) {
//		List<Attribute> attributes = new ArrayList<Attribute>(6);
//		addBasicAttributes(true, cpr, systemOwner, systemName, systemVersion, responsibleOrg, attributes);
//		return HealthcareServiceUserIdentificationHeaderUtil.createHealthcareServiceUserIdentificationHeader(issuer, attributes);
//	}

//	private static void addBasicAttributes(boolean isCitizen, String cpr, String systemOwner, String systemName, String systemVersion, String responsibleOrg, List<Object> attributes) {
//		attributes.add(new UserTypeAttribute(isCitizen));
//		attributes.add(new ActingUserCivilRegistrationNumberAttribute("0987654321"));
//		attributes.add(new CitizenCivilRegistrationNumberAttribute(cpr));
//		attributes.add(new SystemVendorNameAttribute(systemOwner));
//		attributes.add(new SystemNameAttribute(systemName));
//		attributes.add(new SystemVersionAttribute(systemVersion));
//		attributes.add(new OperationsOrganisationNameAttribute(responsibleOrg));
//	}

	@Produces
	private Config createTestConfig() {
		TestConfig config = new TestConfig();
		config.set(SXAConfigKeys.PASSWORD.name(), "password");
		config.set(SXAConfigKeys.REQUEST_TIMEOUT_SECONDS.name(), "1");
		config.set(SXAConfigKeys.ROOT_ID.name(), "11.22.33");
		config.set(SXAConfigKeys.SAMTYKKE_KODE.name(), "samtykkekode");
		config.set(SXAConfigKeys.SAMTYKKE_TEKST.name(), "samtykketekst");
		config.set(SXAConfigKeys.SLUTBRUGER_AUTORISATIONSKODE.name(), "slutautorisationskode");
		config.set(SXAConfigKeys.SLUTBRUGER_CPR.name(), "slutcpr");
		config.set(SXAConfigKeys.SLUTBRUGER_EFTERNAVN.name(), "slutefternavn");
		config.set(SXAConfigKeys.SLUTBRUGER_EMAIL.name(), "slutemail");
		config.set(SXAConfigKeys.SLUTBRUGER_FORNAVN.name(), "slutfornavn");
		config.set(SXAConfigKeys.SLUTBRUGER_IDENTIFIKATION.name(), "slutidentifikation");
		config.set(SXAConfigKeys.SLUTBRUGER_ORGANISATION.name(), "slutorganisation");
		config.set(SXAConfigKeys.SLUTBRUGER_ROLLE.name(), "slutrolle");
		config.set(SXAConfigKeys.SLUTBRUGER_STILLING.name(), "slutstilling");
		return config;
	}
}
