/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentprovider.invoker;

import static org.mockito.Matchers.any;
import static org.mockito.Mockito.when;

import java.io.StringReader;
import java.nio.charset.StandardCharsets;

import javax.enterprise.inject.Produces;
import javax.inject.Inject;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.transform.stream.StreamSource;

import org.jglue.cdiunit.CdiRunner;
import org.junit.Assert;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.Mock;
import org.mockito.Mockito;

import com.cgi.svareksponering.ResultatPakkePIH;

import dk.nsi.dds.projects.sxa.documentprovider.exceptions.InvokerCallException;
import dk.nsi.dds.projects.sxa.documentprovider.exceptions.InvokerCallTimedoutException;
import dk.nsi.dds.projects.sxa.documentprovider.util.SXAInvocationAttributes;
import dk.nsi.privacy.pih.PIHHeaderBuilder;
import dk.nsi.schema.privacy.pih._1.PrivacyInformationHeaderType;
import dk.nsi.schema.privacy.piw._1.PrivacyInformationWrapperType;

@RunWith(CdiRunner.class)
public class SXAInvokeHandlerTest {
  
  @Produces
  @Mock
  private SXAInvoker sxaInvoker;
  
  @Produces
  @Mock
  private SXAInvocationAttributes sxaInvocationAttributes;
  
  @Inject
  private SXAInvokeHandler sxaInvokeHandler;
  
  @Inject
  private JAXBContext jaxbContext;
  
  private String requestdDocumentId = "";
  
  @Test
  public void getDocumentsTest() throws Exception {
    // Arrange
    when(sxaInvoker.invoke(any())).thenReturn(new SXAResponseWrapper(new ResultatPakkePIH(), new PIHHeaderBuilder().addSummary(1, 2).build()));
    // Act
    String response = bytesToString(sxaInvokeHandler.getDocument(requestdDocumentId, sxaInvocationAttributes));
    // Assert
   JAXBElement<PrivacyInformationWrapperType> piwElement = jaxbContext.createUnmarshaller().unmarshal(new StreamSource(new StringReader(response)), PrivacyInformationWrapperType.class);
   Assert.assertNotNull(piwElement);
   Assert.assertNotNull(piwElement.getValue());
   Assert.assertNotNull(piwElement.getValue().getDocument());
   Assert.assertNotNull(piwElement.getValue().getDocument().getValue());
   Assert.assertTrue(piwElement.getValue().getDocument().getValue().length > 0);
   Mockito.verify(sxaInvoker).invoke(any());
  }
  
  @Test(expected = InvokerCallException.class)
  public void getDocumentsInvokeReturningNull() throws Exception {
    // Arrange
    when(sxaInvoker.invoke(any())).thenReturn(null);
    // Act
    sxaInvokeHandler.getDocument(requestdDocumentId, sxaInvocationAttributes);
  }
  
  @Test(expected = InvokerCallException.class)
  public void getDocumentsInvokeThrowsInvokerCallException() throws Exception {
    // Arrange
    when(sxaInvoker.invoke(any())).thenThrow(new InvokerCallException("Test"));
    // Act
    sxaInvokeHandler.getDocument(requestdDocumentId, sxaInvocationAttributes);
  }
  
  @Test(expected = InvokerCallTimedoutException.class)
  public void getDocumentsInvokeThrowsInvokerCallTimedOutException() throws Exception {
    // Arrange
    when(sxaInvoker.invoke(any())).thenThrow(new InvokerCallTimedoutException("Test"));
    // Act
    sxaInvokeHandler.getDocument(requestdDocumentId, sxaInvocationAttributes);
  }
  
  @Test(expected = InvokerCallException.class)
  public void getDocumentsInvokeReturningMissingPIH() throws Exception {
    // Arrange
    when(sxaInvoker.invoke(any())).thenReturn(new SXAResponseWrapper(new ResultatPakkePIH(), null));
    // Act
    sxaInvokeHandler.getDocument(requestdDocumentId, sxaInvocationAttributes);
  }
  
  @Test(expected = InvokerCallException.class)
  public void getDocumentsInvokeReturningMissingResultatPakkePIH() throws Exception {
    // Arrange
    when(sxaInvoker.invoke(any())).thenReturn(new SXAResponseWrapper(null, new PIHHeaderBuilder().addSummary(1, 2).build()));
    // Act
    sxaInvokeHandler.getDocument(requestdDocumentId, sxaInvocationAttributes);
  }
  
  private static String bytesToString(byte[] document) {
    return new String(document, StandardCharsets.UTF_8);
  }

  @Produces
  public JAXBContext getJAXBContext() {
    try {
      return JAXBContext.newInstance(PrivacyInformationWrapperType.class, PrivacyInformationHeaderType.class, ResultatPakkePIH.class);
    } catch (JAXBException e) {
      throw new RuntimeException(e);
    }
  }
}
