/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentprovider.invoker;

import java.net.SocketTimeoutException;

import javax.enterprise.inject.Produces;
import javax.inject.Inject;
import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.namespace.QName;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPFactory;
import javax.xml.soap.SOAPFault;
import javax.xml.ws.WebServiceException;
import javax.xml.ws.soap.SOAPFaultException;

import org.jglue.cdiunit.CdiRunner;
import org.jglue.cdiunit.InRequestScope;
import org.joda.time.DateTime;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.Mock;
import org.mockito.Mockito;

import com.cgi.svareksponering.ISvarEksponering;
import com.cgi.svareksponering.Relationship;
import com.cgi.svareksponering.ResultatPakkePIH;
import com.cgi.svareksponering.Speciale;
import com.cgi.svareksponering.UserRole;

import dk.nsi.common.servicestatuscheck.ServiceStatusCheckBean;
import dk.nsi.common.util.SLALoggingDataHandler;
import dk.nsi.dds.projects.sxa.documentprovider.exceptions.InvokerCallException;
import dk.nsi.dds.projects.sxa.documentprovider.exceptions.InvokerCallTimedoutException;
import dk.nsi.dds.projects.sxa.documentprovider.util.SXAInvocationAttributes;
import dk.nsi.util.Config;
import dk.sdsd.nsp.slalog.api.SLALogItem;
import org.junit.Assert;

@InRequestScope
@RunWith(CdiRunner.class)
public class SXAInvokerTest {

	@Produces
	@Mock
	private ServiceStatusCheckBean serviceStatusCheckBean;
	@Produces
	@Mock
	private ISvarEksponering iSvarEksponering;
	@Produces
	@Mock
	private Config config;
	@Produces
	@Mock
	private SLALoggingDataHandler slaLoggingDataHandler;
	@Inject
	private SXAInvoker sxaInvoker;
	@Mock
	private SLALogItem slaLogItem;

	@Produces
	private GetDataFromPeriodParametersFactory createInstance() {
		try {
			return new GetDataFromPeriodParametersFactory(DatatypeFactory.newInstance());
		} catch (DatatypeConfigurationException e) {
			throw new RuntimeException(e);
		}
	}

	@Before
	public void setupMocks() {
		Mockito.when(slaLoggingDataHandler.createSLALogItemFromTemplate("GetDataFromPeriod")).thenReturn(slaLogItem);
	}

	@Test
	public void invokeSuccessTest() {
		// Arrange
		Mockito.when(iSvarEksponering.getDataFromPeriod(Mockito.any(), Mockito.any())).thenReturn(new ResultatPakkePIH());
		// Act
		SXAResponseWrapper wrapper = sxaInvoker.invoke(createDefaultInvocationAttributes());
		Assert.assertNotNull(wrapper);
		Assert.assertNull(wrapper.getPrivacyInformationHeaderType());
		Assert.assertNotNull(wrapper.getResultatPakkePIH());
	}

	@Test(expected = InvokerCallTimedoutException.class)
	public void invokeThrowsTimeoutTest() {
		// Arrange
		Mockito.when(iSvarEksponering.getDataFromPeriod(Mockito.any(), Mockito.any())).thenThrow(new WebServiceException(new SocketTimeoutException()));
		// Act
		sxaInvoker.invoke(createDefaultInvocationAttributes());
	}

	@Test(expected = InvokerCallException.class)
	public void invokeThrowsSoapFaultTest() throws SOAPException {
		// Arrange
		SOAPFault soapFault = SOAPFactory.newInstance().createFault("reasonText", new QName(""));
		Mockito.when(iSvarEksponering.getDataFromPeriod(Mockito.any(), Mockito.any())).thenThrow(new SOAPFaultException(soapFault));
		// Act
		sxaInvoker.invoke(createDefaultInvocationAttributes());
	}

	@Test(expected = InvokerCallException.class)
	public void invokeThrowsRuntimeExceptionTest() throws SOAPException {
		// Arrange
		Mockito.when(iSvarEksponering.getDataFromPeriod(Mockito.any(), Mockito.any())).thenThrow(new RuntimeException());
		// Act
		sxaInvoker.invoke(createDefaultInvocationAttributes());
	}

	private SXAInvocationAttributes createDefaultInvocationAttributes() {
		SXAInvocationAttributes attributes = new SXAInvocationAttributes();
		attributes.setBrugerType(UserRole.BORGER);
		attributes.setFraDato(new DateTime());
		attributes.setLoginPassword("password");
		attributes.setLoginUsername("username");
		attributes.setPatientCpr("patientcpr");
		attributes.setRelation(Relationship.SELF);
		attributes.setSamtykkeKode("samtykkekode");
		attributes.setSamtykkeTekst("samtykketekst");
		attributes.setSlutBrugerAutorisationsKode("slutautorisationskode");
		attributes.setSlutBrugerCpr("slutcpr");
		attributes.setSlutBrugerEfternavn("slutefternavn");
		attributes.setSlutBrugerEmail("slutemail");
		attributes.setSlutBrugerFornavn("slutfornavn");
		attributes.setSlutBrugerIdentifikation("slutidentifikation");
		attributes.setSlutBrugerOrganisation("slutorganisation");
		attributes.setSlutBrugerRolle("slutrolle");
		attributes.setSlutBrugerStilling("slutstilling");
		attributes.setSpecialer(new Speciale[] { Speciale.MIKROBIOLOGI });
		attributes.setTilDato(new DateTime());
		return attributes;
	}
}
