/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentprovider.util;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import org.joda.time.DateTime;
import org.junit.Before;
import org.junit.Test;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Code;

import com.cgi.svareksponering.Relationship;
import com.cgi.svareksponering.Speciale;
import com.cgi.svareksponering.UserRole;

import dk.nsi.dds.projects.sxa.common.DocumentIdEncoderDecoder;
import dk.nsi.dds.projects.sxa.common.InvalidSXAQueryParameters;
import dk.nsi.dds.projects.sxa.common.QueryDetails;
import dk.nsi.dds.projects.sxa.common.TestDataHelper;
import dk.nsi.dds.projects.sxa.common.configuration.SXAMappedValue;
import dk.nsi.dds.projects.sxa.common.xds.impl.ValueConverter;
import dk.nsi.dds.projects.sxa.documentprovider.context.HttpHeaderContext;

public class SXAInvocationAttributesTest {
	private static DateTime FROM = ValueConverter.convertFromDTM("20150101081500");
	private static DateTime TO = ValueConverter.convertFromDTM("20151201081500");
	private static String PATIENT_ID = "0123456789";
	private static TestDataHelper configTestData = new TestDataHelper();
	private static SXAMappedValue pathologyCode = configTestData.practiceSettingCodeForPathology;
	private static SXAMappedValue microbiologyCode = configTestData.practiceSettingCodeForMicrobiology;
	private static List<SXAMappedValue> validPracticeSettingCodesList = configTestData.getPracticeSettingCodes();
	private TestConfig config = new TestConfig();
	private HttpHeaderContext context;
	
	@Before
	public void setup() {
		context = new HttpHeaderContext();
	}

	@Test
	public void invocationAttributesFromQuery() throws Exception {
		final QueryDetails queryDetails = createTestQueryDetails(PATIENT_ID, FROM, TO, TestDataHelper.asCodeList(microbiologyCode, pathologyCode));
		SXAInvocationAttributes attributes = SXAInvocationAttributesFactory.createInvocationAttributes(queryDetails, config, context);
		assertEquals(FROM, attributes.getFraDato());
		assertEquals(TO, attributes.getTilDato());
		assertEquals(PATIENT_ID, attributes.getPatientCpr());
		assertNotNull(attributes.getSpecialer());
		assertEquals(2, attributes.getSpecialer().length);
		assertTrue(Arrays.asList(attributes.getSpecialer()).contains(Speciale.fromValue(configTestData.practiceSettingCodeForPathology.getMappedValue())));
		assertTrue(Arrays.asList(attributes.getSpecialer()).contains(Speciale.fromValue(configTestData.practiceSettingCodeForMicrobiology.getMappedValue())));
		assertFalse(Arrays.asList(attributes.getSpecialer()).contains(Speciale.fromValue(configTestData.practiceSettingCodeForBiochemistry.getMappedValue())));
	}

	@Test
	public void invocationAttributesFromNoSpecialties() throws Exception {
		final QueryDetails emptySpecialtiesQueryDetails = createTestQueryDetails(Collections.emptyList());
		SXAInvocationAttributes attributes = SXAInvocationAttributesFactory.createInvocationAttributes(emptySpecialtiesQueryDetails, config, context);
		assertNotNull(attributes.getSpecialer());
		assertEquals(3, attributes.getSpecialer().length);
		assertTrue(Arrays.asList(attributes.getSpecialer()).contains(Speciale.fromValue(configTestData.practiceSettingCodeForPathology.getMappedValue())));
		assertTrue(Arrays.asList(attributes.getSpecialer()).contains(Speciale.fromValue(configTestData.practiceSettingCodeForMicrobiology.getMappedValue())));
		assertTrue(Arrays.asList(attributes.getSpecialer()).contains(Speciale.fromValue(configTestData.practiceSettingCodeForBiochemistry.getMappedValue())));
	}
	
	public void addToContext(String key, String value) {
		Map<String, List<String>> httpHeaders = context.getHttpHeaders();

		List<String> values = httpHeaders.get(key);
		if(values == null) {
			values = new ArrayList<String>();
			httpHeaders.put(key, values);
		}
		
		values.add(value);
	}

	@Test
	public void invocationAttributesForCitizenSelf() throws Exception {
		
		addToContext("userrole", "Borger");
		addToContext("relationship", "Self");
		
		final QueryDetails createTestQueryDetails = createTestQueryDetails(configTestData.getPracticeSettingCodesAsCodeList());
		SXAInvocationAttributes attributes = SXAInvocationAttributesFactory.createInvocationAttributes(createTestQueryDetails, config, context);
		assertEquals(UserRole.BORGER, attributes.getBrugerType());
		assertEquals(Relationship.SELF, attributes.getRelation());
	}

	@Test
	public void invocationAttributesForParent() throws Exception {
		addToContext("userrole", "Borger");
		addToContext("relationship", "Parent");

		SXAInvocationAttributes attributes = SXAInvocationAttributesFactory.createInvocationAttributes(createTestQueryDetails(configTestData.getPracticeSettingCodesAsCodeList()), config, context);
		assertEquals(UserRole.BORGER, attributes.getBrugerType());
		assertEquals(Relationship.PARENT, attributes.getRelation());
	}

	@Test
	public void invocationAttributesForGuardian() throws Exception {
		addToContext("userrole", "Borger");
		addToContext("relationship", Relationship.GUARDIAN.value());

		SXAInvocationAttributes attributes = SXAInvocationAttributesFactory.createInvocationAttributes(createTestQueryDetails(configTestData.getPracticeSettingCodesAsCodeList()), config, context);
		assertEquals(UserRole.BORGER, attributes.getBrugerType());
		assertEquals(Relationship.GUARDIAN, attributes.getRelation());
	}

	private QueryDetails createTestQueryDetails(List<Code> practiceSettingCodesList) throws InvalidSXAQueryParameters {
		return createTestQueryDetails(PATIENT_ID, FROM, TO, practiceSettingCodesList);
	}

	private static QueryDetails createTestQueryDetails(String patientId, DateTime from, DateTime to, List<Code> practiceSettingCodesList) throws InvalidSXAQueryParameters {
		String encoded = new DocumentIdEncoderDecoder(patientId, from, to, practiceSettingCodesList, validPracticeSettingCodesList).encode();
		assertTrue(encoded.length() > 0);
		assertTrue(encoded.length() <= 47);
		return new QueryDetails(encoded, validPracticeSettingCodesList);
	}
}
