/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.common;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.joda.time.DateTime;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Code;

import dk.nsi.dds.projects.sxa.common.configuration.SXAMappedValue;
import dk.nsi.dds.projects.sxa.common.xds.impl.ValueConverter;

/**
 * Performs encoding and decoding of: mandatory patient id consisting of digits only optional from datetime optional to datetime The encoded string conforms to OID rules, consisting of digits and dots
 * only, where a non-present datetime is represented by noDateTimeValue
 */
public class DocumentIdEncoderDecoder {

	private static final int PATIENT_ID_INDEX = 1;
	private static final int DATETIME_FROM_INDEX = 2;
	private static final int DATETIME_TO_INDEX = 3;
	private static final int PRACTICE_SETTINGS_CODES_INDEX = 4;
	private static final String OID_REGEXP_STRING = "([\\d]+)\\.([\\d]+)\\.([\\d]+).([\\d]+)";
	private final static Pattern encodedPattern = Pattern.compile(OID_REGEXP_STRING);
	private static final String magic = "b";
	private static final String sep = ".";
	private static final int noDateTimeValue = 1;

	private String patientId;
	private DateTime from;
	private DateTime to;
	private List<Code> practiceSettingCodes = new ArrayList<>();

	private PracticeSettingCodesEncoderDecoder practiceSettingCodesEncoder;

	public DocumentIdEncoderDecoder(String patientId, DateTime from, DateTime to, List<Code> selectedPracticeSettingCodes, List<SXAMappedValue> validPracticeSettingCodes) {
		super();
		if (patientId == null || patientId.trim().isEmpty() || !patientId.matches("[0-9]+")) {
			throw new IllegalArgumentException("patientId must contain digits and digits only");
		}
		this.patientId = patientId;
		this.from = from;
		this.to = to;
		this.practiceSettingCodesEncoder = new PracticeSettingCodesEncoderDecoder(validPracticeSettingCodes);
		setPracticeSettingCodes(selectedPracticeSettingCodes);
	}

	public DocumentIdEncoderDecoder(String encoded, List<SXAMappedValue> validPracticeSettingCodesListPracticeSettingCodesList) {
		super();
		if (encoded == null) {
			throw new IllegalArgumentException("Empty encoded string");
		}
		Matcher matcher = encodedPattern.matcher(encoded);
		if (matcher.matches()) {
			this.patientId = decodeId(matcher.group(PATIENT_ID_INDEX));
			this.from = getDateTimeOrNull(matcher.group(DATETIME_FROM_INDEX));
			this.to = getDateTimeOrNull(matcher.group(DATETIME_TO_INDEX));
			this.practiceSettingCodesEncoder = new PracticeSettingCodesEncoderDecoder(validPracticeSettingCodesListPracticeSettingCodesList);
			this.practiceSettingCodes = practiceSettingCodesEncoder.decode(Integer.parseInt(matcher.group(PRACTICE_SETTINGS_CODES_INDEX)));
		} else {
			throw new IllegalArgumentException("Unexpected format");
		}
	}

	private DateTime getDateTimeOrNull(String value) {
		if (value == null || value.isEmpty()) {
			throw new IllegalArgumentException("Unexpected format");
		}
		if (value.equals(String.valueOf(noDateTimeValue))) {
			return null;
		}
		BigInteger result = new BigInteger(value).subtract(BigInteger.valueOf(noDateTimeValue));
		return ValueConverter.convertFromDTM(result.toString());
	}

	public String encode() {
		String result = encodeId(magic + patientId);
		result += sep + encodeDateTime(from);
		result += sep + encodeDateTime(to);
		result += sep + practiceSettingCodesEncoder.encode(practiceSettingCodes);
		return result;
	}

	private String encodeDateTime(DateTime dateTime) {
		if (dateTime != null) {
			return new BigInteger(ValueConverter.convertToDTM(dateTime)).add(BigInteger.valueOf(noDateTimeValue)).toString();
		}
		return String.valueOf(noDateTimeValue);
	}

	private String encodeId(String text) {
		return new BigInteger(text, 16).toString();
	}

	private String decodeId(String encoded) {
		String decoded = new BigInteger(encoded).toString(16);
		if (decoded == null || !decoded.toLowerCase().startsWith(magic.toLowerCase())) {
			throw new IllegalArgumentException();
		}
		return decoded.substring(magic.length());
	}

	public String getPatientId() {
		return patientId;
	}

	public DateTime getFrom() {
		return from;
	}

	public DateTime getTo() {
		return to;
	}

	public List<Code> getPracticeSettingCodes() {
		return this.practiceSettingCodes;
	}

	private void setPracticeSettingCodes(List<Code> codes) {
		if (codes != null) {
			for (Code code : codes) {
				if (!this.practiceSettingCodes.contains(code)) {
					this.practiceSettingCodes.add(code);
				}
			}
		}
	}
}
