/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.common;

import org.joda.time.DateTime;
import org.joda.time.DateTimeZone;

/**
 * Helper for deriving information from a person id. This is possible when the person id is
 * given as a Danish centralized person register (CPR, in Danish) identifier.  
 *
 */
public class PersonIdInfoDerivation {

    /**
     * Calculates a person's birthday from person id. Algorithm for determining birth year taoes two
     * first digits from current year, adds two digits from person id, and compares the resulting
     * year with current year. If resulting year is greater than current year, 100 years are
     * subtracted from the resulting year.
     * 
     * Example:
     * 	CurrentYear:	2015
     * 	PersonId:		010116
     * 	Output:			1916
     * 
     * 	CurrentYear:	2015
     * 	personId:		010115
     * 	Output:			2015
     * 
     * @throws PatientIdDerivationException
     */
    public static DateTime deriveDateOfBirth(String patientId) throws PatientIdDerivationException {
	assertValidPatientDate(patientId);
	try {
	    return calculateBirthdayFromCurrentYear(patientId);
	} catch (InvalidDateException e) {
	    throw new PatientIdDerivationException("Cannot derive date of birth from patient id", e);
	}
    }

    public static String determinePatientGender(String patientId) throws PatientIdDerivationException {
	assertValidPatientIdForm(patientId);
	try {
	    int lastControlDigit = Integer.parseInt(patientId.substring(patientId.length() -1));
	    if (lastControlDigit % 2 == 0) {
		return "F";
	    } else {
		return "M";
	    }
	} catch (Exception ex) {
	    // It is assumed ill-advised to add the id to error message as it might be viewable in logs
	    throw new PatientIdDerivationException("Could not determine gender from PatientId");
	}
    }

    /**
     * Verifies that given patient id has CPR number form. This must not be used as validation whether
     * the patient id identifies a real person or not.
     * @param patientId
     * @throws PatientIdDerivationException
     */
    public static void assertValidPatientIdForm(String patientId) throws PatientIdDerivationException {
	if (patientId == null) {
	    throw new PatientIdDerivationException("patientId is null");
	}
	if (patientId.length() != 10 || !patientId.matches("[0-9]+")) { // Ten char digits only
	    // It is assumed ill-advised to add the id to error message as it might be viewable in logs
	    throw new PatientIdDerivationException("Patient id does not adhere to CPR standard");
	}
    }

    /**
     * Verifies that given patient id has CPR number form and contains date information which
     * is sensible. The method must not be used as validation whether the patient id identifies a
     * real person or not.
     * @param patientId
     * @throws PatientIdDerivationException
     */
    public static void assertValidPatientDate(String patientId) throws PatientIdDerivationException {
	assertValidPatientIdForm(patientId);
	int day = 0;
	int month = 0;
	try {
	    day = Integer.parseInt(patientId.substring(0, 2));
	    month = Integer.parseInt(patientId.substring(2, 4));
	    Integer.parseInt(patientId.substring(4, 6)); // year sensible?
	} catch (Exception ex) {
	    throw new PatientIdDerivationException("Cannot derive sensible date from patient id");
	}
	if (day < 1 || day > 31 || month < 1 || month > 12) {
	    throw new PatientIdDerivationException("Day and/or month not of sensible values in patient id");
	}	    
    }


    /**
     * Calculates person birthday from person id. Algorithm for determining birth year takes two
     * first digits from current year, adds two digits from person id, and compares the resulting
     * year with current year. If resulting year is greater than current year, 100 years are
     * subtracted from the resulting year.
     * 
     * Example:
     * 	CurrentYear:	2015
     * 	PersonId:		010116
     * 	Output:			1916
     * 
     * 	CurrentYear:	2015
     * 	personId:		010115
     * 	Output:			2015
     * 
     * @param patientId
     * @return
     * @throws InvalidDateException 
     */
    private static DateTime calculateBirthdayFromCurrentYear(String patientId) throws InvalidDateException {
	try {
	    int day = Integer.parseInt(patientId.substring(0, 2));
	    int month = Integer.parseInt(patientId.substring(2, 4));
	    int year = Integer.parseInt(patientId.substring(4, 6));
	    DateTime now = new DateTime(DateTimeZone.UTC);
	    int currentYear = now.getYear();
	    String currentYearTekst = currentYear + "";
	    int proposedYear = Integer.parseInt(currentYearTekst.substring(0,
		    currentYearTekst.length() - 2) + String.format("%02d", year));
	    DateTime result = new DateTime(proposedYear, month, day, 0, 0, DateTimeZone.UTC);
	    if (result.isAfter(now)) {
		result = result.minusYears(100);
	    }
	    return result;
	} catch (Exception ex) {
	    throw new InvalidDateException(ex);
	}
    }
}
