/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.common;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

import org.openehealth.ipf.commons.ihe.xds.core.metadata.Code;

import dk.nsi.dds.projects.sxa.common.configuration.SXAMappedValue;

public class PracticeSettingCodesEncoderDecoder {
	private final int MAX_VALUES_IN_LIST = 32;
	private List<SXAMappedValue> validPracticeSettingCodes;

	public PracticeSettingCodesEncoderDecoder(List<SXAMappedValue> validPracticeSettingCodes) {
		if (validPracticeSettingCodes.size() > MAX_VALUES_IN_LIST) {
			throw new IllegalArgumentException("Too many values in list");
		}
		this.validPracticeSettingCodes = validPracticeSettingCodes;
	}

	public int encode(List<Code> codes) {
		int result = 0;
		int count = 0;
		for (Code code : codes) {
			int maskedValue = getMaskValueForCode(code);
			if (maskedValue > 0) {
				count++;
			}
			result += maskedValue;
		}
		if (count > MAX_VALUES_IN_LIST) {
			throw new IllegalArgumentException("Too many values in list");
		}
		return result;
	}

	public List<Code> decode(int code) {
		if (code < 0) {
			throw new IllegalArgumentException("Negative parameter");
		} else if (code >= Math.pow((double) 2, (double) validPracticeSettingCodes.size())) {
			throw new IllegalArgumentException("Value to decode is larger than validPracticeCodes can handle");
		}
		List<Code> result = new ArrayList<>();
		for (int i = 0; i < validPracticeSettingCodes.size(); i++) {
			int output = code & (int) Math.pow(2, (double) i);
			if (output > 0) {
				final SXAMappedValue valid = validPracticeSettingCodes.get(i);
				result.add(new Code(valid.getCode(), valid.getDisplayName(), valid.getSchemeName()));
			}
		}
		return result;
	}

	private int getMaskValueForCode(Code code) {
		final List<SXAMappedValue> matching = validPracticeSettingCodes.stream().filter(c -> c.getCode().equals(code.getCode()) && c.getSchemeName().equals(code.getSchemeName()))
				.collect(Collectors.toList());
		if (matching == null || matching.isEmpty()) {
			return 0;
		}
		double index = validPracticeSettingCodes.indexOf(matching.stream().findFirst().get());
		return (int) Math.pow((double) 2, (double) index);
	}
}
