package dk.nsi.dds.projects.sxa.common.security;

import dk.nsi.dds.projects.sxa.common.configuration.SXAConfigKeys;
import dk.nsi.util.Config;
import dk.nsi.util.ConfigException;
import org.apache.cxf.binding.soap.SoapFault;

import javax.inject.Inject;
import javax.interceptor.AroundInvoke;
import javax.interceptor.InvocationContext;
import javax.servlet.http.HttpServletRequest;
import javax.xml.namespace.QName;
import javax.xml.ws.WebServiceContext;
import javax.xml.ws.handler.MessageContext;

public class AccessValidationInterceptor {
    @Inject
    private WebServiceContext webServiceContext;

    @Inject
    private Config config;

    @AroundInvoke
    public Object validateAccess(InvocationContext context) throws Exception {
        String credentialHeaderName = getCredentialHeaderName();
        String serverCredential = getServerCredential();

        if(credentialHeaderName == null && serverCredential == null) {
            // Validation was not configured, so simply skip validation.
            return context.proceed();
        }
        if(credentialHeaderName == null || serverCredential == null) {
            throw new IllegalStateException("SXA was misconfigured, expected credentialHeader and serverCredential to be simultaneouly present or absent!");
        }

        checkCredential(serverCredential, credentialHeaderName);

        return context.proceed();
    }

    private String getCredentialHeaderName() {
        return getPropertyFromConfig(SXAConfigKeys.SXA_CREDENTIAL_HEADER);
    }

    private String getServerCredential() {
        return getPropertyFromConfig(SXAConfigKeys.SXA_CREDENTIAL);
    }

    private String getPropertyFromConfig(SXAConfigKeys configKey) {
        try {
            return config.getProperty(configKey);
        }
        catch(ConfigException e) {
            return null;
        }
    }

    private void checkCredential(String serverCredential, String credentialHeaderName) {
        String clientCredential = getCredentialHeaderValueFromRequest(credentialHeaderName);
        if(!serverCredential.equals(clientCredential)) {
            throwSoapFault(String.format("Incorrect %s header provided!", credentialHeaderName));
        }
    }

    private String getCredentialHeaderValueFromRequest(String credentialHeaderName) {
        HttpServletRequest request = getHttpRequest();

        String headerValue = request.getHeader(credentialHeaderName);
        if(headerValue == null || headerValue.isEmpty()) {
            throwSoapFault(String.format("No header of type %s was present!", credentialHeaderName));
        }

        return headerValue;
    }

    private HttpServletRequest getHttpRequest() {
        MessageContext messageContext = extractMessageContext();
        return extractHttpRequest(messageContext);
    }

    private MessageContext extractMessageContext() {
        if(webServiceContext == null) {
            throw new IllegalStateException("webServiceContext was null!");
        }

        MessageContext messageContext = webServiceContext.getMessageContext();
        if(messageContext == null) {
            throw new IllegalStateException("No MessageContext was present on webServiceContext!");
        }

        return messageContext;
    }

    private HttpServletRequest extractHttpRequest(MessageContext messageContext) {
        if(!messageContext.containsKey("HTTP.REQUEST") || !(messageContext.get("HTTP.REQUEST") instanceof HttpServletRequest)) {
            throw new IllegalStateException("Could not extract http request!");
        }

        HttpServletRequest request = (HttpServletRequest)messageContext.get("HTTP.REQUEST");
        return request;
    }

    private void throwSoapFault(String message) {
        throw new SoapFault(message, new QName("http://sxa", "bad_request"));
    }
}
