/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (SDS), Denmark
 * (http://sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (SDS), Denmark (http://sundhedsdatastyrelsen.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.common.xds.impl;

import java.util.ArrayList;
import java.util.List;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.joda.time.DateTime;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.AvailabilityStatus;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Code;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.DocumentEntryType;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Identifiable;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.TimeRange;
import org.openehealth.ipf.commons.ihe.xds.core.requests.query.FindDocumentsQuery;
import org.openehealth.ipf.commons.ihe.xds.core.requests.query.GetDocumentsQuery;
import org.openehealth.ipf.commons.ihe.xds.core.requests.query.QueryList;
import org.openehealth.ipf.commons.ihe.xds.core.requests.query.QueryReturnType;
import org.openehealth.ipf.commons.ihe.xds.core.requests.query.QueryType;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.query.AdhocQueryRequest;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.query.ResponseOptionType;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.rim.AdhocQueryType;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.rim.SlotType1;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.rim.ValueListType;
import org.openehealth.ipf.commons.ihe.xds.core.transform.requests.QueryParameter;

import dk.nsi.dds.projects.sxa.common.core.query.QueryParameterFormatter;
import dk.nsi.dds.projects.sxa.common.xds.UnexpectedDataException;



public class AdhocQueryRequestBuilder {
	private org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.rim.ObjectFactory rimObjFactory;
	private org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.query.ObjectFactory queryObjFactory;

	public AdhocQueryRequestBuilder() {
		rimObjFactory = new org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.rim.ObjectFactory();
		queryObjFactory = new org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.query.ObjectFactory();
	}

	public AdhocQueryRequest createRegistryStoredQueryFindDocumentsRequest(FindDocumentsQuery query) throws UnexpectedDataException {
		AdhocQueryRequest request = createAdhocQueryRequest(QueryType.FIND_DOCUMENTS);
		addParameters(request, query);
		return request;
	}

	public AdhocQueryRequest createRegistryStoredQueryGetDocumentsRequest(GetDocumentsQuery query) throws UnexpectedDataException {
		AdhocQueryRequest request = createAdhocQueryRequest(QueryType.GET_DOCUMENTS);
		addParameters(request, query);
		return request;
	}

	private AdhocQueryRequest createAdhocQueryRequest(QueryType queryType) {
		AdhocQueryRequest adhocQueryRequest = queryObjFactory.createAdhocQueryRequest();
		ResponseOptionType responseOption = queryObjFactory.createResponseOptionType();
		responseOption.setReturnComposedObjects(true);
		responseOption.setReturnType(QueryReturnType.LEAF_CLASS.getCode());
		adhocQueryRequest.setResponseOption(responseOption);
		AdhocQueryType adhocQueryType = rimObjFactory.createAdhocQueryType();
		adhocQueryType.setId(queryType.getId());
		adhocQueryRequest.setAdhocQuery(adhocQueryType);
		return adhocQueryRequest;
	}

	private void addParameters(AdhocQueryRequest request, GetDocumentsQuery query) throws UnexpectedDataException {
		if (CollectionUtils.isEmpty(query.getUuids()) && CollectionUtils.isEmpty(query.getUniqueIds())) {
			throw new UnexpectedDataException("GetDocumentsQuery must have either document unique ids or entry uuids");
		} else if (CollectionUtils.isNotEmpty(query.getUuids()) && CollectionUtils.isNotEmpty(query.getUniqueIds())) {
			throw new UnexpectedDataException("Both document unique id and entry uuid cannot have values in a GetDocumentsQuery");
		}

		// Required parameters
		addQueryParameterSlotValues(request, QueryParameter.DOC_ENTRY_UNIQUE_ID, query.getUniqueIds());
		addQueryParameterSlotValues(request, QueryParameter.DOC_ENTRY_UUID, query.getUuids());
		// Optional parameters
		if(StringUtils.isNotEmpty(query.getHomeCommunityId())) {
			request.getAdhocQuery().setHome(query.getHomeCommunityId());
		}
	}

	private void addParameters(AdhocQueryRequest request, FindDocumentsQuery query) throws UnexpectedDataException {
		// Required parameters
		addDocumentEntryPatientIdQueryParameter(request, query.getPatientId());
		addDocumentEntryStatusQueryParameters(request, query.getStatus());
		// Optional parameters
		addDocumentEntryClassCodeQueryParameters(request, query);
		addDocumentEntryTypeCodeQueryParameters(request, query);
		addDocumentEntryPracticeSettingCodeQueryParameters(request, query);
		addDocumentEntryAllTimeQueryParameters(request, query);
		addDocumentEntryHealthcareFacilityTypeCodeQueryParameters(request, query);
		addDocumentEntryEventTypeCodeQueryParameters(request, query);
		addDocumentEntryConfidentialityCodeQueryParameters(request, query);
		addDocumentEntryAuthorPersonQueryParameters(request, query);
		addDocumentEntryFormatCodeQueryParameters(request, query);
		addDocumentEntryTypeQueryParameters(request, query.getDocumentEntryTypes());
	}

	private void addDocumentEntryPatientIdQueryParameter(AdhocQueryRequest request, Identifiable patientId) throws UnexpectedDataException {
		request.getAdhocQuery().getSlot().add(createQueryParameterSlot(QueryParameter.DOC_ENTRY_PATIENT_ID, ValueEncoderDecoder.encodePatientId(patientId)));
	}

	private void addDocumentEntryStatusQueryParameters(AdhocQueryRequest request, List<AvailabilityStatus> statuses) {
		addStatusQueryParameterSlot(request, QueryParameter.DOC_ENTRY_STATUS, statuses);
	}

	private void addDocumentEntryClassCodeQueryParameters(AdhocQueryRequest request, FindDocumentsQuery query) {
		addCodeQueryParameterSlot(request, QueryParameter.DOC_ENTRY_CLASS_CODE, query.getClassCodes());
	}

	private void addDocumentEntryTypeCodeQueryParameters(AdhocQueryRequest request, FindDocumentsQuery query) {
		addCodeQueryParameterSlot(request, QueryParameter.DOC_ENTRY_TYPE_CODE, query.getTypeCodes());
	}

	private void addDocumentEntryPracticeSettingCodeQueryParameters(AdhocQueryRequest request, FindDocumentsQuery query) {
		addCodeQueryParameterSlot(request, QueryParameter.DOC_ENTRY_PRACTICE_SETTING_CODE, query.getPracticeSettingCodes());
	}

	private void addDocumentEntryAllTimeQueryParameters(AdhocQueryRequest request, FindDocumentsQuery query) {
		if (query.getCreationTime() != null) {
			addTimeIntervalQueryParameter(request, query.getCreationTime(), QueryParameter.DOC_ENTRY_CREATION_TIME_FROM, QueryParameter.DOC_ENTRY_CREATION_TIME_TO);
		}
		if (query.getServiceStartTime() != null) {
			addTimeIntervalQueryParameter(request, query.getServiceStartTime(), QueryParameter.DOC_ENTRY_SERVICE_START_TIME_FROM, QueryParameter.DOC_ENTRY_SERVICE_START_TIME_TO);
		}
		if (query.getServiceStopTime() != null) {
			addTimeIntervalQueryParameter(request, query.getServiceStopTime(), QueryParameter.DOC_ENTRY_SERVICE_STOP_TIME_FROM, QueryParameter.DOC_ENTRY_SERVICE_STOP_TIME_TO);
		}
	}

	private void addDocumentEntryHealthcareFacilityTypeCodeQueryParameters(AdhocQueryRequest request, FindDocumentsQuery query) {
		addCodeQueryParameterSlot(request, QueryParameter.DOC_ENTRY_HEALTHCARE_FACILITY_TYPE_CODE, query.getHealthcareFacilityTypeCodes());
	}

	private void addDocumentEntryEventTypeCodeQueryParameters(AdhocQueryRequest request, FindDocumentsQuery query) {
		addQueryParameterSlot(request, QueryParameter.DOC_ENTRY_EVENT_CODE, query.getEventCodes());
	}

	private void addDocumentEntryConfidentialityCodeQueryParameters(AdhocQueryRequest request, FindDocumentsQuery query) {
		addQueryParameterSlot(request, QueryParameter.DOC_ENTRY_CONFIDENTIALITY_CODE, query.getConfidentialityCodes());
	}

	private void addDocumentEntryAuthorPersonQueryParameters(AdhocQueryRequest request, FindDocumentsQuery query) {
		addQueryParameterSlotValues(request, QueryParameter.DOC_ENTRY_AUTHOR_PERSON, query.getAuthorPersons());
	}

	private void addDocumentEntryFormatCodeQueryParameters(AdhocQueryRequest request, FindDocumentsQuery query) {
		addCodeQueryParameterSlot(request, QueryParameter.DOC_ENTRY_FORMAT_CODE, query.getFormatCodes());
	}

	private void addDocumentEntryTypeQueryParameters(AdhocQueryRequest request, List<DocumentEntryType> entryTypes) {
		List<String> values = new ArrayList<String>();

		for (DocumentEntryType entryType : entryTypes) {
			values.add(entryType.getUuid());
		}

		addQueryParameterSlotValues(request, QueryParameter.DOC_ENTRY_TYPE, values);
	}

	private void addQueryParameterSlotValues(AdhocQueryRequest request, QueryParameter slotName, List<String> values) {
		if (CollectionUtils.isNotEmpty(values)) {
			request.getAdhocQuery().getSlot().add(createQueryParameterSlot(slotName, values));
		}
	}

	private SlotType1 createQueryParameterSlot(QueryParameter name, List<String> values) {
		SlotType1 slot = createSlot(name);
		slot.setValueList(createValueList(values));
		return slot;
	}

	private SlotType1 createSlot(QueryParameter name) {
		SlotType1 slot = rimObjFactory.createSlotType1();
		slot.setName(name.getSlotName());
		return slot;
	}

	// Generic helpers
	private void addCodeQueryParameterSlot(AdhocQueryRequest request, QueryParameter queryParameter, List<Code> list) {
		if (CollectionUtils.isNotEmpty(list)) {
			List<String> values = new ArrayList<String>();

			for (Code codedValue : list) {
				values.add(formatCodedValue(codedValue));
			}
			addQueryParameterSlotValues(request, queryParameter, values);
		}
	}

	private String formatCodedValue(Code codedValue) {
		// Description field intentionally left out
		return codedValue.getCode() + "^^" + codedValue.getSchemeName();
	}

	private void addStatusQueryParameterSlot(AdhocQueryRequest request, QueryParameter queryParameter, List<AvailabilityStatus> statuses) {
		if (CollectionUtils.isNotEmpty(statuses)) {
			List<String> values = new ArrayList<String>();

			for (AvailabilityStatus status : statuses) {
				values.add(status.getQueryOpcode());
			}
			addQueryParameterSlotValues(request, queryParameter, values);
		}
	}

	private void addQueryParameterSlot(AdhocQueryRequest request, QueryParameter queryParam, QueryList<Code> queryList) {
		if (queryList != null) {
			List<List<Code>> outerList = queryList.getOuterList();
			if (outerList != null) {
				for (List<Code> list : outerList) {
					addCodeQueryParameterSlot(request, queryParam, list);
				}
			}
		}
	}

	private void addTimeIntervalQueryParameter(AdhocQueryRequest request, TimeRange timeInterval, QueryParameter fromParamName, QueryParameter toParamName) {
		if (timeInterval.getFrom() != null) {
			addDateTimeQueryParameter(request, timeInterval.getFrom(), fromParamName);
		}
		if (timeInterval.getTo() != null) {
			addDateTimeQueryParameter(request, timeInterval.getTo(), toParamName);
		}
	}

	private void addDateTimeQueryParameter(AdhocQueryRequest request, DateTime dateTime, QueryParameter queryParameter) {
		request.getAdhocQuery().getSlot().add(createQueryParameterSlot(queryParameter, ValueConverter.convertToDTM(dateTime)));
	}

	private SlotType1 createQueryParameterSlot(QueryParameter queryParameter, String value) {
		SlotType1 slot = createSlot(queryParameter);
		ValueListType valueListType = rimObjFactory.createValueListType();
		valueListType.getValue().add(QueryParameterFormatter.createSingleValuedQueryParameter(value));
		slot.setValueList(valueListType);
		return slot;
	}

	private ValueListType createValueList(List<String> values) {
		ValueListType valueListType = rimObjFactory.createValueListType();
		for (String value : values) {
			valueListType.getValue().add(QueryParameterFormatter.createMultiValuedQueryParameter(value));
		}
		return valueListType;
	}

}
