/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (SDS), Denmark
 * (http://sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (SDS), Denmark (http://sundhedsdatastyrelsen.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.common.xds.impl;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.openehealth.ipf.commons.ihe.xds.core.metadata.AssigningAuthority;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Hl7v2Based;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Identifiable;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Organization;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Person;

import dk.nsi.dds.projects.sxa.common.xds.UnexpectedDataException;
import dk.nsi.dds.projects.sxa.common.xds.UnexpectedEncodingException;
import dk.nsi.dds.projects.sxa.common.xds.constants.XdsConstants;

public class ValueEncoderDecoder {
	private final static Pattern repositoryUniqueIdPattern = Pattern.compile("(.*" + XdsConstants.XDS_REGISTRYERROR_CODECONTEXT_REPOSITORY_UNIQUE_ID + ")([\\d\\.\\d]+)(.*)");
	private final static Pattern homeCommunityIdPattern = Pattern.compile("(.*" + XdsConstants.XDS_REGISTRYERROR_CODECONTEXT_HOME_COMMUNITY_ID + ")([\\d\\.\\d]+)(.*)");
	private final static Pattern documentUniqueIdPattern = Pattern.compile("(.*" + XdsConstants.XDS_REGISTRYERROR_CODECONTEXT_DOCUMENT_UNIQUE_ID + ")([^\\s\\,]*)(.*)");
	// Previous: <Organisationsnavn>^^^^^&<Assigning authority>&ISO^^^^<org type>=<Organisation id>
	public static Pattern oldOrganisationPattern = Pattern.compile("(.*)" // org. name
			+ "\\^\\^\\^\\^\\^&([\\d\\.\\d]+)&ISO" // Assigning authority
			+ "\\^\\^\\^\\^(.*)=(.*)"); // org. id

	public static Identifiable decodePatientId(String encodedPatientId) {
		if (encodedPatientId == null) {
			throw new IllegalArgumentException("encodedPatientId missing");
		}
		return Hl7v2Based.parse(encodedPatientId, Identifiable.class);
	}

	public static String encodePatientId(Identifiable patientId) throws UnexpectedDataException {
		return Hl7v2Based.render(patientId);
	}

	public static Person decodePerson(String encodedPerson) {
		if (encodedPerson == null) {
			throw new IllegalArgumentException("encodedPerson missing");
		}
		return Hl7v2Based.parse(encodedPerson, Person.class);
	}

	public static String encodePerson(Person person) throws UnexpectedDataException {
		if (person == null) {
			throw new UnexpectedDataException("Values missing for person");
		}
		return Hl7v2Based.render(person);
	}

	public static Organization decodeOrganisation(String encodedOrganisation) {
		if (encodedOrganisation == null) {
			throw new IllegalArgumentException("encodedOrganisation missing");
		}
		return Hl7v2Based.parse(encodedOrganisation, Organization.class);
	}

	public static Organization decodeOrganisationOldEncoding(String encodedOrganisation) throws UnexpectedEncodingException {
		if (encodedOrganisation == null) {
			throw new IllegalArgumentException("encodedOrganisation missing");
		}
		Matcher matcher = oldOrganisationPattern.matcher(encodedOrganisation);
		if (!matcher.matches()) {
			throw new UnexpectedEncodingException("Organisation not encoded in accepted formatting");
		}
		String assigningAuthority = matcher.group(2) + "." + matcher.group(3);
		return new Organization(matcher.group(1), matcher.group(4), new AssigningAuthority(assigningAuthority, null));
	}

	public static String encodeOrganisation(Organization organisation) throws UnexpectedDataException {
		if (organisation == null) {
			throw new UnexpectedDataException("Values missing for organisation");
		}
		return Hl7v2Based.render(organisation);
	}

	public static String encodeErrorCodeContext(String documentUniqueId, String homeCommunityId, String repositoryUniqueId) {
		return String.format("[%s, %s, %s]", 
				XdsConstants.XDS_REGISTRYERROR_CODECONTEXT_DOCUMENT_UNIQUE_ID + documentUniqueId, 
				XdsConstants.XDS_REGISTRYERROR_CODECONTEXT_HOME_COMMUNITY_ID + homeCommunityId, 
				XdsConstants.XDS_REGISTRYERROR_CODECONTEXT_REPOSITORY_UNIQUE_ID + repositoryUniqueId);
	}

	public static String extractRepositoryUniqueId(String errorCodeContext) {
		Matcher matcher = repositoryUniqueIdPattern.matcher(errorCodeContext);
		if (matcher.matches()) {
			return matcher.group(2);
		}
		return null;
	}

	public static String extractHomeCommunityId(String errorCodeContext) {
		Matcher matcher = homeCommunityIdPattern.matcher(errorCodeContext);
		if (matcher.matches()) {
			return matcher.group(2);
		}
		return null;
	}

	public static String extractDocumentUniqueId(String errorCodeContext) {
		Matcher matcher = documentUniqueIdPattern.matcher(errorCodeContext);
		if (matcher.matches()) {
			return matcher.group(2);
		}
		return null;
	}
}
