/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.common;

import java.io.UnsupportedEncodingException;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import org.joda.time.DateTime;
import org.junit.BeforeClass;
import org.junit.Test;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Code;

import dk.nsi.dds.projects.sxa.common.configuration.SXAMappedValue;
import dk.nsi.dds.projects.sxa.common.util.ConverterUtil;
import dk.nsi.dds.projects.sxa.common.xds.impl.ValueConverter;
import org.junit.Assert;

public class DocumentIdEncoderDecoderTest {

  private String patientId = "2512484916";
  private DateTime from = ValueConverter.convertFromDTM("20150101081500");
  private DateTime to = ValueConverter.convertFromDTM("20151201081500");
  
  private static TestDataHelper testDataHelper = new TestDataHelper();
  private static List<SXAMappedValue> validPracticeSettingCodesList = Arrays.asList();
  
  @BeforeClass
  public static void setup() {
    validPracticeSettingCodesList = Arrays.asList(
      testDataHelper.practiceSettingCodeForMicrobiology,
      testDataHelper.practiceSettingCodeForBiochemistry,
      testDataHelper.practiceSettingCodeForPathology);
  }
  
  @Test
  public void encode() throws UnsupportedEncodingException {
    String encoded = new DocumentIdEncoderDecoder(patientId, from, to, null, validPracticeSettingCodesList).encode();
    Assert.assertTrue(encoded.length() > 0);
    Assert.assertTrue(encoded.length() <= 47); 
  }

  @Test
  public void encodeNoFrom() {
    String encoded = new DocumentIdEncoderDecoder(patientId, null, to, null, validPracticeSettingCodesList).encode();
    Assert.assertTrue(encoded.length() > 0);
  }

  @Test
  public void encodeNoTo() {
    String encoded = new DocumentIdEncoderDecoder(patientId, from, null, null, validPracticeSettingCodesList).encode();
    Assert.assertTrue(encoded.length() > 0);
  }

  @Test
  public void encodeNoFromTo() {
    String encoded = new DocumentIdEncoderDecoder(patientId, null, null, null, validPracticeSettingCodesList).encode();
    Assert.assertTrue(encoded.length() > 0);
  }

  @Test
  public void encodeDecode() throws UnsupportedEncodingException {
    assertEncodeDecode(patientId, from, to);
  }

  @Test
  public void encodeDecodeNoFrom() throws UnsupportedEncodingException {
    assertEncodeDecode(patientId, null, to);
  }

  @Test
  public void encodeDecodeNoTo() throws UnsupportedEncodingException {
    assertEncodeDecode(patientId, from, null);
  }

  @Test
  public void encodeDecodeNoFromTo() throws UnsupportedEncodingException {
    assertEncodeDecode(patientId, null, null);
  }
  
  @Test
  public void encodeDecodeWithPracticeSettingCodes() throws UnsupportedEncodingException {
    
    List<Code> practiceSettingsCodesList = Arrays.asList(
      ConverterUtil.convert(testDataHelper.practiceSettingCodeForMicrobiology),
      ConverterUtil.convert(testDataHelper.practiceSettingCodeForBiochemistry));
    
    String encoded = new DocumentIdEncoderDecoder(patientId, from, to, practiceSettingsCodesList, validPracticeSettingCodesList).encode();
    DocumentIdEncoderDecoder decoded = new DocumentIdEncoderDecoder(encoded, validPracticeSettingCodesList);

    Assert.assertEquals(practiceSettingsCodesList, decoded.getPracticeSettingCodes());
  }

  @Test
  public void encodeDecodeWithEmptyPracticeSettingCodes() throws UnsupportedEncodingException {
    
    List<Code> practiceSettingsCodesList = Collections.emptyList();
    
    String encoded = new DocumentIdEncoderDecoder(patientId, from, to, practiceSettingsCodesList, validPracticeSettingCodesList).encode();
    DocumentIdEncoderDecoder decoded = new DocumentIdEncoderDecoder(encoded, validPracticeSettingCodesList);

    Assert.assertEquals(practiceSettingsCodesList, decoded.getPracticeSettingCodes());
  }

  @Test
  public void encodeDecodeWithDuplicatesInPracticeSettingCodes() throws UnsupportedEncodingException {
    
    List<Code> practiceSettingsCodesList = Arrays.asList(
      ConverterUtil.convert(testDataHelper.practiceSettingCodeForMicrobiology),
      ConverterUtil.convert(testDataHelper.practiceSettingCodeForMicrobiology),
      ConverterUtil.convert(testDataHelper.practiceSettingCodeForBiochemistry));
    List<Code> expectedSettingsCodesList = Arrays.asList(
      ConverterUtil.convert(testDataHelper.practiceSettingCodeForMicrobiology),
      ConverterUtil.convert(testDataHelper.practiceSettingCodeForBiochemistry));
    
    String encoded = new DocumentIdEncoderDecoder(patientId, from, to, practiceSettingsCodesList, validPracticeSettingCodesList).encode();
    DocumentIdEncoderDecoder decoded = new DocumentIdEncoderDecoder(encoded, validPracticeSettingCodesList);
    
    Assert.assertEquals(expectedSettingsCodesList, decoded.getPracticeSettingCodes());
  }

  @Test(expected=IllegalArgumentException.class)
  public void illegalPatientId() throws UnsupportedEncodingException {
    new DocumentIdEncoderDecoder("12A", null, null, null, validPracticeSettingCodesList);
  }

  private void assertEncodeDecode(String id, DateTime fromDateTime, DateTime toDateTime) {
    String encoded = new DocumentIdEncoderDecoder(id, fromDateTime, toDateTime, null, validPracticeSettingCodesList).encode();
    DocumentIdEncoderDecoder decoded = new DocumentIdEncoderDecoder(encoded,validPracticeSettingCodesList);
    Assert.assertEquals(id, decoded.getPatientId());
    Assert.assertEquals(fromDateTime, decoded.getFrom());
    Assert.assertEquals(toDateTime, decoded.getTo());
  }
}
