/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.common;

import org.joda.time.DateTime;
import org.joda.time.DateTimeZone;
import org.junit.Test;

import org.junit.Assert;

public class PersonIdInfoDerivationTest {
	
	@Test
	public void minSupportedAgeTest() throws Exception {
		DateTime birthDay = createBirthday();
		assertDeriveDateOfBirth(birthDay);
	}
	
	@Test
	public void maxSupportedAgeTest() throws Exception {
		DateTime birthDay = createBirthday().minusYears(100).plusDays(1);
		assertDeriveDateOfBirth(birthDay);
	}
	
	@Test
	public void tooOldTest() throws Exception {
		DateTime birthDay = createBirthday().minusYears(100);
		assertUnequalDeriveDateOfBirth(birthDay);
	}
	
	@Test(expected=PatientIdDerivationException.class)
	public void patientIdNullTest() throws Exception {
		PersonIdInfoDerivation.deriveDateOfBirth(null);
	}
	
	@Test(expected=PatientIdDerivationException.class)
	public void patientIdEmptyTest() throws Exception {
		PersonIdInfoDerivation.deriveDateOfBirth("");
	}
	
	@Test(expected=PatientIdDerivationException.class)
	public void patientIdTooShortTest() throws Exception {
		PersonIdInfoDerivation.deriveDateOfBirth("010112123");
	}
	
	@Test(expected=PatientIdDerivationException.class)
	public void patientIdTooLongTest() throws Exception {
		PersonIdInfoDerivation.deriveDateOfBirth("01011212341");
	}
	
	@Test(expected=PatientIdDerivationException.class)
	public void lettersInPatientIdTest() throws Exception {
		PersonIdInfoDerivation.deriveDateOfBirth("u101011234");
	}
	
	private DateTime createBirthday() {
		DateTime now = DateTime.now(DateTimeZone.UTC);
		return new DateTime(now.getYear(), now.getMonthOfYear(), now.getDayOfMonth(), 0, 0, DateTimeZone.UTC);
	}
	
	private static void assertDeriveDateOfBirth(DateTime time) throws Exception {
		Assert.assertEquals(time,
				PersonIdInfoDerivation.deriveDateOfBirth(dateTimeToId(time)));
	}
	
	private static void assertUnequalDeriveDateOfBirth(DateTime time) throws Exception {
		Assert.assertNotSame(time,
				PersonIdInfoDerivation.deriveDateOfBirth(dateTimeToId(time)));
	}
	
	private static String dateTimeToId(DateTime time) {
		return time.toString("ddMMyy") + "1234"; // add control ciffer
	}
	
	@Test
	public void determineGenderSuccessTest() throws PatientIdDerivationException {
		// Last digit is equal
		Assert.assertEquals("F", PersonIdInfoDerivation.determinePatientGender("0101011230"));
		Assert.assertEquals("F", PersonIdInfoDerivation.determinePatientGender("0101011232"));
		Assert.assertEquals("F", PersonIdInfoDerivation.determinePatientGender("0101011234"));
		Assert.assertEquals("F", PersonIdInfoDerivation.determinePatientGender("0101011236"));
		Assert.assertEquals("F", PersonIdInfoDerivation.determinePatientGender("0101011238"));
		// Last digit is unequal
		Assert.assertEquals("M", PersonIdInfoDerivation.determinePatientGender("0101011231"));
		Assert.assertEquals("M", PersonIdInfoDerivation.determinePatientGender("0101011233"));
		Assert.assertEquals("M", PersonIdInfoDerivation.determinePatientGender("0101011235"));
		Assert.assertEquals("M", PersonIdInfoDerivation.determinePatientGender("0101011237"));
		Assert.assertEquals("M", PersonIdInfoDerivation.determinePatientGender("0101011239"));
		// Testing max day and month
		Assert.assertEquals("F", PersonIdInfoDerivation.determinePatientGender("3112011230"));
		Assert.assertEquals("M", PersonIdInfoDerivation.determinePatientGender("3112011231"));
	}

	@Test
	public void determineGenderDisregardsDate() throws PatientIdDerivationException {
	    PersonIdInfoDerivation.determinePatientGender("3201011234");
	}

	@Test(expected=PatientIdDerivationException.class)
	public void determineGenderFailTestNull() throws PatientIdDerivationException {
		PersonIdInfoDerivation.determinePatientGender(null);
	}

	@Test(expected=PatientIdDerivationException.class)
	public void determineGenderFailTestEmpty() throws PatientIdDerivationException {
	    PersonIdInfoDerivation.determinePatientGender("");
	}

	@Test(expected=PatientIdDerivationException.class)
	public void determineGenderFailTest1() throws PatientIdDerivationException {
	    PersonIdInfoDerivation.determinePatientGender("01010112341");
	}

	@Test(expected=PatientIdDerivationException.class)
	public void determineGenderFailTest2() throws PatientIdDerivationException {
	    PersonIdInfoDerivation.determinePatientGender("010101123");
	}

}
