package dk.nsi.dds.projects.sxa.common.security;

import dk.nsi.dds.projects.sxa.common.configuration.SXAConfigKeys;
import dk.nsi.util.Config;
import dk.nsi.util.ConfigException;
import org.apache.cxf.binding.soap.SoapFault;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.runners.MockitoJUnitRunner;

import javax.interceptor.InvocationContext;
import javax.servlet.http.HttpServletRequest;
import javax.xml.ws.WebServiceContext;
import javax.xml.ws.handler.MessageContext;

import static org.junit.Assert.*;

@RunWith(MockitoJUnitRunner.class)
public class AccessValidationInterceptorTest {
    @InjectMocks
    private AccessValidationInterceptor subject;

    @Mock
    private WebServiceContext webServiceContext;

    @Mock
    private Config config;

    @Mock
    private InvocationContext context;

    private static final String credential = "00000000-0000-0000-0000-000000000000";
    private static final String credentialHeader = "X-Labsvar-Credential";

    @Before
    public void setup() {
        Mockito.when(config.getProperty(SXAConfigKeys.SXA_CREDENTIAL)).thenReturn(credential);
        Mockito.when(config.getProperty(SXAConfigKeys.SXA_CREDENTIAL_HEADER)).thenReturn(credentialHeader);

        setupWebServiceContext(credential);
    }

    @Test
    public void validateAccess_success() throws Exception {
        // Arrange
        String s = "success";
        Mockito.when(context.proceed()).thenReturn(s);

        // Act
        Object result = subject.validateAccess(context);

        // Assert
        assertEquals(s, result);
    }

    @Test
    public void validateAccess_configurationOmitted_skipsValidation() throws Exception {
        // Arrange
        String s = "success";
        Mockito.when(context.proceed()).thenReturn(s);

        Mockito.when(config.getProperty(SXAConfigKeys.SXA_CREDENTIAL)).thenThrow(new ConfigException());
        Mockito.when(config.getProperty(SXAConfigKeys.SXA_CREDENTIAL_HEADER)).thenThrow(new ConfigException());

        // Act
        Object result = subject.validateAccess(context);

        // assert
        assertEquals(s, result);
    }

    @Test(expected = IllegalStateException.class)
    public void validateAccess_configurationIncomplete_failure() throws Exception {
        // Arrange
        String s = "success";
        Mockito.when(context.proceed()).thenReturn(s);

        Mockito.when(config.getProperty(SXAConfigKeys.SXA_CREDENTIAL_HEADER)).thenThrow(new ConfigException());

        // Act
        Object result = subject.validateAccess(context);

        // Assert
    }

    @Test(expected = SoapFault.class)
    public void validateAccess_headerMissingFromRequest_failure() throws Exception {
        // Arrange
        String s = "success";
        Mockito.when(context.proceed()).thenReturn(s);

        String cred = null;
        setupWebServiceContext(cred);

        // Act
        Object result = subject.validateAccess(context);

        // Assert
        assertEquals(s, result);
    }

    @Test(expected = SoapFault.class)
    public void validateAccess_wrongHeader_failure() throws Exception {
        // Arrange
        String s = "success";
        Mockito.when(context.proceed()).thenReturn(s);

        String cred = "11111111-1111-1111-1111-111111111111";
        setupWebServiceContext(cred);

        // Act
        Object result = subject.validateAccess(context);

        // Assert
        assertEquals(s, result);
    }

    private void setupWebServiceContext(String credential) {
        HttpServletRequest request = Mockito.mock(HttpServletRequest.class);
        Mockito.when(request.getHeader(credentialHeader)).thenReturn(credential);

        MessageContext messageContext = Mockito.mock(MessageContext.class);
        Mockito.when(messageContext.containsKey(Mockito.anyString())).thenReturn(true);
        Mockito.when(messageContext.get(Mockito.anyString())).thenReturn(request);

        Mockito.when(webServiceContext.getMessageContext()).thenReturn(messageContext);
    }
}