/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (SDS), Denmark
 * (http://sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (SDS), Denmark (http://sundhedsdatastyrelsen.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.common.testutil;

import java.math.BigInteger;
import java.util.Arrays;
import java.util.List;

import org.junit.Assert;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Code;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Identifiable;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.LocalizedString;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Organization;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Person;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.rim.InternationalStringType;

import dk.nsi.dds.projects.sxa.common.configuration.SXACodeValue;
import dk.nsi.dds.projects.sxa.common.util.ConverterUtil;

public class AssertUtil {
	public static void assertTrue(boolean... conditions) {
		for (boolean condition : conditions) {
			Assert.assertTrue(condition);
		}
	}
	
	public static void assertFalse(boolean... conditions) {
		for (boolean condition : conditions) {
			Assert.assertFalse(condition);
		}
	}
	
	/*Assert Equals*/
	public static void assertEquals(SXACodeValue expected, Code actual) {
		assertEquals(expected, actual, false);
	}
	
	public static void assertEquals(SXACodeValue expected, Code actual, boolean ignoreLangAndCharset) {
		assertEquals(ConverterUtil.convert(expected), actual, ignoreLangAndCharset);
	}

	public static void assertEquals(Organization expected, Organization actual) {
		Assert.assertEquals(expected, actual);
	}
	
	public static void assertEquals(Identifiable expected, Identifiable actual) {
		Assert.assertEquals(expected, actual);
	}

	public static void assertEqualsIgnoreLangAndCharset(Code expected, Code actual) {
		assertEquals(expected, actual, true);
	}
	
	public static void assertEquals(Code expected, Code actual) {
		assertEquals(expected, actual, false);
	}
	
	public static void assertEquals(Code expected, Code actual, boolean ignoreLangAndCharset) {
		if(expected == null) {
			Assert.assertNull(actual);
		} else {
			Assert.assertEquals(expected.getCode(), actual.getCode());
			Assert.assertEquals(expected.getSchemeName(), actual.getSchemeName());
			if(expected.getDisplayName() == null) {
				Assert.assertNull(expected.getDisplayName());
			} else {
				assertEquals(expected.getDisplayName(), actual.getDisplayName(), ignoreLangAndCharset);
			}
		}
	}

	public static void assertEquals(LocalizedString expected, LocalizedString actual) {
		assertEquals(expected, actual, false);
	}
	
	public static void assertEquals(LocalizedString expected, LocalizedString actual, boolean ignoreLangAndCharset) {
		if(expected == null) {
			Assert.assertNull(actual);
		} else {
			Assert.assertEquals(expected.getValue(), actual.getValue());
			if(ignoreLangAndCharset == false) {
				Assert.assertEquals(expected.getLang(), actual.getLang());
				Assert.assertEquals(expected.getCharset(), actual.getCharset());
			}
		}	
	}

	public static void assertEquals(Person expected, Person actual) {
		Assert.assertEquals(expected, actual);		
	}

	public static void assertEquals(String expected, String actual) {
		Assert.assertEquals(expected, actual);			
	}
	
	public static void assertEquals(String expected, InternationalStringType actual) {
		if(expected == null) {
			assertNull(actual);
		} else {
			assertNotNull(actual);
			assertNotNull(actual.getLocalizedString());
			assertEquals(1, actual.getLocalizedString().size());
			Assert.assertEquals(expected, actual.getLocalizedString().get(0).getValue());
		}			
	}
	
	public static void assertOneOf(InternationalStringType actual, String... expected) {
		assertOneOf(actual, false, expected);
	}
	
	public static void assertOneOf(InternationalStringType actual, boolean allowNull, String... expected) {
		if(expected == null || expected.length == 0) {
			assertNull(actual);
		} else if (actual == null && allowNull){
			assertNull(actual);
		} else {
			assertNotNull(actual);
			assertNotNull(actual.getLocalizedString());
			assertEquals(1, actual.getLocalizedString().size());
			assertOneOf(actual.getLocalizedString().get(0).getValue(), allowNull, expected);
		}			
	}
	
	public static void assertOneOf(String actual, String... expected) {
		assertOneOf(actual, false, expected);
	}
	
	public static void assertOneOf(String actual, boolean allowNull, String... expected) {
		if(expected == null || expected.length == 0) {
			assertNull(actual);
		} else if (actual == null && allowNull){
			assertNull(actual);
		} else {
			assertNotNull(actual);
			Arrays.sort(expected);
			if(Arrays.binarySearch(expected, actual) < 0) {
				throw new AssertionError("expected:<" + Arrays.toString(expected) + "> but was:<"+ actual + ">");
			}
		}			
	}
	
	public static void assertEquals(long expected, long actual) {
		Assert.assertEquals(expected, actual);			
	}

	public static void assertEquals(long expected, BigInteger actual) {
		assertEquals(expected, actual.longValue());
	}

	public static void assertNull(Object obj) {
		Assert.assertNull(obj);		
	}
	
	public static void assertNotNull(Object obj) {
		Assert.assertNotNull(obj);		
	}

	public static void assertEquals(Class<?> expected, Class<?> actual) {
		Assert.assertEquals(expected, actual);
	}

	public static void assertEmpty(List<?> list) {
		assertNotNull(list);
		assertEquals(0, list.size());
	}

	public static <E> void assertEquals(List<E> list1, List<E> list2) {
		Assert.assertEquals(list1, list2);
	}

	public static void assertContains(String expected, String actual) {
		Assert.assertTrue("Forventede at finde "+expected, actual.contains(expected));
	}

	public static void assertAntal(int expected, String afTekst, String iTekst) {
		int actual = 0;
		int index = 0;
		while((index = iTekst.indexOf(afTekst, index == 0 ? 0 : ++index)) > 0) {
			actual++;
		}
		Assert.assertEquals("Forventede at finde "+afTekst+" "+expected+" antal gange. Fandt den "+actual+" antal gange", expected, actual);
	}
}
