/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (SDS), Denmark
 * (http://sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (SDS), Denmark (http://sundhedsdatastyrelsen.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.common.xds.impl;

import static org.junit.Assert.fail;

import org.junit.Assert;
import org.junit.Test;

public class ValueEncoderDecoderOfErrorCodeContextTest {
	private static final String VALID_ENCODED_PATIENT_ID = "[Document Unique Id: doc#1234, Home Community Id: 1234.5678, Repository Unique Id: 2468.1357]";
	private static final String INVALID_ENCODED_PATIENT_ID = "[Document Unique Id:            , Home Community Id: ABC, Repository Unique Id: XYZ]";

	@Test
	public void testDecodeNullValues() {
		try {
			ValueEncoderDecoder.extractDocumentUniqueId(null);
			fail("Expected to fail with NullpointerException");
		} catch (NullPointerException npe) {
			// Expected
		}

		try {
			ValueEncoderDecoder.extractHomeCommunityId(null);
			fail("Expected to fail with NullpointerException");
		} catch (NullPointerException npe) {
			// Expected
		}

		try {
			ValueEncoderDecoder.extractRepositoryUniqueId(null);
			fail("Expected to fail with NullpointerException");
		} catch (NullPointerException npe) {
			// Expected
		}
	}

	@Test
	public void testDecodeValidValues() {
		String documentUID = ValueEncoderDecoder.extractDocumentUniqueId(VALID_ENCODED_PATIENT_ID);
		Assert.assertEquals("doc#1234", documentUID);

		String homeCommunityId = ValueEncoderDecoder.extractHomeCommunityId(VALID_ENCODED_PATIENT_ID);
		Assert.assertEquals("1234.5678", homeCommunityId);

		String repositoryUID = ValueEncoderDecoder.extractRepositoryUniqueId(VALID_ENCODED_PATIENT_ID);
		Assert.assertEquals("2468.1357", repositoryUID);
	}
	
	@Test
	public void testDecodeInvalidValues() {
		String documentUID = ValueEncoderDecoder.extractDocumentUniqueId(INVALID_ENCODED_PATIENT_ID);
		Assert.assertEquals("", documentUID);

		String homeCommunityId = ValueEncoderDecoder.extractHomeCommunityId(INVALID_ENCODED_PATIENT_ID);
		Assert.assertNull(homeCommunityId);

		String repositoryUID = ValueEncoderDecoder.extractRepositoryUniqueId(INVALID_ENCODED_PATIENT_ID);
		Assert.assertNull(repositoryUID);
	}

	@Test
	public void testEncodeValidValues() {
		final String documentUID = "doc#1234";
		final String homeCommunityId = "1234.5678";
		final String repositoryUID = "2468.1357";
		String encodedErrorCodeContext = ValueEncoderDecoder.encodeErrorCodeContext(documentUID, homeCommunityId, repositoryUID);
		Assert.assertEquals("[Document Unique Id: doc#1234, Home Community Id: 1234.5678, Repository Unique Id: 2468.1357]", encodedErrorCodeContext);
	}

	@Test
	public void testEncodeNullValues() {
		final String documentUID = "doc#1234";
		final String homeCommunityId = "1234.5678";
		final String repositoryUID = "2468.1357";
		String encodedErrorCodeContext;
		encodedErrorCodeContext = ValueEncoderDecoder.encodeErrorCodeContext(null, homeCommunityId, repositoryUID);
		Assert.assertEquals("[Document Unique Id: null, Home Community Id: 1234.5678, Repository Unique Id: 2468.1357]", encodedErrorCodeContext);

		encodedErrorCodeContext = ValueEncoderDecoder.encodeErrorCodeContext(documentUID, null, repositoryUID);
		Assert.assertEquals("[Document Unique Id: doc#1234, Home Community Id: null, Repository Unique Id: 2468.1357]", encodedErrorCodeContext);

		encodedErrorCodeContext = ValueEncoderDecoder.encodeErrorCodeContext(documentUID, homeCommunityId, null);
		Assert.assertEquals("[Document Unique Id: doc#1234, Home Community Id: 1234.5678, Repository Unique Id: null]", encodedErrorCodeContext);
	}
}
