/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (SDS), Denmark
 * (http://sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (SDS), Denmark (http://sundhedsdatastyrelsen.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.common.xds.impl;

import static org.junit.Assert.fail;

import org.junit.Assert;
import org.junit.Test;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.AssigningAuthority;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Organization;

import dk.nsi.dds.projects.sxa.common.xds.UnexpectedDataException;
import dk.nsi.dds.projects.sxa.common.xds.UnexpectedEncodingException;

public class ValueEncoderDecoderOfOrganizationTest {
	private static final Organization VALID_DECODED_ORGANIZATION = new Organization("name", "id", new AssigningAuthority("uid", "type"));
	private static final String VALID_ENCODED_ORGANIZATION = "name^^^^^&uid&type^^^^id";
	private static final String VALID_ENCODED_ORGANIZATION_OLD_ENCODING = "name^^^^^&123.456&ISO^^^^abc=xyz";

	@Test
	public void testDecodeNull() {
		String encodedOrganization = null;
		try {
			ValueEncoderDecoder.decodeOrganisation(encodedOrganization);
			fail("Expected to throw IllegalArgumentException");
		} catch (IllegalArgumentException e) {
			Assert.assertEquals("encodedOrganisation missing", e.getMessage());
		}
	}

	@Test
	public void testDecodeValidEncoded() {
		String encodedOrganization = VALID_ENCODED_ORGANIZATION;
		Organization organization = ValueEncoderDecoder.decodeOrganisation(encodedOrganization);
		Assert.assertEquals("Organization[name=name,assigningAuthority=AssigningAuthority[universalId=uid,universalIdType=type],idNumber=id]", organization.toString());

		Assert.assertEquals("name", organization.getOrganizationName());
		Assert.assertEquals("id", organization.getIdNumber());
		Assert.assertNotNull(organization.getAssigningAuthority());
		Assert.assertEquals("uid", organization.getAssigningAuthority().getUniversalId());
		Assert.assertEquals("type", organization.getAssigningAuthority().getUniversalIdType());
	}

	@Test
	public void testDecodeNullOldEncoding() throws UnexpectedEncodingException {
		String encodedOrganization = null;
		try {
			ValueEncoderDecoder.decodeOrganisationOldEncoding(encodedOrganization);
			fail("Expected to throw IllegalArgumentException");
		} catch (IllegalArgumentException e) {
			Assert.assertEquals("encodedOrganisation missing", e.getMessage());
		}
	}
	
	@Test
	public void testDecodeInvalidOldEncoding() {
		String encodedOrganization = "123^^^^^^^^^^456";
		try {
			ValueEncoderDecoder.decodeOrganisationOldEncoding(encodedOrganization);
			fail("Expected to throw UnexpectedEncodingException");
		} catch (UnexpectedEncodingException e) {
			Assert.assertEquals("Organisation not encoded in accepted formatting", e.getMessage());
		}
	}

	@Test
	public void testDecodeValidOldEncoding() throws UnexpectedEncodingException {
		// decodeOrganisationOldEncoding
		
		String encodedOrganization = VALID_ENCODED_ORGANIZATION_OLD_ENCODING;
		Organization organization = ValueEncoderDecoder.decodeOrganisationOldEncoding(encodedOrganization);
		Assert.assertEquals("Organization[name=name,assigningAuthority=AssigningAuthority[universalId=123.456.abc,universalIdType=<null>],idNumber=xyz]", organization.toString());

		Assert.assertEquals("name", organization.getOrganizationName());
		Assert.assertEquals("xyz", organization.getIdNumber());
		Assert.assertNotNull(organization.getAssigningAuthority());
		Assert.assertEquals("123.456.abc", organization.getAssigningAuthority().getUniversalId());
		Assert.assertNull(organization.getAssigningAuthority().getUniversalIdType());
	}

	@Test
	public void testEncodeNull() {
		Organization organization = null;
		try {
			ValueEncoderDecoder.encodeOrganisation(organization);
			fail("Expected to throw UnexpectedDataException");
		} catch (UnexpectedDataException e) {
			Assert.assertEquals("Values missing for organisation", e.getMessage());
		}
	}

	@Test
	public void testEncodeValid() throws UnexpectedDataException {
		Organization organization = VALID_DECODED_ORGANIZATION;
		String encodeOrganisation = ValueEncoderDecoder.encodeOrganisation(organization);
		Assert.assertEquals(VALID_ENCODED_ORGANIZATION, encodeOrganisation);
	}
}
